# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved
import ctypes
import enum
import sys
import time
from builtins import bytes

import numpy
import numpy as np
import six

from braintech.drivers.blinker.blinker import AbstractBlinker

is_py2 = sys.version[0] == '2'


class MessageType(enum.IntEnum):
    ENABLE_FIELD = 1
    DISABLE_FIELD = 2
    CHANNEL_DATA = 3
    FLUSH = 4
    FREQUENCY = 5
    START = 6
    STOP = 7
    INTENSITY = 8
    TIME_SYNC = 9
    DEFINE_FIELD = 10
    SET_FIELDS = 11
    ERROR_MESSAGE = 12
    SYNCHRO_OUTPUT = 13
    VERSION_REQ = 14
    TEMP_REQ = 15
    SET_MODE = 16


class BlinkerState(enum.IntEnum):
    ON = 0
    DEMO = 1
    OFF = 2


class Command(ctypes.Structure):
    START_BYTE = 0xAA
    _pack_ = 1
    _fields_ = [
        ('start', ctypes.c_byte),
        ('type', ctypes.c_byte),
    ]
    receivable_commands = {}

    def __init__(self, *args, **kwargs):
        self.first_byte_timestamp = local_clock()
        super(Command, self).__init__(self.START_BYTE, self.TYPE.value, *args, **kwargs)

    @classmethod
    def read(cls, stream):
        while stream.is_open:
            start = bytes(stream.read(1))
            if start and start[0] == cls.START_BYTE:
                break
        first_byte = local_clock()
        msg_type = stream.read(1)
        while msg_type == b'':
            msg_type = stream.read(1)
        msg_type = bytes(msg_type)[0]
        try:
            MsgClass = Command.receivable_commands[msg_type]
        except KeyError:
            raise ValueError("Unknown Command: %s", msg_type)
        buffer = bytes([cls.START_BYTE, msg_type])
        while len(buffer) < ctypes.sizeof(MsgClass):
            buffer += stream.read(ctypes.sizeof(MsgClass) - len(buffer))
        msg = MsgClass.from_buffer_copy(buffer)
        msg.first_byte_timestamp = first_byte
        return msg

    def __str__(self):
        args = ["%s" % getattr(self, f[0]) for f in self._fields_]
        return self.__class__.__name__ + "(" + ", ".join(args) + ")"


def receivable(klass):
    Command.receivable_commands[klass.TYPE] = klass
    return klass


class Start(Command):
    TYPE = MessageType.START


class Stop(Command):
    TYPE = MessageType.STOP


@receivable
class Flush(Command):
    TYPE = MessageType.FLUSH
    _fields_ = [
        ('last_update', ctypes.c_uint32),
        ('period', ctypes.c_uint16)
    ]


class Field(Command):
    _fields_ = [
        ('field', ctypes.c_uint16)
    ]


class Synchro(Field):
    TYPE = MessageType.SYNCHRO_OUTPUT


class Enable(Field):
    TYPE = MessageType.ENABLE_FIELD


class Disable(Field):
    TYPE = MessageType.DISABLE_FIELD
    _fields_ = [
        ('brightness', ctypes.c_byte)
    ]

    def __init__(self, field, brightness):
        super(Disable, self).__init__(field, int(brightness * 255))

    def get_brightness(self):
        return self.brightness / 255.0


@receivable
class TimeSync(Command):
    TYPE = MessageType.TIME_SYNC
    _fields_ = [
        ('rx_time', ctypes.c_uint32),
        ('tx_time', ctypes.c_uint32),
    ]


class Frequency(Field):
    FREQUENCY_UNIT = 0.01
    PHASE_UNIT = 0.001
    TYPE = MessageType.FREQUENCY
    MAX_PHASE = AbstractBlinker.MAX_PHASE
    _fields_ = [
        ('frequency', ctypes.c_uint16),
        ('phase', ctypes.c_uint16)
    ]

    def __init__(self, field, frequency, phase):
        phase = phase / self.MAX_PHASE
        super(Frequency, self).__init__(field, int(frequency / self.FREQUENCY_UNIT),
                                        int((phase - int(phase)) / self.PHASE_UNIT))

    def get_frequency(self):
        return self.frequency * self.FREQUENCY_UNIT

    def get_phase(self):
        return self.phase * self.PHASE_UNIT * self.MAX_PHASE


class FieldData(Field):
    TYPE = MessageType.CHANNEL_DATA
    _fields_ = [
        ('length', ctypes.c_uint16)
    ]

    def __init__(self, field, channel_data):
        super(FieldData, self).__init__(field, len(channel_data))
        self.channel_data[:] = (np.array(channel_data) * 255).astype(dtype=numpy.uint8)

    def __new__(cls, field, channel_data, *args):
        if cls is not FieldData:
            return super(FieldData, cls).__new__(cls, field, channel_data, *args)
        CHANNEL_DATA = ctypes.c_ubyte * len(channel_data)

        class FieldData_(FieldData):
            _fields_ = [
                ('channel_data', CHANNEL_DATA)
            ]

        return FieldData_(field, channel_data)


class Intensity(Command):
    TYPE = MessageType.INTENSITY
    _fields_ = [
        ('intensity', ctypes.c_byte)
    ]
    MAX_INTENSITY = 80

    def __init__(self, intensity, max_intensity=MAX_INTENSITY):
        super(Intensity, self).__init__(min(int(intensity * max_intensity), max_intensity))
        self.max_intensity = max_intensity

    def get_intensity(self):
        return self.intensity / self.max_intensity


class SetFields(Command):
    TYPE = MessageType.SET_FIELDS
    _fields_ = [
        ('fields_num', ctypes.c_uint16)
    ]


class DefineField(Field):
    TYPE = MessageType.DEFINE_FIELD
    _fields_ = [
        ('length', ctypes.c_uint16)
    ]

    def __init__(self, field, channel_indexes):
        super(DefineField, self).__init__(field, len(channel_indexes) * 2)
        self.channel_indexes[:] = np.array(channel_indexes).astype(dtype=numpy.uint16)

    def __new__(cls, field, channel_indexes, *args):
        if cls is not DefineField:
            return super(DefineField, cls).__new__(cls, field, channel_indexes, *args)
        CHANNEL_INDEXES = ctypes.c_uint16 * len(channel_indexes)

        class DefineField_(DefineField):
            _fields_ = [
                ('channel_indexes', CHANNEL_INDEXES)
            ]

        return DefineField_(field, channel_indexes)


@receivable
class Error(Field):
    TYPE = MessageType.ERROR_MESSAGE
    _fields_ = [
        ('errno', ctypes.c_uint16)
    ]

    class Error(enum.IntEnum):
        NO_SUCH_FIELD = 0x01

    def get_error(self):
        return Error.Error(self.errno)


@receivable
class FirmwareVersion(Command):
    TYPE = MessageType.VERSION_REQ
    _fields_ = [
        ('version', ctypes.c_uint16),
    ]


class SetState(Command):
    TYPE = MessageType.SET_MODE
    _fields_ = [
        ('state', ctypes.c_uint8),
    ]


def to_bytes(command):
    if six.PY2:
        from ctypes import create_string_buffer, memmove, sizeof, addressof
        buffer = create_string_buffer(sizeof(command))
        memmove(buffer, addressof(command), sizeof(command))
        return buffer.raw
    else:
        return bytes(command)


def local_clock():
    if is_py2:
        return time.clock()
    else:
        return time.perf_counter()
