#!/usr/bin/env python3
# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.

"""Module provides logic for BCI Appliance. Contains single :class 'LogicBCIAppliance'."""
import asyncio

from braintech.drivers.blinker.blinker import CHANNELS_NUMBER
from braintech.obci.core.broker import messages
from braintech.obci.experiment.peer.configured_peer import ConfiguredPeer
from braintech.obci.core.utils.properties import param_property, join_strings, split_ints, split_floats
from .blinker_peer import OBCIApplianceConfigurationMsg

__all__ = ('LogicBCIAppliance',)


class LogicBCIAppliance(ConfiguredPeer):
    """Peer provides logic for BCI Appliance."""

    def __init__(self, urls, **kwargs):
        """Peer keeps configuration parameters for BCIAppliance experiment."""
        super().__init__(urls, **kwargs)
        self.managed_peer_url = None

    async def _dependencies_are_ready(self):
        await self._prepare_peer()
        await self._set_fields_config()

    @property
    def managed_peer_id(self):
        """Return blinker peer id from launch dependencies."""
        return self.config.launch_deps['managed_peer']

    fields_function = param_property('fields_function')

    number_of_trials = param_property('number_of_trials', int)

    duty_cycle = param_property('duty_cycle', float)

    trial_duration = param_property('trial_duration', float)

    break_duration = param_property('break_duration', float)

    luminosity = param_property('luminosity', float)

    phase_offset = param_property('phase_offset', split_ints, join_strings)

    active_fields = param_property('active_fields', split_ints, join_strings)

    frequency = param_property('frequency', split_floats, join_strings)

    async def _start(self):
        await super()._start()
        await self._run_experiment()

    async def _prepare_peer(self):
        """Extract ids of experiment peers from scenario config. Retrieve peer urls from Broker."""
        resp = await self.ask_broker(messages.PeerUrlQuery(target=self.managed_peer_id,
                                                           target_url='',
                                                           sender_id=self.peer_id))
        if resp.target_url and resp.target == self.managed_peer_id:
            self.managed_peer_url = resp.target_url

    async def _run_experiment(self):
        for _ in range(self.number_of_trials):
            await self._start_peer()
            await asyncio.sleep(self.trial_duration + self.break_duration)  # required
            await self._stop_peer()

    async def _set_fields_config(self):
        self.field_config = dict()
        for field_number in range(CHANNELS_NUMBER):
            if field_number in self.active_fields:
                self.field_config[field_number] = {'field_number': field_number,
                                                   'action': self.fields_function,
                                                   'frequency': self.frequency[field_number],
                                                   'phase_offset': self.phase_offset[field_number],
                                                   'luminosity': self.luminosity,
                                                   'duty_cycle': self.duty_cycle
                                                   }
            else:
                self.field_config[field_number] = {'field_number': field_number,
                                                   'action': 'still',
                                                   'luminosity': self.luminosity
                                                   }

    async def _start_peer(self):
        """Send configuration and start messages to BCIAppliancePeer."""
        message = OBCIApplianceConfigurationMsg(fields_config=self.field_config)
        await self.ask_peer(self.managed_peer_url, message)
        message = messages.PeerControlMessage(peer_id=self.managed_peer_id, action='start')
        await self.ask_peer(self.managed_peer_url, message)

    async def _stop_peer(self):
        """Send stop message to BCIAppliancePeer."""
        message = messages.PeerControlMessage(peer_id=self.managed_peer_id, action='stop')
        await self.ask_peer(self.managed_peer_url, message)


if __name__ == "__main__":
    pass
