# -*- coding: utf-8 -*-
# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.

import enum
import sys
import termios
import threading
import time
from collections import namedtuple
from functools import reduce
from logging import getLogger

import numpy
import numpy as np
import serial

from braintech.drivers.blinker import commands
from braintech.drivers.blinker.blinker import AbstractBlinker, run_demo, FieldConfig
from braintech.drivers.blinker.blinker_preview import Emulator
from braintech.drivers.blinker.commands import Command, to_bytes, local_clock, Synchro

is_py2 = sys.version[0] == '2'
if is_py2:
    from Queue import Queue, Empty
else:
    from queue import Queue, Empty

LOGGER = getLogger(__name__)


class PTPTimes(namedtuple('PTPTimes_', 't1_tx,t1_rx,t2_tx,t2_rx')):
    def time_offset(self):
        return ((self.t1_rx - self.t1_tx) - (self.t2_rx - self.t2_tx)) / 2.0

    def duration(self):
        return self.t2_rx - self.t1_tx


MICROSECONDS = 0.000001
MAX_FIRMWARE_VERSION = 5


class BlinkerException(Exception):
    pass


class SlimBlinkerDevice(serial.Serial):
    MAX_BLINKER_TIME = 4000
    RECEIVE_TIMEOUT = 2

    class TimestampMethod(enum.Enum):
        TRANSFER_DURATION = 'duration'
        CLOCK_OFFSET = 'clock_offset'

    def __init__(self, port, preview=None):
        super(SlimBlinkerDevice, self).__init__(None, timeout=1)
        self.dtr = False
        self.rts = False
        self._closed = False
        self.port = port
        try:
            self.open()
        except termios.error as e:
            if 'ioctl' in str(e):
                raise BlinkerException("IOCTL ERROR OCCURED:\n"
                                       "TO FIX THIS ERROR TRY RESTARTING PC OR "
                                       "DISCONNECT BLINKER POWER, WRITETHIS IN TERMINAL: \n\n"
                                       "sudo rmmod -f cdc_acm\nsudo modprobe cdc_acm\n"
                                       "AND RECONNECT") from e
            else:
                raise

        self._queue = []
        self._preview = preview
        self._receive_queue = Queue()
        self._receive_thread = threading.Thread(target=self._receive, name=self.__class__.__name__ + 'ReceiveThread')
        self._receive_thread.daemon = True
        self._receive_thread.start()
        self._time_period = 0
        self._last_blinker_time = 0
        self._duration = 0
        self.version = 0
        self._determine_firmware_version()
        self.timestamp_method = self.TimestampMethod.TRANSFER_DURATION
        try:
            self.synchronize(50)
        except Exception:
            self.close()
            raise

    def _blinker_time(self, blinker_ts):
        blinker_ts *= MICROSECONDS
        if blinker_ts < self._last_blinker_time:
            self._time_period += 1
        self._last_blinker_time = blinker_ts
        return blinker_ts + self._time_period * self.MAX_BLINKER_TIME

    def _unix_time_using_clock_offset(self, blinker_time):
        blinker_time_in_local_clock = blinker_time - self._blinker_offset
        return self._local_clock_to_unix_time(blinker_time_in_local_clock)

    def _unix_time_using_transfer_duration(self, receive_time):
        return self._local_clock_to_unix_time(receive_time - self._duration)

    def _local_clock_to_unix_time(self, l_clock):
        return l_clock + (time.time() - local_clock())

    def _determine_firmware_version(self):
        command = self._prepare_bytes(commands.FirmwareVersion())
        self.write(command)
        try:
            resp = self._get_response()
            self.version = resp.version
        except Empty:
            self.version = 3

    def synchronize(self, len=10):
        command = self._prepare_bytes(commands.TimeSync())
        results = []
        for i in range(len):
            start_t = local_clock()
            self.write(command)
            resp = self._get_response()
            ptp = PTPTimes(t1_tx=start_t,
                           t1_rx=self._blinker_time(resp.rx_time),
                           t2_tx=self._blinker_time(resp.tx_time),
                           t2_rx=resp.first_byte_timestamp
                           )
            results.append(ptp)
        self._blinker_offset = reduce(lambda s, p: s + p.time_offset(), results, 0) / len
        self._duration = reduce(lambda s, p: s + p.t1_rx - self._blinker_offset - p.t1_tx, results, 0) / len

    def _receive(self):
        while not self._closed:
            try:
                msg = Command.read(self)
            except Exception:
                if self._closed:
                    return
                else:
                    raise
            self._receive_queue.put(msg)
            if isinstance(msg, commands.Error):
                LOGGER.error(msg.get_error())

    def _reconfigure_port(self, *args, **kwargs):
        try:
            super(SlimBlinkerDevice, self)._reconfigure_port(*args, **kwargs)
        except Exception:
            pass

    def _prepare_bytes(self, command):
        return to_bytes(command)

    def send(self, command, flush=False):
        if self._preview:
            self._preview.write(command)
        self._queue.append(self._prepare_bytes(command))
        if flush:
            return self.flush()
        return -1

    def _get_response(self):
        return self._receive_queue.get(timeout=self.RECEIVE_TIMEOUT)

    def flush(self):
        self._queue.append(self._prepare_bytes(commands.Flush()))
        self._receive_queue = Queue()
        self.write(b"".join(self._queue))
        self._queue = []
        if self._preview:
            self._preview.flush()
        resp = self._get_response()
        if self.timestamp_method == self.TimestampMethod.TRANSFER_DURATION:
            return self._unix_time_using_transfer_duration(resp.first_byte_timestamp + resp.period * MICROSECONDS)
        else:
            blinker_time = self._blinker_time(resp.last_update)
            return self._unix_time_using_clock_offset(blinker_time + resp.period * MICROSECONDS)

    def close(self):
        self._closed = True
        super(SlimBlinkerDevice, self).close()
        self._receive_thread.join()

    def set_state(self, state):
        self.write(to_bytes(commands.SetState(state)))


class Blinker16Mixin:
    FIELDS_NUMBER = 16
    ROWS = 4
    COLUMNS = 4

    def _define_initial_fields(self):
        self.set_number_of_fields(1)
        border_segments = []
        for i in range(self.ROWS):
            first_segment = int(i * self.SEGMENT_COUNT / self.ROWS)
            border_segments.extend(range(first_segment, first_segment + self.SEGMENTS_PER_FIELD))
            last_segment = int((i + 1) * self.SEGMENT_COUNT / self.ROWS)
            border_segments.extend(range(last_segment - self.SEGMENTS_PER_FIELD, last_segment))
        self.define_field(0, border_segments)
        self.disable(0, 1.0 if hasattr(self, '_deleting') else 0)
        self.set_number_of_fields(self.FIELDS_NUMBER)
        for i in range(self.FIELDS_NUMBER):
            row = int(i / self.COLUMNS)
            column = int(i % self.COLUMNS)
            first_segment = int(
                row * self.SEGMENT_COUNT / 4 + self.SEGMENTS_PER_FIELD + column * self.SEGMENTS_PER_FIELD * 2)
            channel_indexes = list(range(first_segment, first_segment + self.SEGMENTS_PER_FIELD * 2))
            self.define_field(i, channel_indexes)

    def close(self):
        self._deleting = True
        if self._device:
            self.reset()
        super().close()


class SlimBlinkerPreview(Emulator):
    def write(self, data):
        if isinstance(data, (commands.Start, commands.Stop)):
            self.s.blinking = isinstance(data, commands.Start)
        elif isinstance(data, commands.SetFields):
            self.s.set_fields(data.fields_num)
        elif isinstance(data, commands.Intensity):
            self.s.intensity = data.get_intensity()
        elif isinstance(data, (commands.Intensity, commands.TimeSync)):
            pass
        else:
            field = self.s.fields[data.field]
            if isinstance(data, commands.FieldData):
                field.buffer = data.channel_data[:]
                field.buffer = list(
                    np.roll(field.buffer, -int(field.phase / SlimBlinker.MAX_PHASE * len(field.buffer))))
            elif isinstance(data, commands.Frequency):
                field.frequency = data.get_frequency()
                field.phase = data.get_phase()
            elif isinstance(data, commands.Enable):
                field.enabled = True
            elif isinstance(data, commands.Disable):
                field.enabled = False
                field.brightness = data.get_brightness()
            elif isinstance(data, commands.DefineField):
                field.channel_indexes = data.channel_indexes[:]

    def flush(self):
        self._send_state()


class DummySlimBlinkerDevice(object):
    def __init__(self, preview):
        self._preview = preview
        self.version = MAX_FIRMWARE_VERSION

    def send(self, command, flush=False):
        if self._preview:
            self._preview.write(command)
        if flush:
            return self.flush()
        return -1

    def flush(self):
        if self._preview:
            self._preview.flush()
        return time.time()

    def close(self):
        pass

    def set_state(self, state):
        pass

    def get_temperatures(self):
        return [0, 0]


class Segment(namedtuple('_Segment', 'id,x,y,width,height')):
    def overlaps(self, x, y, width, height):
        xb = x + width
        yb = y + height
        if self.xb < x or self.x > xb:
            return False
        if self.yb < y or self.yb > yb:
            return False
        return True

    @property
    def xb(self):
        return self.x + self.width

    @property
    def yb(self):
        return self.y + self.height


def _calculate_segments(screen_width, screen_height, rows, columns):
    segments = []
    segment_width = screen_width / columns
    segment_height = screen_height / rows
    for r in range(rows):
        for c in range(columns):
            segments.append(
                Segment(r * columns + c,
                        c * segment_width,
                        r * segment_height,
                        segment_width,
                        segment_height))
    return segments


class SlimBlinker(AbstractBlinker):
    FIELDS_NUMBER = 40
    SEGMENT_COUNT = 40
    SEGMENTS_PER_FIELD = 1
    ROWS = 4
    COLUMNS = 10
    DEVICE_CLASS = SlimBlinkerDevice
    DUMMY_DEVICE_CLASS = DummySlimBlinkerDevice
    MAX_INTENSITY = 80
    RESOLUTION = (1366, 768)
    SEGMENTS = _calculate_segments(RESOLUTION[0], RESOLUTION[1], ROWS, COLUMNS)

    @classmethod
    def get_segments_for_rect(cls, x, y, width, height, in_pixels=False):
        result = []
        if not in_pixels:
            res_width, res_height = cls.RESOLUTION
            x, y, width, height = x * res_width, y * res_height, width * res_width, height * res_height
        for s in cls.SEGMENTS:
            if s.overlaps(x, y, width, height):
                result.append(s)
        return result

    def _create_preview(self):
        return SlimBlinkerPreview(self.ROWS, self.COLUMNS)

    def _init_device(self, port):
        if port:
            self._device = self.DEVICE_CLASS(port, self._preview)
        else:
            self._device = self.DUMMY_DEVICE_CLASS(self._preview)

    def _update_field(self, field_number, buffer=None, frequency=None, phase_offset=None, luminosity=None):
        config = self._fields[field_number]
        if luminosity is not None:
            config.brightness = luminosity
        if frequency is not None:
            config.frequency = frequency
        if phase_offset is not None:
            config.phase = phase_offset
        if frequency is not None or phase_offset is not None:
            self._device.send(commands.Frequency(field_number, config.frequency, config.phase))
        if buffer is not None:
            config.buffer = numpy.array(buffer) * config.brightness
            self._device.send(commands.FieldData(field_number, config.buffer))

    def start(self, flush=True):
        return self._device.send(commands.Start(), flush=flush)

    def stop(self, flush=True):
        return self._device.send(commands.Stop(), flush=flush)

    def enable(self, field_number, flush=False):
        self._fields[field_number].enabled = True
        return self._device.send(commands.Enable(field_number), flush=flush)

    def disable(self, field_number, brightness, flush=False):
        super(SlimBlinker, self).disable(field_number, brightness)
        return self._device.send(commands.Disable(field_number, brightness), flush=flush)

    def set_still(self, field_number, luminosity):
        return self._update_field(field_number, [1, 1], luminosity=luminosity)

    def _define_initial_fields(self):
        self.set_number_of_fields(self.FIELDS_NUMBER)
        for i in range(self.fields_num):
            channel_indexes = list(range(i * self.SEGMENTS_PER_FIELD, (i + 1) * self.SEGMENTS_PER_FIELD))
            self.define_field(i, channel_indexes)

    def reset(self):
        self._define_initial_fields()
        for i in range(self.fields_num):
            self.disable(i, 1.0, flush=False)
        self.stop()
        return self.flush()

    def flush(self):
        return self._device.flush()

    def define_field(self, field_number, channel_indexes):
        if len(channel_indexes) < 1:
            raise BlinkerException("Field {} - empty channel indexes".format(field_number))
        if field_number >= self.fields_num:
            raise BlinkerException("Can't configure field number {} there is only {} fields".format(field_number,
                                                                                                    self.fields_num))
        if len(channel_indexes) and isinstance(channel_indexes[0], Segment):
            channel_indexes = [s.id for s in channel_indexes]
        assert all(0 <= i < self.SEGMENT_COUNT for i in channel_indexes)
        self._device.send(commands.DefineField(field_number, channel_indexes))

    def set_number_of_fields(self, channels_number):
        self._fields = [FieldConfig(i) for i in range(channels_number)]
        self._device.send(commands.SetFields(channels_number))

    def set_intensity(self, value):
        """Set the overall brightness.

        value - float [0;1]
        """
        self._device.send(commands.Intensity(value, self.MAX_INTENSITY), flush=True)

    def set_state(self, state):
        self._device.set_state(state)

    def close(self):
        super().close()
        if self._device:
            self._device.close()
            self._device = None

    @property
    def version(self):
        return self._device.version

    def set_synchro_field(self, field: int, flush=True):
        self._device.send(Synchro(field), flush)


class Slim16Blinker(Blinker16Mixin, SlimBlinker):
    pass


if __name__ == '__main__':
    port = sys.argv[1] if len(sys.argv) > 1 else None
    blinker = SlimBlinker(port, port is None)
    run_demo(blinker, 5)
    del blinker
