from unittest import mock

import numpy as np
import pytest

from braintech.obci.signal_processing import buffers
from braintech.obci.signal_processing.signal.data_generic_write_proxy import (
    Impedance, SamplePacket
)


def test_ring_buffer_add_and_get():
    ring_buffer = buffers.RingSampleBuffer(size=2, number_of_channels=2, copy_on_ret=True)

    last = np.array([1, 1])
    ring_buffer.add(last)
    new = np.array([2, 2])
    ring_buffer.add(new)
    expected = np.array([[1, 2],
                         [1, 2]])
    assert (ring_buffer.get(0, 2) == expected).all()

    new = np.array([3, 3])
    ring_buffer.add(new)
    expected = np.array([[2, 3],
                         [2, 3]])
    assert (ring_buffer.get(0, 2) == expected).all()

    new = np.array([4, 4])
    ring_buffer.add(new)
    expected = np.array([[3, 4],
                         [3, 4]])
    assert (ring_buffer.get(0, 2) == expected).all()


def test_ring_buffer_get():
    ring_buffer = buffers.RingSampleBuffer(size=3, number_of_channels=2, copy_on_ret=True)

    def assertArrayEqual(actual, expected):
        assert np.array_equal(actual, expected)

    assert len(ring_buffer.get(0, 0)) == 0
    assertArrayEqual(ring_buffer.get(0, 1), np.array([[0], [0]]))
    assertArrayEqual(ring_buffer.get(0, 2), np.array([[0, 0],
                                                      [0, 0]]))
    assertArrayEqual(ring_buffer.get(0, 3), np.array([[0, 0, 0],
                                                      [0, 0, 0]]))

    ring_buffer.add(np.array([1, 1]))

    assert len(ring_buffer.get(0, 0)) == 0
    assertArrayEqual(ring_buffer.get(0, 1), np.array([[1],
                                                      [1]]))
    assertArrayEqual(ring_buffer.get(0, 2), np.array([[1, 0],
                                                      [1, 0]]))
    assertArrayEqual(ring_buffer.get(0, 3), np.array([[1, 0, 0],
                                                      [1, 0, 0]]))
    ring_buffer.add(np.array([2, 2]))

    assert len(ring_buffer.get(0, 0)) == 0
    assertArrayEqual(ring_buffer.get(0, 1), np.array([[1],
                                                      [1]]))
    assertArrayEqual(ring_buffer.get(0, 2), np.array([[1, 2],
                                                      [1, 2]]))
    assertArrayEqual(ring_buffer.get(0, 3), np.array([[1, 2, 0],
                                                      [1, 2, 0]]))
    assert len(ring_buffer.get(0, 0)) == 0
    assertArrayEqual(ring_buffer.get(1, 1), np.array([[2],
                                                      [2]]))
    assertArrayEqual(ring_buffer.get(1, 2), np.array([[2, 0],
                                                      [2, 0]]))
    assertArrayEqual(ring_buffer.get(1, 3), np.array([[2, 0, 1],
                                                      [2, 0, 1]]))
    assert len(ring_buffer.get(0, 0)) == 0
    assertArrayEqual(ring_buffer.get(2, 1), np.array([[0],
                                                      [0]]))
    assertArrayEqual(ring_buffer.get(2, 2), np.array([[0, 1],
                                                      [0, 1]]))
    assertArrayEqual(ring_buffer.get(2, 3), np.array([[0, 1, 2],
                                                      [0, 1, 2]]))

    assert len(ring_buffer.get(0, 0)) == 0
    assertArrayEqual(ring_buffer.get(3, 1), np.array([[1],
                                                      [1]]))
    assertArrayEqual(ring_buffer.get(3, 2), np.array([[1, 2],
                                                      [1, 2]]))
    assertArrayEqual(ring_buffer.get(3, 3), np.array([[1, 2, 0],
                                                      [1, 2, 0]]))

    with pytest.raises(AssertionError):
        ring_buffer.get(-1, 1)
    with pytest.raises(AssertionError):
        ring_buffer.get(1, -1)
    with pytest.raises(AssertionError):
        ring_buffer.get(0, 100)

    ring_buffer.add(np.array([3, 3]))
    assertArrayEqual(ring_buffer[:2], ring_buffer.get(0, 2))
    assertArrayEqual(ring_buffer[1:2], ring_buffer.get(1, 1))
    assertArrayEqual(ring_buffer[-2:], ring_buffer.get(1, 2))
    assertArrayEqual(ring_buffer[-3:-1], ring_buffer.get(0, 2))
    assertArrayEqual(ring_buffer[1], ring_buffer.get(0, 3)[:, 1])
    assertArrayEqual(ring_buffer[-1], ring_buffer.get(0, 3)[:, 2])


@pytest.mark.incremental
class TestAutoBufferCreationWithImpedanceFlag:
    def test_buffer_should_call_callback_with_additional_argument(self, sample_packet):
        kwargs = self.creation_kwargs(sample_packet)
        kwargs['return_impedance'] = True
        kwargs['ret_func'] = ret_func = mock.MagicMock()
        instance = buffers.AutoRingBuffer(**kwargs)
        instance.handle_sample_packet(sample_packet)
        call_args_with_impedance, call_kwargs_with_impedance = ret_func.call_args
        assert len(call_args_with_impedance) == 2
        assert isinstance(call_args_with_impedance[0], np.ndarray)
        assert isinstance(call_args_with_impedance[1], Impedance)

        kwargs['return_impedance'] = False
        instance = buffers.AutoRingBuffer(**kwargs)
        instance.handle_sample_packet(sample_packet)
        call_args_without_impedance, call_kwargs_without_impedance = ret_func.call_args
        assert len(call_args_without_impedance) == 1
        assert isinstance(call_args_without_impedance[0], np.ndarray)

    def test_buffer_last_argument_should_be_a_valid_impedance_value(self, sample_packet):
        kwargs = self.creation_kwargs(sample_packet)
        kwargs['return_impedance'] = True
        kwargs['ret_func'] = ret_func = mock.MagicMock()
        instance = buffers.AutoRingBuffer(**kwargs)
        instance.handle_sample_packet(sample_packet)
        call_args_with_impedance, call_kwargs_with_impedance = ret_func.call_args

        impedance = call_args_with_impedance[-1]
        assert isinstance(impedance, Impedance)
        for channel_idx in range(sample_packet.channel_count):
            expected = sample_packet.impedance.for_channel(channel_idx)
            if isinstance(expected, np.ndarray):
                computed = impedance.for_channel(channel_idx)
                assert isinstance(computed, np.ndarray)
                assert len(computed) == kwargs['samples_count']
                assert (computed == expected).all()
            else:
                expected = None
                computed = impedance.for_channel(channel_idx)
                assert computed is expected

    def creation_kwargs(self, sample_packet):
        samples_count, channels_count = sample_packet.samples.shape
        buffer_size = samples_count
        number_of_samples_to_skip_in_each_call_to_get = buffer_size
        kwargs = dict(
            samples_count=samples_count,
            num_of_channels=channels_count,
            copy_on_ret=True,
            ret_func=mock.MagicMock(),
            from_sample=buffer_size,
            every=number_of_samples_to_skip_in_each_call_to_get,
            return_impedance=False,
        )
        return kwargs


@pytest.mark.incremental
class TestAutoBlinkBufferCreationWithImpedanceFlag:
    def test_buffer_should_call_callback_with_additional_argument(self, sample_packet):
        kwargs = self.creation_kwargs(sample_packet)
        kwargs['return_impedance'] = True
        kwargs['ret_func'] = ret_func = mock.MagicMock()
        instance = buffers.AutoBlinkBuffer(**kwargs)
        self.make_callback_be_called(instance, sample_packet)
        call_args_with_impedance, call_kwargs_with_impedance = ret_func.call_args
        assert len(call_args_with_impedance) == 3
        assert isinstance(call_args_with_impedance[0], buffers.Blink)
        assert isinstance(call_args_with_impedance[1], np.ndarray)
        assert isinstance(call_args_with_impedance[2], Impedance)

        kwargs['return_impedance'] = False
        instance = buffers.AutoBlinkBuffer(**kwargs)
        self.make_callback_be_called(instance, sample_packet)
        call_args_without_impedance, call_kwargs_without_impedance = ret_func.call_args
        assert len(call_args_without_impedance) == 2
        assert isinstance(call_args_without_impedance[0], buffers.Blink)
        assert isinstance(call_args_without_impedance[1], np.ndarray)

    def test_buffer_last_argument_should_be_a_valid_impedance_value(self, sample_packet):
        kwargs = self.creation_kwargs(sample_packet)
        kwargs['return_impedance'] = True
        kwargs['ret_func'] = ret_func = mock.MagicMock()
        instance = buffers.AutoBlinkBuffer(**kwargs)
        self.make_callback_be_called(instance, sample_packet)
        call_args_with_impedance, call_kwargs_with_impedance = ret_func.call_args

        impedance = call_args_with_impedance[-1]
        assert isinstance(call_args_with_impedance[-1], Impedance)
        for channel_idx in range(sample_packet.channel_count):
            expected = sample_packet.impedance.for_channel(channel_idx)
            if isinstance(expected, np.ndarray):
                computed = impedance.for_channel(channel_idx)
                assert isinstance(computed, np.ndarray)
                assert len(computed) == kwargs['samples_count']
                assert (computed == expected).all()
            else:
                expected = None
                computed = impedance.for_channel(channel_idx)
                assert computed is expected

    def creation_kwargs(self, sample_packet):
        samples_count, channels_count = sample_packet.samples.shape
        sampling_frequency = 1
        kwargs = dict(
            samples_count=samples_count,
            num_of_channels=channels_count,
            copy_on_ret=True,
            from_blink=0,
            sampling=sampling_frequency,
            ret_func=mock.MagicMock(),
            return_impedance=False,
        )
        return kwargs

    def make_callback_be_called(self, instance, sample_packet):
        instance.handle_sample_packet(sample_packet)
        self._fill_buffer_with_the_same_packet(instance, sample_packet)
        instance.handle_blink(buffers.Blink(timestamp=0, index=0))
        instance.handle_sample_packet(sample_packet)

    def _fill_buffer_with_the_same_packet(self, instance, sample_packet):
        samples_handled = 0
        while not instance._is_full:
            instance.handle_sample_packet(sample_packet)
            samples_handled += 1
        return samples_handled


@pytest.fixture(
    params=[
        SamplePacket(
            samples=np.array([[0.0],
                              [0.5],
                              [1.0],
                              [0.5],
                              [0.0]]),
            ts=np.array([0.0,
                         1.0,
                         2.0,
                         3.0,
                         4.0]),
            impedance=Impedance(
                ids=np.array([Impedance.UNKNOWN]),
                data=np.array([[]])
            )
        ),
        SamplePacket(
            samples=np.array([[0.0, 1.0],
                              [0.5, 1.75],
                              [1.0, 2.5],
                              [0.5, 1.75],
                              [0.0, 1.0]]),
            ts=np.array([0.0,
                         1.0,
                         2.0,
                         3.0,
                         4.0]),
            impedance=Impedance(
                ids=np.array([Impedance.PRESENT, Impedance.UNKNOWN]),
                data=np.array([[0.5],
                               [0.75],
                               [1.0],
                               [0.75],
                               [0.5]])
            )
        ),
        SamplePacket(
            samples=np.array([[0.0, 0.5, 1.0],
                              [0.0, 0.75, 1.5],
                              [0.0, 1.0, 2.0]]),
            ts=np.array([0.0,
                         1.0,
                         2.0]),
            impedance=Impedance(
                ids=np.array([Impedance.PRESENT, Impedance.UNKNOWN, Impedance.NOT_APPLICABLE]),  # noqa
                data=np.array([[0.5],
                               [0.75],
                               [0.5]])
            )
        ),
        SamplePacket(
            samples=np.array([[0.0, 0.5, 1.0, 0.5, 0.0],
                              [0.0, 0.75, 1.5, 0.75, 0.0]]),
            ts=np.array([0.0,
                         1.0]),
            impedance=Impedance(
                ids=np.array([Impedance.PRESENT, Impedance.UNKNOWN, Impedance.NOT_APPLICABLE, Impedance.PRESENT,
                              Impedance.UNKNOWN]),  # noqa
                data=np.array([[0.5, 1.0],
                               [0.75, 2.0]])
            )
        ),
    ],
)
def sample_packet(request):
    return request.param
