import os.path
from collections import namedtuple

import numpy as np
import pytest

from braintech.obci.signal_processing import writers as signal_writers
from braintech.obci.signal_processing.read_manager import ReadManager
from braintech.obci.signal_processing.signal import containers as signal_containers
from braintech.obci.signal_processing.signal.containers import ArrayWithImpedance
from braintech.obci.signal_processing.signal.data_generic_write_proxy import (
    Impedance, SamplePacket
)
from braintech.obci.signal_processing.signal.data_read_proxy import DataReadProxy
from braintech.obci.signal_processing.signal.data_write_proxy import get_proxy
from braintech.obci.signal_processing.signal.info_file_proxy import InfoFileReadProxy, InfoFileWriteProxy

START_TIMESTAMP = 1536753278.7339857


def test_saving_and_retrieving_data(data_writer, data_reader, data_kwargs, sample_packet):
    all_channels_count = sample_packet.channel_count
    if data_kwargs['append_ts']:
        all_channels_count += 1
    if data_kwargs['save_impedance']:
        all_channels_count += sample_packet.impedance.impedance_channel_count

    data_writer.set_first_sample_timestamp(START_TIMESTAMP)
    data_writer.data_received(sample_packet)
    data_writer.finish_saving()
    data_reader.start_reading()
    saved_data = data_reader.get_all_values(all_channels_count)
    data_reader.finish_reading()
    saved_data = saved_data.T
    active_channels = sample_packet.channel_count
    assert np.array_equal(saved_data[:, :active_channels], sample_packet.samples)
    if data_kwargs['append_ts']:
        impedance_data_start = active_channels + 1
        assert np.array_equal(saved_data[:, active_channels] + START_TIMESTAMP, sample_packet.ts)
    else:
        impedance_data_start = active_channels
    if data_kwargs['save_impedance']:
        assert np.array_equal(saved_data[:, impedance_data_start:], sample_packet.impedance.data)
    else:
        assert impedance_data_start == saved_data.shape[1]


def test_saving_and_retrieving_info(signal_writer, sample_packet,
                                    files_paths, sample_type, data_kwargs, metadata):
    expected = {
        'blocks_per_page': '5',
        'byte_order': 'LITTLE_ENDIAN',
        'calibration': '1.0',
        'channels_gains': list(map(str, metadata['channel_gains'])),
        'channels_names': list(metadata['channel_names']),
        'channels_offsets': list(map(str, metadata['channel_offsets'])),
        'export_file_name': 'name',
        'file': files_paths.data_file_path,
        'file_format': [''],
        'first_sample_timestamp': repr(sample_packet.ts[0]),
        'number_of_channels': str(sample_packet.channel_count),
        'number_of_samples': str(sample_packet.samples.size),
        'page_size': '20.0',
        'sample_type': sample_type,
        'sampling_frequency': str(metadata['sampling_rate']),
    }
    if data_kwargs['append_ts']:
        expected['channels_names'].append('TSS')
        expected['channels_gains'].append('1.0')
        expected['channels_offsets'].append('0.0')
        expected['number_of_channels'] = str(int(expected['number_of_channels']) + 1)

    if data_kwargs['save_impedance']:
        impedance_channels_names = [
            'Imp_{name}'.format(name=name)
            for name, flag in zip(expected['channels_names'], sample_packet.impedance.flags)
            if flag == Impedance.PRESENT
        ]
        expected['channels_names'] += impedance_channels_names
        expected['channels_gains'] += ['1'] * len(impedance_channels_names)
        expected['channels_offsets'] += ['0'] * len(impedance_channels_names)
        channels_flags = list(map(str, sample_packet.impedance.flags))
        if data_kwargs['append_ts']:
            channels_flags += [str(Impedance.NOT_APPLICABLE)]
        expected['channels_flags'] = channels_flags
        expected['number_of_channels'] = str(int(expected['number_of_channels']) + len(impedance_channels_names))

    _save_packet_to_file(signal_writer, sample_packet)
    info_reader = InfoFileReadProxy(files_paths.info_file_path)
    assert expected == info_reader.get_params()
    data_reader = DataReadProxy(
        p_file_path=files_paths.data_file_path,
        sample_type=sample_type,
    )
    if data_kwargs['append_ts']:
        num_of_channels = len(info_reader.get_param("channels_names"))
        ts_chnl_nr = info_reader.get_param("channels_names").index("TSS")
        TSS_samples = data_reader.get_all_values(num_of_channels)[ts_chnl_nr, :]
        assert TSS_samples[0] + START_TIMESTAMP == float(info_reader.get_param('first_sample_timestamp'))
    data_reader.finish_reading()


def test_read_manager_get_all_samples(signal_writer, sample_packet,
                                      files_paths, data_kwargs, metadata):
    _save_packet_to_file(signal_writer, sample_packet)
    signal_reader = ReadManager(
        p_info_source=files_paths.info_file_path,
        p_data_source=files_paths.data_file_path,
        p_tags_source=files_paths.tags_file_path,
    )
    samples = signal_reader.get_all_samples()
    if data_kwargs['append_ts']:
        expected_samples = np.hstack((
            sample_packet.samples,
            sample_packet.ts.reshape((sample_packet.sample_count, 1)) - START_TIMESTAMP,
        ))
    else:
        expected_samples = sample_packet.samples
    expected_samples = expected_samples.T
    expected_impedance = sample_packet.impedance.data.T
    assert np.array_equal(samples, expected_samples)
    all_samples = list(signal_reader.iter_samples())
    assert np.array_equal(np.array(all_samples).T, samples)

    if data_kwargs['save_impedance']:
        assert isinstance(all_samples[0], ArrayWithImpedance)
        if data_kwargs['append_ts']:
            expected_flags = np.append(sample_packet.impedance.flags, Impedance.NOT_APPLICABLE)
            assert np.array_equal(expected_flags, samples.impedance_flags)
            assert len(samples.impedance.shape) == len(sample_packet.samples.shape)
            assert samples.impedance.shape[0] == expected_samples.shape[0]
            assert samples.impedance.shape[1] == expected_samples.shape[1]
            assert float(signal_reader.get_param("first_sample_timestamp")) > 0
        else:
            expected_flags = sample_packet.impedance.flags
            assert np.array_equal(expected_flags, samples.impedance_flags)
            assert samples.impedance.shape == expected_samples.shape
        impedance_data_without_nans = samples.impedance[np.logical_not(np.isnan(samples.impedance[:, 0]))]
        assert np.array_equal(expected_impedance, impedance_data_without_nans)
    else:
        with pytest.raises(AttributeError):
            samples.impedance_flags
        with pytest.raises(AttributeError):
            samples.impedance


def test_read_manager_get_channel_samples(signal_writer, sample_packet,
                                          files_paths, data_kwargs, metadata):
    _save_packet_to_file(signal_writer, sample_packet)
    signal_reader = ReadManager(
        p_info_source=files_paths.info_file_path,
        p_data_source=files_paths.data_file_path,
        p_tags_source=files_paths.tags_file_path,
    )
    channels_names = list(signal_reader.get_param('channels_names'))
    sample_channels_names = [
        channel_name
        for channel_name in channels_names
        if not channel_name.startswith('Imp_')
    ]
    all_samples = signal_reader.get_all_samples()

    for idx, channel_name in enumerate(sample_channels_names):
        samples = signal_reader.get_channel_samples(channel_name)
        expected = all_samples[idx, :]
        assert np.array_equal(expected, samples)
        if data_kwargs['save_impedance']:
            assert all_samples.impedance_flags[idx] == samples.impedance_flags[0]
            if samples.impedance_flags[0] == signal_containers.Impedance.PRESENT:
                expected = all_samples.impedance[idx:idx + 1, :]
                computed = samples.impedance
                assert np.array_equal(expected, computed)
            else:
                assert (np.isnan(samples.impedance)).all()
        else:
            with pytest.raises(AttributeError):
                samples.impedance_flags
            with pytest.raises(AttributeError):
                samples.impedance
    if sample_packet is GET_CHANNELS_SAMPLE_PACKET and data_kwargs['save_impedance']:
        _test_read_manager_get_channels_samples(sample_packet, signal_reader)


def _test_read_manager_get_channels_samples(sample_packet, signal_reader):
    samples = signal_reader.get_channels_samples(['EXG_4', 'EXG_3', 'EXG_1'], 0, 1)
    assert samples.shape == (3, 1)
    assert samples.impedance.shape == (3, 1)
    assert samples.impedance[0, 0] == sample_packet.impedance.for_channel(3)[0]
    assert np.isnan(samples.impedance[1, 0])
    assert samples.impedance[2, 0] == sample_packet.impedance.for_channel(0)[0]


def _save_packet_to_file(signal_writer, packet):
    signal_writer.write(packet, packet.samples.size)
    signal_writer.finish_saving_signal()


@pytest.fixture()
def metadata(sample_packet):
    active_channels = [
        'EXG_{}'.format(idx)
        for idx in range(1, sample_packet.channel_count + 1)
    ]
    return dict(
        active_channels=active_channels,
        sampling_rate=128,
        channel_names=list(active_channels),
        channel_gains=[idx for idx in range(1, sample_packet.channel_count + 1)],
        channel_offsets=[idx for idx in range(1, sample_packet.channel_count + 1)],
    )


@pytest.fixture()
def signal_writer(files_paths, data_kwargs, sample_type, sample_packet, metadata):
    file_path = os.path.dirname(files_paths.data_file_path)
    file_name = os.path.basename(files_paths.data_file_path).rstrip(signal_writers.SignalWriter.FILE_EXTENSIONS['data'])
    writer = signal_writers.SignalWriter(
        save_file_path=file_path,
        save_file_name=file_name,
        append_timestamps=data_kwargs['append_ts'],
        use_own_buffer=data_kwargs['use_own_buffer'],
        sample_type=(sample_type if data_kwargs['format'] == 'non-ascii' else 'ascii'),
        save_tags=True,
        channels_info=[],
        save_impedance=data_kwargs['save_impedance'],
        **metadata
    )
    writer.first_sample_timestamp = sample_packet.ts[0]
    if data_kwargs['save_impedance']:
        writer.impedance_flags = sample_packet.impedance.flags
    return writer


@pytest.fixture()
def data_writer(sample_type, files_paths, data_kwargs):
    return get_proxy(
        file_path=files_paths.data_file_path,
        append_ts=data_kwargs['append_ts'],
        use_own_buffer=data_kwargs['use_own_buffer'],
        format=(sample_type if data_kwargs['format'] == 'non-ascii' else 'ascii'),
        save_impedance=data_kwargs['save_impedance'],
    )


@pytest.fixture()
def data_reader(sample_type, files_paths):
    data_reader = DataReadProxy(
        p_file_path=files_paths.data_file_path,
        sample_type=sample_type,
    )
    yield data_reader
    data_reader.finish_reading()


@pytest.fixture(
    params=[
        dict(append_ts=False, use_own_buffer=False, format='non-ascii', save_impedance=False),
        dict(append_ts=False, use_own_buffer=True, format='non-ascii', save_impedance=False),
        dict(append_ts=True, use_own_buffer=False, format='non-ascii', save_impedance=False),
        dict(append_ts=True, use_own_buffer=True, format='non-ascii', save_impedance=False),
        dict(append_ts=False, use_own_buffer=False, format='non-ascii', save_impedance=True),
        dict(append_ts=False, use_own_buffer=True, format='non-ascii', save_impedance=True),
        dict(append_ts=True, use_own_buffer=False, format='non-ascii', save_impedance=True),
        dict(append_ts=True, use_own_buffer=True, format='non-ascii', save_impedance=True),
    ],
    ids=lambda kwargs: 'appaned_ts:{}-use_own_buffer:{}-format:{}-save_impedance:{}'.format(
        kwargs['append_ts'], kwargs['use_own_buffer'], kwargs['format'], kwargs['save_impedance'])
)
def data_kwargs(request):
    return request.param


@pytest.fixture(
    params=['DOUBLE', 'FLOAT'],
    ids=lambda arg: 'sample_type:{}'.format(arg)
)
def sample_type(request):
    return request.param


@pytest.fixture()
def files_paths(tmpdir):
    FilePaths = namedtuple(
        'FilePaths', ['data_file_path', 'info_file_path', 'tags_file_path']
    )
    saved_files_dir = tmpdir.mkdir('test_saving_and_retrieving')
    data_file = saved_files_dir.join('test.obci.raw')
    data_file.ensure(file=True)
    data = str(data_file)
    info_file = saved_files_dir.join('test.obci.xml')
    info_file.ensure(file=True)
    info = str(info_file)
    tags_file = saved_files_dir.join('test.obci.tag')
    tags_file.ensure(file=True)
    tags = str(tags_file)
    return FilePaths(data_file_path=data, info_file_path=info, tags_file_path=tags)


GET_CHANNELS_SAMPLE_PACKET = SamplePacket(
    samples=np.array([[0.0, 0.5, 1.0, 0.5, 0.0],
                      [0.0, 0.75, 1.5, 0.75, 0.0]]),
    ts=np.array([0.0,
                 1.0], dtype=np.float64) + START_TIMESTAMP,
    impedance=Impedance(
        ids=np.array(
            [Impedance.PRESENT, Impedance.UNKNOWN, Impedance.NOT_APPLICABLE, Impedance.PRESENT, Impedance.UNKNOWN]),
        # noqa
        data=np.array([[0.5, 1.0],
                       [0.75, 2.0]])
    )
)


@pytest.fixture(
    params=[
        SamplePacket(
            samples=np.array([[0.0],
                              [0.5],
                              [1.0],
                              [0.5],
                              [0.0]]),
            ts=np.array([0.0,
                         1.0,
                         2.0,
                         3.0,
                         4.0], dtype=np.float64) + START_TIMESTAMP,
            impedance=Impedance.create_unknown(1, 5)
        ),
        SamplePacket(
            samples=np.array([[0.0, 1.0],
                              [0.5, 1.75],
                              [1.0, 2.5],
                              [0.5, 1.75],
                              [0.0, 1.0]]),
            ts=np.array([0.0,
                         1.0,
                         2.0,
                         3.0,
                         4.0], dtype=np.float64) + START_TIMESTAMP,
            impedance=Impedance(
                ids=np.array([Impedance.PRESENT, Impedance.UNKNOWN]),
                data=np.array([[0.5],
                               [0.75],
                               [1.0],
                               [0.75],
                               [0.5]], dtype=np.float64) + START_TIMESTAMP
            )
        ),
        SamplePacket(
            samples=np.array([[0.0, 0.5, 1.0],
                              [0.0, 0.75, 1.5],
                              [0.0, 1.0, 2.0]]),
            ts=np.array([0.0,
                         1.0,
                         2.0], dtype=np.float64) + START_TIMESTAMP,
            impedance=Impedance(
                ids=np.array([Impedance.PRESENT, Impedance.UNKNOWN, Impedance.NOT_APPLICABLE]),  # noqa
                data=np.array([[0.5],
                               [0.75],
                               [0.5]])
            )
        ),
        GET_CHANNELS_SAMPLE_PACKET
    ],
)
def sample_packet(request):
    return request.param


# to ensure backups work
def test_info_file_writer_attributes_update(files_paths):
    writer = InfoFileWriteProxy(files_paths.info_file_path)
    writer.set_attributes({"number_of_samples": 1})
    writer.save_info()

    reader = InfoFileReadProxy(files_paths.info_file_path)
    assert reader.get_param("number_of_samples") == '1'

    writer.set_attributes({"number_of_samples": 3})
    writer.save_info()

    reader = InfoFileReadProxy(files_paths.info_file_path)
    assert reader.get_param("number_of_samples") == '3'

    writer.set_attributes({"number_of_samples": 10}, overwrite=False)
    writer.save_info()

    reader = InfoFileReadProxy(files_paths.info_file_path)
    assert reader.get_param("number_of_samples") == '3', "overwriting False broke"
