# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.

import collections

from braintech.obci.core.broker.message_handler_mixin import subscribe_message_handler
from . import analysis
from . import messages as camera_messages

StateChangePoint = collections.namedtuple('StateChangePoint', ['state', 'time'])


class CameraAggregatorMixin:
    def __init__(self, *args, **kwargs):
        self.camera_aggregator = None
        super().__init__(*args, **kwargs)

    @subscribe_message_handler(camera_messages.CameraEyesStateMessage)
    def handle_eyes_state_change(self, msg: camera_messages.CameraEyesStateMessage):
        if self.camera_aggregator:
            state = analysis.EyesState(msg.state)
            self.camera_aggregator.add_eyes_state(state=state, unix_time=msg.timestamp)

    @subscribe_message_handler(camera_messages.CameraHeadStateMessage)
    def handle_head_state_change(self, msg: camera_messages.CameraHeadStateMessage):
        if self.camera_aggregator:
            state = analysis.HeadState(msg.state)
            self.camera_aggregator.add_head_state(state=state, unix_time=msg.timestamp)


class StateAggregator:
    DEFAULT_EYES_STATE = analysis.EyesState.NONE_PRESENT
    DEFAULT_HEAD_STATE = analysis.HeadState.NONE_PRESENT
    DEFAULT_STATE_TIME = 0
    CONTROL_STATE_MINIMAL_TIME_PERCENTAGE = 0.8

    def __init__(self):
        self._eyes_states = collections.deque()
        self._head_states = collections.deque()

    def add_eyes_state(self, state: analysis.EyesState, unix_time):
        state_change_point = StateChangePoint(state=state, time=unix_time)
        target_index = self._find_first_later_index(unix_time, self._eyes_states)
        self._eyes_states.insert(target_index, state_change_point)

    def add_head_state(self, state: analysis.HeadState, unix_time):
        state_change_point = StateChangePoint(state=state, time=unix_time)
        target_index = self._find_first_later_index(unix_time, self._head_states)
        self._head_states.insert(target_index, state_change_point)

    def _find_first_later_index(self, unix_time, current_points):
        for i in range(len(current_points)):
            state_change = current_points[i]
            is_later = state_change.time > unix_time
            if is_later:
                first_later_index = i
                return first_later_index
        else:
            index_after_last = len(current_points)
            return index_after_last

    def was_control_state(self, start_unix_time, end_unix_time):
        artificial_start_point = StateChangePoint(state=self.DEFAULT_EYES_STATE,
                                                  time=self.DEFAULT_STATE_TIME)
        relevant_points = self._get_relevant_points(start_unix_time, end_unix_time,
                                                    self._eyes_states, artificial_start_point)
        percentage = self._count_control_state_percentage(start_unix_time, end_unix_time,
                                                          relevant_points)
        return self.CONTROL_STATE_MINIMAL_TIME_PERCENTAGE <= percentage

    def had_artifact(self, start_unix_time, end_unix_time):
        artificial_start_point = StateChangePoint(state=self.DEFAULT_HEAD_STATE,
                                                  time=self.DEFAULT_STATE_TIME)
        relevant_points = self._get_relevant_points(start_unix_time, end_unix_time,
                                                    self._head_states, artificial_start_point)
        artifact_states = [analysis.HeadState.MOVING]
        return any(p.state in artifact_states for p in relevant_points)

    def _get_relevant_points(self, start_unix_time, end_unix_time, points,
                             artificial_start_point):
        relevant = []
        for i in range(len(points)):
            index_from_behind = -(i + 1)
            point = points[index_from_behind]
            if end_unix_time < point.time:
                continue
            elif start_unix_time < point.time <= end_unix_time:
                relevant.append(point)
            else:
                relevant.append(point)
                break
        else:
            relevant.append(artificial_start_point)
        relevant.reverse()
        return relevant

    def _count_control_state_percentage(self, start_unix_time, end_unix_time,
                                        points):
        eye_state_to_control_state = {
            analysis.EyesState.OPEN: True,
            analysis.EyesState.CLOSED: False,
            analysis.EyesState.NONE_PRESENT: False,
        }
        control_state_time = 0
        total_involved_time = 0
        artificial_last_end = StateChangePoint(state=None, time=9999999999)
        starts = points
        ends = points[1:] + [artificial_last_end]
        for start, end in zip(starts, ends):
            is_control_state = eye_state_to_control_state[start.state]
            time_involved = self._count_time_intersection(start.time, end.time,
                                                          start_unix_time, end_unix_time)
            if is_control_state:
                control_state_time += time_involved
            total_involved_time += time_involved
        return control_state_time / total_involved_time

    def _count_time_intersection(self, start1, end1, start2, end2):
        later_start = max(start1, start2)
        earlier_end = min(end1, end2)
        return max(earlier_end - later_start, 0)
