# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.

import datetime
import time

from braintech.obci.core.broker import messages as obci_messages
from braintech.obci.experiment.peer.configured_peer import ConfiguredPeer
from braintech.obci.core.utils.properties import param_property

from . import analysis
from . import messages

__all__ = ('CameraAnalyzingPeer',)

FAKE_FRAME_DURATION = datetime.timedelta(seconds=1 / 30)


class CameraAnalyzingPeer(ConfiguredPeer):
    angle_difference_for_movement = param_property('angle_difference_for_movement',
                                                   deserializer=float, serializer=str)
    eyes_open_borderline_ratio = param_property('eyes_open_borderline_ratio',
                                                deserializer=float, serializer=str)
    display_preview_window = param_property('display_preview_window',
                                            deserializer=int, serializer=str)

    def run(self):
        while not self.is_shutting_down:
            if self.is_running:
                self._broadcast_state_changes()
            else:
                time.sleep(0.1)
        super().run()

    def _broadcast_state_changes(self):
        last_state = None
        states_generator = analysis.get_current_state_from_camera(
            self.angle_difference_for_movement,
            self.display_preview_window,
            self.eyes_open_borderline_ratio,
        )
        while self.is_running:
            state = next(states_generator)
            frame_time = datetime.datetime.now()
            self._send_state_change(state, last_state, frame_time)
            last_state = state

    def _send_state_change(self, new, old, timestamp):
        if self._is_eyes_state_new(new, old):
            self._logger.info('New state: %s', new.eyes)
            message = messages.CameraEyesStateMessage(
                timestamp=timestamp.timestamp(),
                state=new.eyes.value,
            )
            self.send_message(message)
            self._send_state_change_tag_message(new.eyes, timestamp)
        if self._is_head_state_new(new, old):
            self._logger.info('New state: %s', new.head)
            message = messages.CameraHeadStateMessage(
                timestamp=timestamp.timestamp(),
                state=new.head.value,
            )
            self.send_message(message)
            self._send_state_change_tag_message(new.head, timestamp)

    def _is_eyes_state_new(self, new, old):
        return old is None or new.eyes != old.eyes

    def _is_head_state_new(self, new, old):
        return old is None or new.head != old.head

    def _send_state_change_tag_message(self, new_state, frame_time):
        fake_end_time = frame_time + FAKE_FRAME_DURATION
        no_channels_flag = '-1'
        message = obci_messages.TagMsg(
            start_timestamp=frame_time.timestamp(),
            end_timestamp=fake_end_time.timestamp(),
            name=new_state.value,
            channels=no_channels_flag,
        )
        self.send_message(message)
