# -*- coding: utf-8 -*-
# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.
from collections import namedtuple

import numpy as np
from scipy import signal

from braintech.obci.signal_processing.buffers import RingSampleBuffer
from obci_p300.utils import ArrayCascadeFilter

Thresholds = namedtuple('thresholds', 'abs,std,med')


def bad_segment2bad_samples(bad_segments, width, size, shift=0):
    bad_samples = np.repeat(bad_segments, width)
    if shift:
        bad_samples = np.roll(bad_samples, shift)
        bad_samples[:shift] = 0
    return bad_samples[:size]


def minimax(a, w):
    chunked_size = int(len(a) / w) * w
    a_chunks = a[:chunked_size].reshape(-1, w)
    mini_max = a_chunks.max(axis=1) - a_chunks.min(axis=1)
    if chunked_size != len(a):
        mini_max = np.concatenate((mini_max, [a[chunked_size:].max() - a[chunked_size:].min()]))
    return mini_max


def power(s, w, rng, Fs):
    freq = np.fft.rfftfreq(w, 1. / Fs)
    a = np.argmin(np.abs(rng[0] - freq))
    b = np.argmin(np.abs(rng[1] - freq))
    assert rng[0] < rng[1], "Wrong range {}".format(rng)
    if a == b:
        print("UWAGA! W podanym zakresie {} znalazła się tylko jedna częstość: {}".format(rng, freq[a]))
    if b < a:
        print("UWAGA! W podanym zakresie {} zostały niepoprawnie wybrane indeksy a b: {} {}".format(rng, a, b))

    window = signal.hanning(w)
    chunks = s[:(s.size // w) * w].reshape(-1, w)
    f = np.fft.rfft(window * chunks, norm="ortho")
    ps = np.mean(np.abs(f[:, a:b + 1]) ** 2, axis=1)
    return ps, (freq[a], freq[b])


class ArtifactDetector:
    ARTIFACT_NAME = 'artifact'
    MIN_DURATION = 0.03
    DEFAULT_FILTERS = [[1., 0.5, 3, 6.97, "butter"],
                       [30, 60, 3, 12.4, "butter"],
                       [[47.5, 52.5], [49.9, 50.1], 3, 25, "cheby2"]]

    def __init__(self, thresholds, channels, sampling_rate, merge_window, forget=0, max_data_duration=60):
        self._thresholds = thresholds
        self._training_data = np.empty((channels, 0), dtype=np.float64)
        self._channels = channels
        self._forget = forget
        self._sampling_rate = sampling_rate
        self._merge_window = int(merge_window * sampling_rate)
        self._filters = ArrayCascadeFilter()
        self.set_filters(self.DEFAULT_FILTERS)
        self._current_thresholds = np.array([thresholds.abs] * self._channels)
        self._channel_thresholds = [('abs', thresholds)] * self._channels
        self._filtered_data = RingSampleBuffer(int(max_data_duration * self._sampling_rate), channels, False)

    def add_data(self, samples):
        filtered_samples = self.filter_data(samples)
        self._filtered_data.add(filtered_samples)

    def do_training(self, training_data=None):
        training_data = self._prepare_training_data(training_data)

        for i, ch_data in enumerate(training_data):
            std, median = self._calculate_std_and_median(ch_data)
            self._set_threshold(i, std, median)

    def set_filters(self, filters):
        offline_filters = [
            Filter(self._sampling_rate, *filter[:4], ftype=filter[4], use_filtfilt=True) for filter in filters
        ]
        filter_def = [
            {"a": filter.a, 'b': filter.b} for filter in offline_filters
        ]
        self._filter_def = filter_def
        self._online_filters = self._create_new_filters()

    def _create_new_filters(self):
        return ArrayCascadeFilter(self._filter_def, int(self._forget * self._sampling_rate))

    def filter_data(self, samples, online=True):
        if online:
            filters = self._online_filters
        else:
            filters = self._create_new_filters()
        return filters.get_initialized(samples.T).T

    def get_tags(self, filtered_samples, start_timestamp=0):
        filtered_samples = self._prepare_samples(filtered_samples)
        for i, ch_data in enumerate(filtered_samples):
            threshold = self._current_thresholds[i]
            bad_samples = self._calculate_bad_samples(ch_data, threshold)
            yield from self._generate_tags(i, bad_samples, start_timestamp)

    def _generate_tags(self, channel_number, bad_samples, start_timestamp):
        bad_sample_indexes = np.nonzero(bad_samples)[0]
        if not len(bad_sample_indexes):
            return
        merge_window = self._merge_window
        min_tag_duration = int(self.MIN_DURATION * self._sampling_rate)

        def _gen_tag(tag_start, tag_stop):
            if tag_stop - tag_start < min_tag_duration:
                ext = min_tag_duration - (tag_stop - tag_start)
                tag_start = max(tag_start - ext / 2, 0)
                tag_stop = tag_start + min_tag_duration

            tag = {'channels': str(channel_number),
                   'start_timestamp': 1. * tag_start / self._sampling_rate + start_timestamp,
                   'name': self.ARTIFACT_NAME,
                   'end_timestamp': 1. * tag_stop / self._sampling_rate + start_timestamp,
                   'desc': {'thr': "{}".format(self._current_thresholds[channel_number]),
                            'val': "{}".format(bad_samples[int(tag_start):int(tag_stop + 1)].max()),
                            'typ': self._channel_thresholds[channel_number][0]}}
            return tag

        tag_start = tag_stop = bad_sample_indexes[0]
        for i in bad_sample_indexes[1:]:
            if i - tag_stop > merge_window + 1:  # +1 is for backward compatybility with ArtTagsWriter
                yield _gen_tag(tag_start, tag_stop)
                tag_start = tag_stop = i
            else:
                tag_stop = i
        yield _gen_tag(tag_start, tag_stop)

    def _set_threshold(self, ch_nr, std, median):
        channel_thresholds = Thresholds(abs=self._thresholds.abs,
                                        med=self._thresholds.med * median,
                                        std=self._thresholds.std * std + median)
        threshold_type = Thresholds._fields[np.argmin(channel_thresholds)]
        self._channel_thresholds[ch_nr] = (threshold_type, channel_thresholds)
        self._current_thresholds[ch_nr] = min(*channel_thresholds)

    def _prepare_training_data(self, training_data=None):
        if training_data is None:
            return self._filtered_data[:]
        else:
            return self.filter_data(training_data, False)

    def _calculate_bad_samples(self, ch_data, threshold):
        raise NotImplemented

    def _calculate_std_and_median(self, ch_data):
        raise NotImplemented

    def _prepare_samples(self, samples):
        return samples

    def find_artifacts(self, from_seconds_ago, to_seconds_ago=0, start_timestamp=0):
        from_samples_ago = int(from_seconds_ago * self._sampling_rate)
        if to_seconds_ago == 0:
            to_samples_ago = None
            requested_length = from_samples_ago
        else:
            to_samples_ago = -int(to_seconds_ago * self._sampling_rate)
            requested_length = from_samples_ago + to_samples_ago
        filtered_samples = self._filtered_data[-from_samples_ago:to_samples_ago]
        missing_samples = requested_length - filtered_samples.shape[1]
        start_timestamp += missing_samples / self._sampling_rate
        yield from self.get_tags(filtered_samples, start_timestamp)

    def find_artifacts_in_samples(self, samples):
        filtered_samples = self.filter_data(samples, online=False)
        missing_samples = samples.shape[1] - filtered_samples.shape[1]
        yield from self.get_tags(filtered_samples, missing_samples / self._sampling_rate)


class OutliersArtifactDetector(ArtifactDetector):
    ARTIFACT_NAME = "outlier"

    _mean = None

    def _prepare_training_data(self, training_data=None):
        training_data = super(OutliersArtifactDetector, self)._prepare_training_data(training_data)
        self._mean = training_data.mean(axis=1)[:, np.newaxis]
        return training_data - self._mean

    def _prepare_samples(self, samples):
        samples = super(OutliersArtifactDetector, self)._prepare_samples(samples)
        if self._mean is None:
            return np.abs(samples - samples.mean(axis=1)[:, np.newaxis])
        return np.abs(samples - self._mean)

    def _calculate_std_and_median(self, ch_data):
        std_thr = self._thresholds.std
        std = ch_data.std()
        while True:
            ch_data = ch_data[np.abs(ch_data) < (std_thr * std + np.median(ch_data))]
            new_std = ch_data.std()
            if new_std >= std or np.isnan(new_std):
                break
            std = new_std
        median = np.median(np.abs(ch_data))
        return std, median

    def _calculate_bad_samples(self, ch_data, threshold):
        return (ch_data > threshold) * ch_data


class P2PArtifactDetector(ArtifactDetector):
    ARTIFACT_NAME = 'p2p'

    def _calculate_std_and_median(self, ch_data):
        width = self._merge_window
        StdThr = self._thresholds.std
        mm = minimax(ch_data, width)
        # przesuniecie o polowe szerokosci okna i ponowne obliczenie
        mm_2 = minimax(np.roll(ch_data, int(-width / 2)), width)
        # obliczenie progu dla std
        MM = np.concatenate([mm, mm_2])
        std = MM.std()
        while True:
            MM = MM[MM < StdThr * std + np.median(MM)]
            new_std = MM.std()
            if MM.std() < std:
                std = new_std
            else:
                break
        return std, np.median(MM)

    def _calculate_bad_samples(self, ch_data, threshold):
        width = self._merge_window
        mm = minimax(ch_data, width)
        mm_2 = minimax(np.roll(ch_data, int(-width / 2)), width)
        # jednostronnie, bo interesują nas tylko przypadki o zbyt dużej, a nie zbyt małej amplitudzie
        bad_segment = mm * (mm > threshold)
        bad_segment_2 = mm_2 * (mm_2 > threshold)  # jednostronnie przesunięte
        bad_samples = np.max(
            [bad_segment2bad_samples(bad_segment, width, ch_data.size),
             bad_segment2bad_samples(bad_segment_2, width, ch_data.size, int(width / 2))],
            axis=0)
        return bad_samples


class SlopesArtifactDetector(P2PArtifactDetector):
    ARTIFACT_NAME = 'slope'


class SlowesArtifactDetector(P2PArtifactDetector):
    ARTIFACT_NAME = 'slow'


class MusclesArtifactDetector(ArtifactDetector):
    ARTIFACT_NAME = 'muscle'

    _mean = None

    def __init__(self, *args, **kwargs):
        line_fs = [50.]
        filters_muscle = []
        for line_f in line_fs:
            filters_muscle.append([[line_f - 2.5, line_f + 2.5], [line_f - 0.1, line_f + 0.1], 3.0, 25., "cheby2"])
        self.DEFAULT_FILTERS = filters_muscle
        self._range = kwargs.pop('range')
        super(MusclesArtifactDetector, self).__init__(*args, **kwargs)

    def _prepare_training_data(self, training_data=None):
        training_data = super(MusclesArtifactDetector, self)._prepare_training_data(training_data)
        self._mean = training_data.mean(axis=1)[:, np.newaxis]
        return training_data - self._mean

    def _calculate_std_and_median(self, ch_data):
        width = self._merge_window
        std_thr = self._thresholds.std
        Pmsc, Pmsc_2 = self._get_powers(ch_data, width)
        # obliczenie progu dla std
        PP = np.concatenate([Pmsc, Pmsc_2])
        std = PP.std()
        while True:
            PP = PP[PP < std_thr * std + np.median(PP)]
            new_std = PP.std()
            if new_std < std:
                std = new_std
            else:
                break
        return std, np.median(PP)

    def _prepare_samples(self, samples):
        samples = super(MusclesArtifactDetector, self)._prepare_samples(samples)
        if self._mean is None:
            return samples - samples.mean(axis=1)[:, np.newaxis]
        return samples - self._mean

    def _get_powers(self, ch_data, width):
        Pmsc, freqs = power(ch_data, width, self._range, self._sampling_rate)
        # przesunięcie o połowę szerokości okna i ponowne obliczenie
        Pmsc_2, freqs = power(np.roll(ch_data, int(-width / 2)), width, self._range, self._sampling_rate)
        return Pmsc, Pmsc_2

    def _calculate_bad_samples(self, ch_data, threshold):
        width = self._merge_window
        Pmsc, Pmsc_2 = self._get_powers(ch_data, width)
        # jednostronnie, bo interesują nas tylko przypadki o zbyt dużej, a nie zbyt małej mocy
        bad_segment = Pmsc * (Pmsc > threshold)
        bad_segment_2 = Pmsc_2 * (Pmsc_2 > threshold)  # jednostronnie
        bad_samples = np.max(
            [bad_segment2bad_samples(bad_segment, width, ch_data.size),
             bad_segment2bad_samples(bad_segment_2, width, ch_data.size, int(width / 2))],
            axis=0)
        return bad_samples


class Filter:
    def __init__(self, sampling_frequency, wp, ws, gpass, gstop, analog=0, ftype='ellip', output='ba', unit='hz',
                 use_filtfilt=True,
                 meancorr=1.0):
        if unit == 'radians':
            b, a = signal.iirdesign(wp, ws, gpass, gstop, analog, ftype, output)
        elif unit == 'hz':
            nyquist = float(sampling_frequency) / 2.0
            try:
                wp = wp / nyquist
                ws = ws / nyquist
            except TypeError:
                wp = [i / nyquist for i in wp]
                ws = [i / nyquist for i in ws]
            b, a = signal.iirdesign(wp, ws, gpass, gstop, analog, ftype, output)
        else:
            raise Exception("Unknown Unit!")
        self.a = a
        self.b = b
        self._use_filtfilt = use_filtfilt
        self._mean_corr = meancorr

    def filter(self, samples):
        if self._use_filtfilt:
            filtered = signal.filtfilt(self.b, self.a,
                                       samples - np.mean(samples, axis=1)[:, np.newaxis] * self._mean_corr,
                                       axis=1)
            return filtered
        else:
            filtered = signal.lfilter(self.b, self.a, samples)
            return filtered
