# -*- coding: utf-8 -*-
# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.
import json
import os
import re
import time

import numpy as np

from braintech.obci.signal_processing.buffers import AutoRingBuffer
from braintech.obci.signal_processing.signal.data_generic_write_proxy import SamplePacket
from braintech.obci.core.utils.openbci_logging import get_logger
from .artifact_detector import OutliersArtifactDetector, SlopesArtifactDetector, \
    SlowesArtifactDetector, MusclesArtifactDetector, Thresholds

LOGGER = get_logger('eeg')


def convert_to_underscore(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


#  progi odnoszące się do std lub mediany są szacowane dla każdego kanału osobno
#  tagi są zaznaczane wg najostrzejszego z kryteriów
#  żeby nie korzystać z danego kryterium wystarczy podać wartość np.inf
#  każdy tag ma zapisane informacje:
#   - które kryterium było wykorzystane (typ:'abs', 'std' lub 'med')
#   - jaki był ostateczny próg (thr:wartość) w µV
#   - jaka wartość przekroczyła próg (val:wartość) w µV

class EEGArtifactsDetectorConfig:
    default_config_path = os.path.expanduser("~/.obci/artifact_detector_config.json")
    SLOPE_WINDOW = 0.0704  # [s] #szerokość okna, w którym szukamy iglic i stromych zboczy (slope)
    SLOPES = Thresholds(
        abs=100,  # [µV] bezwzględna wartość amplitudy peak2peak w oknie SlopeWindow
        std=7,  # wartość peak2peak w oknie, jako wielokrotność std7,
        med=6,  # wartość peak2peak w oknie, jako wielokrotność mediany
    )

    SLOW_WINDOW = 0.5  # [s] #szerokość okna, w którym szukamy fal wolnych (slow)
    SLOWS = Thresholds(
        abs=100,  # [µV] bezwzględna wartość amplitudy peak2peak w oknie SlowWindow
        std=6,  # wartość peak2peak w oknie, jako wielokrotność std
        med=4,  # wartość peak2peak w oknie, jako wielokrotność mediany
    )
    OUTLIERS_MERGE_WINDOW = .1  # [s] #szerokość okna, w którym łączymy sąsiadujące obszary z próbkami outlier
    OUTLIERS = Thresholds(
        abs=100,  # [µV] bezwzględna wartość amplitudy (liczona od 0)
        std=7,  # wartość amplitudy, jako wielokrotność std
        med=13,  # wartość amplitudy, jako wielokrotność mediany
    )

    MUSCLES_FREQ_RANGE = [40, 90]  # [Hz] przedział częstości, w którym szukamy artefaktów mięśniowych
    MUSCLES = Thresholds(
        abs=50,  # [µV²] bezwzględna wartość średniej mocy na próbkę w zakresie częstości MusclesFreqRange
        std=7,  # wartość amplitudy, jako wielokrotność std
        med=5,  # wartość amplitudy, jako wielokrotność mediany
    )

    FORGET = 0  # [s] długość fragmentów na początku i końcu sygnału, dla których nie zapisujemy tagów

    def _attr_name(self, name):
        underscored = convert_to_underscore(name)
        if underscored.endswith('thr'):
            cfg, thr, _ = underscored.split('_')
            return cfg.upper(), thr
        elif underscored == 'outliers_merge_win':
            return 'OUTLIERS_MERGE_WINDOW'
        else:
            return underscored.upper()

    def __getattr__(self, item):
        attr_name = self._attr_name(item)
        if isinstance(attr_name, tuple):
            return getattr(getattr(self, attr_name[0]), attr_name[1])
        else:
            raise AttributeError(attr_name)

    def __setattr__(self, key, value):
        if key in self.__class__.__dict__:
            super(EEGArtifactsDetectorConfig, self).__setattr__(key, value)
        else:
            attr_name = self._attr_name(key)
            if isinstance(attr_name, tuple):
                value = getattr(self, attr_name[0])._replace(**{attr_name[1]: value})
                attr_name = attr_name[0]
            setattr(self, attr_name, value)

    @classmethod
    def from_dict(cls, _dict):
        config = cls()
        for key, val in _dict.items():
            setattr(config, key, val)
        return config

    @classmethod
    def _json_param_dict_to_class_params(cls, param_dict):
        for param in cls.params():
            try:
                param_value = param_dict[param]
                try:
                    # thresholds
                    if len(param_value) == 3:
                        param_value_to_write = Thresholds(*param_value)
                    # muscle range
                    else:
                        param_value_to_write = list(param_value)
                except TypeError:
                    param_value_to_write = param_value
                setattr(cls, param, param_value_to_write)
            except IndexError:
                pass

    @classmethod
    def from_json(cls, path=None):
        if path is None:
            path = cls.default_config_path
        try:
            with open(path) as f:
                param_dict = json.load(f)
                cls._json_param_dict_to_class_params(param_dict)
        except FileNotFoundError:
            c = cls()
            c.to_json(path)
            return c
        return cls()

    def _get_param_dict(self):
        param_dict = {}
        for param in self.params():
            param_dict[param] = getattr(self, param)
        return param_dict

    def __repr__(self):
        return "<EEGArtifactDetectorConfig {}>".format(self._get_param_dict())

    def to_json(self, path=None):
        if path is None:
            path = self.default_config_path
        param_dict = self._get_param_dict()
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, 'w') as f:
            json.dump(param_dict, f)

    @classmethod
    def params(cls):
        return [name for name in cls.__dict__ if name == name.upper()]


class EEGArtifactsDetector:
    def __init__(self, sample_rate, channel_names, channel_gains=None,
                 channel_offsets=None, config=None, max_data_duration=0, training_duration=10.0,
                 update_training=False):
        self.sample_rate = sample_rate
        self.channel_names = channel_names
        self.training_duration = training_duration
        self._config = config or EEGArtifactsDetectorConfig()
        self._new_tags = []
        no_channels = len(channel_names)
        self._detectors = [
            OutliersArtifactDetector(config.OUTLIERS, no_channels, sample_rate,
                                     config.OUTLIERS_MERGE_WINDOW, forget=config.FORGET,
                                     max_data_duration=max_data_duration),
            SlopesArtifactDetector(config.SLOPES, no_channels, sample_rate,
                                   config.SLOPE_WINDOW, forget=config.FORGET,
                                   max_data_duration=max_data_duration),
            SlowesArtifactDetector(config.SLOWS, no_channels, sample_rate,
                                   config.SLOW_WINDOW, forget=config.FORGET,
                                   max_data_duration=max_data_duration),
            MusclesArtifactDetector(config.MUSCLES, no_channels, sample_rate,
                                    config.SLOW_WINDOW, forget=config.FORGET, range=config.MUSCLES_FREQ_RANGE,
                                    max_data_duration=max_data_duration)
        ]
        self._gains = np.array(channel_gains or [1.0] * no_channels).reshape(-1, 1)
        self._offsets = np.array(channel_offsets or [0.0] * no_channels).reshape(-1, 1)
        self._update_training = update_training
        self._trained = False
        self._last_training_time = 0
        self._first_sample_timestamp = None
        self._last_sample_end_timestamp = None

        self._last_decision_vis = 0
        self._max_data_duration = max_data_duration

    def do_training_if_needed(self, uvolt_samples=None):
        if not self._trained or self._update_training:
            if self._last_sample_end_timestamp >= self._last_training_time + self.training_duration:
                self.do_training(uvolt_samples)

    def do_training(self, uvolt_samples=None):
        start = time.time()
        for d in self._detectors:
            d.do_training(uvolt_samples)
        self._trained = True
        LOGGER.info("artifact detectors trained in %.2f s", time.time() - start)
        self._last_training_time = self._last_sample_end_timestamp
        self._log_tresholds()

    def _log_tresholds(self):
        detector_texts = ['\n']
        for d in self._detectors:
            artifact_detector_name = "Detector: {}".format(d.ARTIFACT_NAME)

            thr_per_chnl = '\tSelected thresholds per channel:'
            thr_per_chnl_data = []
            for chnl_nr, chnl in enumerate(self.channel_names):
                thr_per_chnl_data.append("\t\t{}: {} uV".format(chnl, d._current_thresholds[chnl_nr]))
            detector_texts.append('\n'.join([artifact_detector_name, thr_per_chnl] + thr_per_chnl_data))
        LOGGER.info('\n\n'.join(detector_texts))

    def _iter_artifacts(self, samples):
        for d in self._detectors:
            yield from d.get_tags(samples)

    def handle_sample_packet(self, sample_packet: SamplePacket):
        microvolt_samples = self.get_microvolt_samples(sample_packet.samples.T)
        for d in self._detectors:
            d.add_data(microvolt_samples)
        if self._first_sample_timestamp is None:
            self._first_sample_timestamp = sample_packet.ts[0]
            self._last_training_time = self._first_sample_timestamp
        self._last_sample_end_timestamp = sample_packet.ts[-1] + 1.0 / self.sample_rate
        self.do_training_if_needed()

    def get_artifacts(self, start_timestamp, end_timestamp):
        from_seconds_ago = max(self._last_sample_end_timestamp - start_timestamp, 0)
        to_seconds_ago = max(self._last_sample_end_timestamp - end_timestamp, 0)
        for d in self._detectors:
            yield from d.find_artifacts(from_seconds_ago, to_seconds_ago, start_timestamp)

    def has_artifacts(self, start_timestamp, end_timestamp):
        try:
            art = list(self.get_artifacts(start_timestamp, end_timestamp))[-1]
            LOGGER.info("Last EEG artifact: {}".format(art))
            return True
        except StopIteration:
            return False
        except IndexError:
            return False

    def get_artifacts_from_samples(self, samples):
        microvolt_samples = self.get_microvolt_samples(samples)
        for d in self._detectors:
            yield from d.find_artifacts_in_samples(microvolt_samples)

    def has_artifacts_in_samples(self, samples):
        try:
            next(self.get_artifacts_from_samples(samples))
            return True
        except StopIteration:
            return False

    @classmethod
    def merge_tags(cls, tags):
        all_tags = ([(t['start_timestamp'], False, t) for t in tags]
                    + [(t['end_timestamp'], True, t) for t in tags])

        all_tags.sort(key=lambda t: t[:2])
        open_tags = 0
        merged_tags = []
        current_tag = {}
        for timestamp, tag_end, tag in all_tags:
            if tag_end:
                open_tags -= 1
                if open_tags == 0:
                    current_tag['end_timestamp'] = timestamp
                    current_tag['channels'] = ' '.join([str(c) for c in sorted(current_tag['channels'])])
                    merged_tags.append(current_tag)
            else:
                open_tags += 1
                if open_tags == 1:
                    current_tag = {
                        'start_timestamp': timestamp,
                        'name': 'eeg_artifact',
                        'desc': {},
                        'channels': set(),
                    }
                current_tag['desc'].setdefault(tag['name'], 0)
                current_tag['desc'][tag['name']] += tag['end_timestamp'] - tag['start_timestamp']
                current_tag['channels'].add(tag['channels'])
        return merged_tags

    def get_microvolt_samples(self, samples):
        return samples * self._gains + self._offsets

    def visualise_decisions(self, bci_panel, update_every_s=1):
        now = time.monotonic()
        if (now - self._last_decision_vis) > update_every_s:
            times, states = self.get_whole_buffer_artifact_state()
            bci_panel.update_plot(field_id="None", feature='Artifact state',
                                  x_data=[times],
                                  y_data=[states],
                                  title="Artifact state",
                                  ylabel="Has artifacts",
                                  ylim=[-0.1, 1.1])
            self._last_decision_vis = now

    def get_whole_buffer_artifact_state(self, max_len=20):
        assert max_len <= self._max_data_duration
        window_len_s = 0.50
        buffer_first_ts = self._last_sample_end_timestamp - max_len
        states = []
        times = []
        for ts_start in np.arange(buffer_first_ts, self._last_sample_end_timestamp, window_len_s):
            state = self.has_artifacts(ts_start, ts_start + window_len_s)
            states.append(state)
            times.append(ts_start)
        return times, states


class EpochingEEGArtifactsDetector(EEGArtifactsDetector):
    def __init__(self, *args, epoch_duration, epoch_interval, **kwargs):
        super(EpochingEEGArtifactsDetector, self).__init__(*args, **kwargs)
        self._epoch_buffer = AutoRingBuffer(
            from_sample=int(epoch_duration * self.sample_rate),
            samples_count=int(epoch_duration * self.sample_rate),
            every=int(epoch_interval * self.sample_rate),
            num_of_channels=1,
            ret_func=self._find_artifacts,
            copy_on_ret=False
        )
        self._new_tags = []

    def handle_sample_packet(self, sample_packet: SamplePacket):
        super(EpochingEEGArtifactsDetector, self).handle_sample_packet(sample_packet)
        timestamps = sample_packet.ts[np.newaxis, :]
        self._epoch_buffer.add_samples_to_buffer(timestamps)
        new_tags = self._new_tags
        self._new_tags = []
        return new_tags

    def _find_artifacts(self, timestamps):
        start_timestamp = timestamps[0, 0]
        end_timestamp_inclusive = timestamps[0, -1] + 1.0 / self.sample_rate
        for tag in self.get_artifacts(start_timestamp, end_timestamp_inclusive):
            self._new_tags.append(tag)
