# -*- coding: utf-8 -*-
# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.
import json
from copy import deepcopy
from pathlib import Path

from unittest import mock
import numpy as np
import pytest

from braintech.obci.signal_processing.read_manager import ReadManager
from braintech.obci.signal_processing.signal.data_generic_write_proxy import SamplePacket
from braintech.obci.signal_processing.tags.tags_file_writer import TagsFileWriter
from bci_artifacts.eeg.artifact_detector import OutliersArtifactDetector, \
    SlopesArtifactDetector, SlowesArtifactDetector, MusclesArtifactDetector
from bci_artifacts.eeg.eeg_artifacts_detector import EEGArtifactsDetector, EEGArtifactsDetectorConfig, \
    EpochingEEGArtifactsDetector

TEST_DATA_FILE = Path(
    __file__).parent / 'test_data' / "with_montage_and_ICA_CLEANED"

VALID_CHANNELS = ['Fp1', 'Fpz', 'Fp2', 'F7', 'F3', 'Fz', 'F4', 'F8', 'T3', 'C3',
                  'Cz', 'C4', 'T4', 'T5', 'P3', 'Pz', 'T6', 'O1', 'Oz', 'O2']


# @pytest.mark.skip()
def test_outliers_detector(read_manager: ReadManager, config: EEGArtifactsDetectorConfig):
    out_detector = OutliersArtifactDetector(config.OUTLIERS, len(read_manager.channels_names),
                                            read_manager.sampling_frequency,
                                            config.OUTLIERS_MERGE_WINDOW, forget=config.FORGET)
    _test_detector(out_detector, read_manager, 'outlier')


def test_slopes_detector(read_manager: ReadManager, config: EEGArtifactsDetectorConfig):
    out_detector = SlopesArtifactDetector(config.SLOPES, len(read_manager.channels_names),
                                          read_manager.sampling_frequency,
                                          config.SLOPE_WINDOW, forget=config.FORGET)
    _test_detector(out_detector, read_manager, 'slope')


def test_slows_detector(read_manager: ReadManager, config: EEGArtifactsDetectorConfig):
    out_detector = SlowesArtifactDetector(config.SLOWS, len(read_manager.channels_names),
                                          read_manager.sampling_frequency,
                                          config.SLOW_WINDOW, forget=config.FORGET)
    _test_detector(out_detector, read_manager, 'slow')


def test_muscles_detector(read_manager: ReadManager, config: EEGArtifactsDetectorConfig):
    out_detector = MusclesArtifactDetector(config.MUSCLES, len(read_manager.channels_names),
                                           read_manager.sampling_frequency,
                                           config.SLOW_WINDOW, forget=config.FORGET, range=config.MUSCLES_FREQ_RANGE)
    _test_detector(out_detector, read_manager, 'muscle')


def _test_detector(detector, read_manager, tag_name):
    signal_length = read_manager.signal_duration
    samples = read_manager.get_microvolt_samples(0, signal_length, 'second')
    for i in range(0, read_manager.number_of_samples, int(read_manager.sampling_frequency)):
        chunk = samples[:, i:int(i + read_manager.sampling_frequency)]
        detector.add_data(chunk)
        detector.do_training()
    new_outlier_tags = list(detector.find_artifacts_in_samples(samples))
    _compare_tags(new_outlier_tags, read_manager.get_tags(tag_name))
    _compare_tags(list(detector.find_artifacts(signal_length, 0)), new_outlier_tags)


def test_eeg_artifacts_detector(read_manager: ReadManager, config: EEGArtifactsDetectorConfig):
    rm = read_manager
    sampling_rate = rm.sampling_frequency
    signal_length = rm.number_of_samples / sampling_rate
    detector = _get_detector(rm,
                             epoch_duration=signal_length,
                             epoch_interval=signal_length,
                             config=config,
                             training_duration=signal_length,
                             max_data_duration=signal_length,
                             update_training=False)
    samples = rm.get_samples(0, signal_length, 'second')
    timestamps = np.arange(0, len(samples.T)) / sampling_rate
    with mock.patch.object(detector, 'do_training', wraps=detector.do_training) as wrapped_do_training:
        tags = detector.handle_sample_packet(SamplePacket(samples.T, timestamps))
        wrapped_do_training.assert_called_with(None)

    writer = TagsFileWriter(str(TEST_DATA_FILE.with_suffix('.artifacts_new.tag')))
    for t in tags:
        writer.tag_received(restore_channel_number(t, rm.channel_indexes))
    writer.finish_saving(0)
    rm_tags = read_manager.get_tags()
    _compare_tags(tags, rm_tags)
    first_artifact = 6.4
    assert detector.has_artifacts(0, first_artifact) is False
    assert detector.has_artifacts_in_samples(samples[:, 0:int(first_artifact * sampling_rate)]) is False
    assert detector.has_artifacts(first_artifact, first_artifact + 2)
    assert detector.has_artifacts_in_samples(
        samples[:, int((first_artifact - config.FORGET) * sampling_rate):int((first_artifact + 2) * sampling_rate)])


def test_training_eeg_artifacts_detector(read_manager: ReadManager, config: EEGArtifactsDetectorConfig):
    new_config = deepcopy(config)
    new_config.FORGET = 2.0
    new_config.MUSCLES = config.MUSCLES._replace(abs=6000)

    rm = read_manager
    sampling_rate = rm.sampling_frequency
    signal_length = rm.number_of_samples / sampling_rate
    epoch_duration = signal_length // 4
    fully_trained_detector = _get_detector(rm,
                                           epoch_duration=epoch_duration,
                                           epoch_interval=epoch_duration,
                                           config=new_config,
                                           training_duration=signal_length,
                                           max_data_duration=signal_length,
                                           update_training=False)
    samples = rm.get_samples(0, signal_length, 'second')
    timestamps = np.arange(0, len(samples.T)) / sampling_rate
    sample_packet = SamplePacket(samples.T, timestamps)
    fully_trained_detector.handle_sample_packet(sample_packet)

    never_trained_detector = _get_detector(rm, epoch_duration=signal_length,
                                           epoch_interval=signal_length,
                                           config=new_config,
                                           training_duration=signal_length * 2,  # never trained
                                           max_data_duration=signal_length,
                                           update_training=False)
    tags = never_trained_detector.handle_sample_packet(sample_packet)

    _round_tags(tags)

    assert tags == [{'channels': '7',
                     'desc': {'thr': '6000', 'typ': 'abs', 'val': '6193'},
                     'end_timestamp': 21.498046875,
                     'name': 'muscle',
                     'start_timestamp': 21.0}
                    ], "All abs thresolds are set to np.inf so only one  tag should be generated"
    trained_detector = _get_detector(rm,
                                     epoch_duration=epoch_duration,
                                     epoch_interval=epoch_duration,
                                     config=new_config,
                                     training_duration=epoch_duration,
                                     max_data_duration=signal_length,
                                     update_training=True)

    half_samples = rm.number_of_samples // 2
    half_packet = SamplePacket(samples[:, :half_samples].T, timestamps[:half_samples])
    tags = trained_detector.handle_sample_packet(half_packet)
    _round_tags(tags)
    assert len(tags) > 10, "Should be partially trained"
    tags_by_fully_trained = fully_trained_detector.handle_sample_packet(half_packet)
    _round_tags(tags_by_fully_trained)
    with pytest.raises(AssertionError):
        _compare_tags(tags, tags_by_fully_trained)

    second_half_packet = SamplePacket(samples[:, half_samples:].T, timestamps[half_samples:])
    tags = trained_detector.handle_sample_packet(second_half_packet)
    _round_tags(tags)
    tags_by_fully_trained = fully_trained_detector.handle_sample_packet(second_half_packet)
    _round_tags(tags_by_fully_trained)
    _compare_tags(tags, tags_by_fully_trained, "Detector should be trained as much as fully_trained one")

    next_packet = SamplePacket(second_half_packet.samples, second_half_packet.ts + signal_length)
    with mock.patch.object(trained_detector, 'do_training', wraps=trained_detector.do_training) as wrapped_do_training:
        tags = trained_detector.handle_sample_packet(next_packet)
        wrapped_do_training.assert_called_with(None)


def _get_detector(rm, **kwargs):
    sampling_rate = rm.sampling_frequency
    return EpochingEEGArtifactsDetector(
        sample_rate=sampling_rate,
        channel_names=rm.get_param('channels_names'),
        channel_gains=[float(g) for g in rm.get_param('channels_gains')],
        channel_offsets=[float(o) for o in rm.get_param('channels_offsets')],
        **kwargs
    )


def test_tag_merging():
    def _tag(start, end, name, channel):
        return {
            'start_timestamp': start,
            'end_timestamp': end,
            'name': name,
            'channels': channel
        }

    def _merged_tag(start, end, channels, **desc):
        return {
            'start_timestamp': start,
            'end_timestamp': end,
            'name': 'eeg_artifact',
            'channels': channels,
            'desc': desc
        }

    tags = [
        _tag(0, 3, 'muscles', 1),
        _tag(0, 3, 'slow', 1),
        _tag(1, 4, 'slow', 2),
        _tag(5, 6, 'slope', 3),
        _tag(6, 7, 'outlier', 4)
    ]
    merged = EEGArtifactsDetector.merge_tags(tags)
    assert merged == [
        _merged_tag(0, 4, "1 2", muscles=3, slow=6),
        _merged_tag(5, 7, "3 4", slope=1, outlier=1)
    ]


def restore_channel_number(tag, channel_indexes):
    new_tag = tag.copy()
    new_tag['channels'] = str(channel_indexes[int(tag['channels'])])
    return new_tag


@pytest.fixture()
def read_manager():
    rm = ReadManager(
        str(TEST_DATA_FILE.with_suffix('.xml')),
        str(TEST_DATA_FILE.with_suffix('.raw')),
        str(TEST_DATA_FILE.with_suffix('.artifacts.tag'))
    )
    channel_indexes = [
        i for i, name in enumerate(rm.channels_names) if name in VALID_CHANNELS
    ]

    def _filter(l):
        return [e for i, e in enumerate(l) if i in channel_indexes]

    samples = rm.get_samples()[channel_indexes, :]
    channel_names = _filter(rm.get_param('channels_names'))
    rm.set_params(dict(channels_names=channel_names,
                       channels_gains=_filter([float(g) for g in rm.get_param('channels_gains')]),
                       channels_offsets=_filter([float(o) for o in rm.get_param('channels_offsets')])))
    rm.set_samples(samples, channel_names)
    tags = rm.get_tags()
    for tag in tags:
        tag['channels'] = str(channel_indexes.index(int(tag['channels'])))
    rm.set_tags(tags)
    rm.channel_indexes = channel_indexes
    yield rm


@pytest.fixture()
def config():
    config = json.load(open(str(TEST_DATA_FILE.with_suffix('.artifact_rej_dict.json'))))
    config = EEGArtifactsDetectorConfig.from_dict(config)
    config.FORGET = 2.0
    return config


def _round_tags(tags):
    for item in tags:
        item['desc']['val'] = str(int(float(item['desc']['val'])))
        item['desc']['thr'] = str(int(float(item['desc']['thr'])))


def _compare_tags(actual_tags, expected_tags, message=None):
    def _key(tag):
        return (tag['channels'], tag['start_timestamp'])

    expected_tags.sort(key=_key)
    actual_tags.sort(key=_key)
    _round_tags(actual_tags)

    assert actual_tags == expected_tags, message
