# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.

import pytest

from bci_artifacts.camera import aggregators
from bci_artifacts.camera import analysis


@pytest.mark.incremental
class TestStateAggregatorEyesStates:
    def test_adding_eye_states(self):
        aggregator = aggregators.StateAggregator()
        aggregator.add_eyes_state(analysis.EyesState.NONE_PRESENT, 1)
        aggregator.add_eyes_state(analysis.EyesState.OPEN, 2)
        aggregator.add_eyes_state(analysis.EyesState.CLOSED, 3)
        aggregator.add_eyes_state(analysis.EyesState.NONE_PRESENT, 4)
        aggregator.add_eyes_state(analysis.EyesState.OPEN, 5)
        aggregator.add_eyes_state(analysis.EyesState.CLOSED, 6)
        assert len(aggregator._eyes_states) == 6

    def test_sorting_added_eye_states(self):
        aggregator = aggregators.StateAggregator()
        aggregator.add_eyes_state(analysis.EyesState.NONE_PRESENT, 4)
        aggregator.add_eyes_state(analysis.EyesState.OPEN, 5)
        aggregator.add_eyes_state(analysis.EyesState.CLOSED, 6)
        aggregator.add_eyes_state(analysis.EyesState.NONE_PRESENT, 1)
        aggregator.add_eyes_state(analysis.EyesState.OPEN, 2)
        aggregator.add_eyes_state(analysis.EyesState.CLOSED, 3)
        assert len(aggregator._eyes_states) == 6
        assert aggregator._eyes_states[0].time == 1
        assert aggregator._eyes_states[1].time == 2
        assert aggregator._eyes_states[2].time == 3
        assert aggregator._eyes_states[3].time == 4
        assert aggregator._eyes_states[4].time == 5
        assert aggregator._eyes_states[5].time == 6

    def test_default_control_state(self):
        aggregator = aggregators.StateAggregator()
        assert aggregator.was_control_state(1, 5) is False
        assert aggregator.was_control_state(1529939694.228137, 1529939709.65216) is False

    def test_states_only_before_selected_range(self):
        aggregator = aggregators.StateAggregator()
        aggregator.add_eyes_state(analysis.EyesState.NONE_PRESENT, 1)
        assert aggregator.was_control_state(2, 5) is False
        aggregator = aggregators.StateAggregator()
        aggregator.add_eyes_state(analysis.EyesState.OPEN, 1)
        assert aggregator.was_control_state(2, 5) is True
        aggregator = aggregators.StateAggregator()
        aggregator.add_eyes_state(analysis.EyesState.CLOSED, 1)
        assert aggregator.was_control_state(2, 5) is False

    def test_states_only_after_selected_range(self):
        aggregator = aggregators.StateAggregator()
        aggregator.add_eyes_state(analysis.EyesState.NONE_PRESENT, 10)
        assert aggregator.was_control_state(2, 5) is False
        aggregator = aggregators.StateAggregator()
        aggregator.add_eyes_state(analysis.EyesState.OPEN, 10)
        assert aggregator.was_control_state(2, 5) is False
        aggregator = aggregators.StateAggregator()
        aggregator.add_eyes_state(analysis.EyesState.CLOSED, 10)
        assert aggregator.was_control_state(2, 5) is False

    def test_getting_first_previous_point_in_relevant_points(self):
        aggregator = aggregators.StateAggregator()
        artificial_start_point = aggregators.StateChangePoint(
            state=analysis.EyesState.NONE_PRESENT,
            time=0,
        )
        expected_default = analysis.EyesState.NONE_PRESENT
        actual = aggregator._get_relevant_points(1, 5, aggregator._eyes_states,
                                                 artificial_start_point)
        assert len(actual) == 1
        assert actual[0].state == expected_default
        aggregator.add_eyes_state(analysis.EyesState.OPEN, 0)
        expected = analysis.EyesState.OPEN
        actual = aggregator._get_relevant_points(1, 5, aggregator._eyes_states,
                                                 artificial_start_point)
        assert len(actual) == 1
        assert actual[0].state == expected
        aggregator.add_eyes_state(analysis.EyesState.CLOSED, 1)
        expected = analysis.EyesState.CLOSED
        actual = aggregator._get_relevant_points(1, 5, aggregator._eyes_states,
                                                 artificial_start_point)
        assert len(actual) == 1
        assert actual[0].state == expected
        aggregator.add_eyes_state(analysis.EyesState.NONE_PRESENT, 2)
        expected_first = analysis.EyesState.CLOSED
        actual = aggregator._get_relevant_points(1, 5, aggregator._eyes_states,
                                                 artificial_start_point)
        assert actual[0].state == expected_first

    def test_getting_relevant_points(self):
        aggregator = aggregators.StateAggregator()
        artificial_start_point = aggregators.StateChangePoint(
            state=analysis.EyesState.NONE_PRESENT,
            time=0,
        )
        aggregator.add_eyes_state(analysis.EyesState.OPEN, 1)
        aggregator.add_eyes_state(analysis.EyesState.CLOSED, 2)
        aggregator.add_eyes_state(analysis.EyesState.NONE_PRESENT, 3)
        aggregator.add_eyes_state(analysis.EyesState.CLOSED, 4)
        aggregator.add_eyes_state(analysis.EyesState.OPEN, 5)
        aggregator.add_eyes_state(analysis.EyesState.NONE_PRESENT, 6)

        actual = aggregator._get_relevant_points(0, 0, aggregator._eyes_states,
                                                 artificial_start_point)
        expected = [
            (analysis.EyesState.NONE_PRESENT, 0),
        ]
        assert [(a.state, a.time) for a in actual] == expected

        actual = aggregator._get_relevant_points(0, 1, aggregator._eyes_states,
                                                 artificial_start_point)
        expected = [
            (analysis.EyesState.NONE_PRESENT, 0),
            (analysis.EyesState.OPEN, 1),
        ]
        assert [(a.state, a.time) for a in actual] == expected

        actual = aggregator._get_relevant_points(0, 2, aggregator._eyes_states,
                                                 artificial_start_point)
        expected = [
            (analysis.EyesState.NONE_PRESENT, 0),
            (analysis.EyesState.OPEN, 1),
            (analysis.EyesState.CLOSED, 2),
        ]
        assert [(a.state, a.time) for a in actual] == expected

        actual = aggregator._get_relevant_points(1, 2, aggregator._eyes_states,
                                                 artificial_start_point)
        expected = [
            (analysis.EyesState.OPEN, 1),
            (analysis.EyesState.CLOSED, 2),
        ]
        assert [(a.state, a.time) for a in actual] == expected

        actual = aggregator._get_relevant_points(1.5, 2, aggregator._eyes_states,
                                                 artificial_start_point)
        expected = [
            (analysis.EyesState.OPEN, 1),
            (analysis.EyesState.CLOSED, 2),
        ]
        assert [(a.state, a.time) for a in actual] == expected

        actual = aggregator._get_relevant_points(2, 5, aggregator._eyes_states,
                                                 artificial_start_point)
        expected = [
            (analysis.EyesState.CLOSED, 2),
            (analysis.EyesState.NONE_PRESENT, 3),
            (analysis.EyesState.CLOSED, 4),
            (analysis.EyesState.OPEN, 5),
        ]
        assert [(a.state, a.time) for a in actual] == expected

        actual = aggregator._get_relevant_points(5, 88, aggregator._eyes_states,
                                                 artificial_start_point)
        expected = [
            (analysis.EyesState.OPEN, 5),
            (analysis.EyesState.NONE_PRESENT, 6),
        ]
        assert [(a.state, a.time) for a in actual] == expected

    def test_time_intersection_counting_when_missing(self):
        aggregator = aggregators.StateAggregator()
        assert aggregator._count_time_intersection(0, 1, 2, 3) == 0
        assert aggregator._count_time_intersection(2, 3, 0, 1) == 0
        assert aggregator._count_time_intersection(0, 1, 1, 2) == 0
        assert aggregator._count_time_intersection(1, 2, 0, 1) == 0

    def test_time_intersection_counting_when_enveloping(self):
        aggregator = aggregators.StateAggregator()
        assert aggregator._count_time_intersection(0, 3, 1, 2) == 1
        assert aggregator._count_time_intersection(1, 2, 0, 3) == 1
        assert aggregator._count_time_intersection(0, 3, 0, 2) == 2
        assert aggregator._count_time_intersection(0, 2, 0, 3) == 2
        assert aggregator._count_time_intersection(0, 3, 0, 3) == 3

    def test_time_intersection_counting_when_both_have_independent_parts(self):
        aggregator = aggregators.StateAggregator()
        assert aggregator._count_time_intersection(0, 3, 2, 4) == 1
        assert aggregator._count_time_intersection(2, 4, 0, 3) == 1

    def test_counting_control_state_percentage(self):
        aggregator = aggregators.StateAggregator()

        states = [
            aggregators.StateChangePoint(state=analysis.EyesState.NONE_PRESENT, time=0),
        ]
        actual = aggregator._count_control_state_percentage(0, 10, states)
        expected = 0
        assert actual == expected

        states = [
            aggregators.StateChangePoint(state=analysis.EyesState.OPEN, time=0),
        ]
        actual = aggregator._count_control_state_percentage(0, 10, states)
        expected = 1
        assert actual == expected

        states = [
            aggregators.StateChangePoint(state=analysis.EyesState.CLOSED, time=0),
        ]
        actual = aggregator._count_control_state_percentage(0, 10, states)
        expected = 0
        assert actual == expected

        states = [
            aggregators.StateChangePoint(state=analysis.EyesState.CLOSED, time=0),
            aggregators.StateChangePoint(state=analysis.EyesState.OPEN, time=5),
        ]
        actual = aggregator._count_control_state_percentage(0, 10, states)
        expected = 0.5
        assert actual == expected

        states = [
            aggregators.StateChangePoint(state=analysis.EyesState.OPEN, time=0),
            aggregators.StateChangePoint(state=analysis.EyesState.CLOSED, time=5),
        ]
        actual = aggregator._count_control_state_percentage(0, 10, states)
        expected = 0.5
        assert actual == expected

        states = [
            aggregators.StateChangePoint(state=analysis.EyesState.OPEN, time=0),
            aggregators.StateChangePoint(state=analysis.EyesState.CLOSED, time=3),
            aggregators.StateChangePoint(state=analysis.EyesState.NONE_PRESENT, time=6),
        ]
        actual = aggregator._count_control_state_percentage(0, 10, states)
        expected = 0.3
        assert actual == expected

        states = [
            aggregators.StateChangePoint(state=analysis.EyesState.OPEN, time=0),
            aggregators.StateChangePoint(state=analysis.EyesState.CLOSED, time=3),
            aggregators.StateChangePoint(state=analysis.EyesState.OPEN, time=5),
            aggregators.StateChangePoint(state=analysis.EyesState.NONE_PRESENT, time=6),
        ]
        actual = aggregator._count_control_state_percentage(0, 10, states)
        expected = 0.4
        assert actual == expected

    def test_adding_head_states(self):
        aggregator = aggregators.StateAggregator()
        aggregator.add_head_state(analysis.HeadState.NONE_PRESENT, 1)
        aggregator.add_head_state(analysis.HeadState.NOT_MOVING, 2)
        aggregator.add_head_state(analysis.HeadState.MOVING, 3)
        aggregator.add_head_state(analysis.HeadState.NONE_PRESENT, 4)
        aggregator.add_head_state(analysis.HeadState.NOT_MOVING, 5)
        aggregator.add_head_state(analysis.HeadState.MOVING, 6)
        assert len(aggregator._head_states) == 6

    def test_sorting_added_head_states(self):
        aggregator = aggregators.StateAggregator()
        aggregator.add_head_state(analysis.HeadState.NONE_PRESENT, 4)
        aggregator.add_head_state(analysis.HeadState.NOT_MOVING, 5)
        aggregator.add_head_state(analysis.HeadState.MOVING, 6)
        aggregator.add_head_state(analysis.HeadState.NONE_PRESENT, 1)
        aggregator.add_head_state(analysis.HeadState.NOT_MOVING, 2)
        aggregator.add_head_state(analysis.HeadState.MOVING, 3)
        assert len(aggregator._head_states) == 6
        assert aggregator._head_states[0].time == 1
        assert aggregator._head_states[1].time == 2
        assert aggregator._head_states[2].time == 3
        assert aggregator._head_states[3].time == 4
        assert aggregator._head_states[4].time == 5
        assert aggregator._head_states[5].time == 6

    def test_checking_if_had_artifact(self):
        aggregator = aggregators.StateAggregator()
        assert aggregator.had_artifact(0, 10) is False

        aggregator.add_head_state(analysis.HeadState.NONE_PRESENT, 1)
        aggregator.add_head_state(analysis.HeadState.NOT_MOVING, 2)
        aggregator.add_head_state(analysis.HeadState.MOVING, 3)
        aggregator.add_head_state(analysis.HeadState.NONE_PRESENT, 4)
        aggregator.add_head_state(analysis.HeadState.NOT_MOVING, 5)
        aggregator.add_head_state(analysis.HeadState.MOVING, 6)

        assert aggregator.had_artifact(0, 1) is False
        assert aggregator.had_artifact(0, 2.5) is False
        assert aggregator.had_artifact(0, 3) is True
        assert aggregator.had_artifact(3, 5) is True
        assert aggregator.had_artifact(4, 5) is False
        assert aggregator.had_artifact(5, 5.5) is False
        assert aggregator.had_artifact(6, 20) is True
        assert aggregator.had_artifact(0, 10) is True
        assert aggregator.had_artifact(7, 10) is True
