# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved
import asyncio
import threading
from asyncio import CancelledError

import numpy as np
import pytest
from mock import MagicMock

from bci_demo_ssvep.simple_ssvep.frontend import Frontend
from bci_demo_ssvep.simple_ssvep.meta_calibrator import SSVEPMetaCalibrator
from obci_p300.common.signal_preprocessor import SignalPreprocessor
from bci_ssvep.logic.ssvep_metacalibrator import FieldFreqsProvider
from obci_p300.common.random_amp_with_sine.random_amplifier_with_sine import RandomAmplifierWithSine


def _sampling_thread(amplifier: RandomAmplifierWithSine, meta_calibrator: SSVEPMetaCalibrator, stop_event):
    while not stop_event.is_set():
        try:
            calib_type = meta_calibrator._current_analysis.calibration_type
            if calib_type is not None:
                amplifier.set_sine_freq(float(calib_type))
        except AttributeError:
            pass
        packet = amplifier.get_samples(100)
        meta_calibrator.handle_sample_packet(packet)


class _TestFrontend(Frontend):

    def __init__(self, loop):
        super().__init__(loop)
        self.decision = asyncio.Event(loop=loop)
        self._last_field = None

    async def display_prediction(self, field, feedback_duration, is_correct=True):
        await super(_TestFrontend, self).display_prediction(field, 0.1, is_correct)
        self._last_field = field
        self.decision.set()


@pytest.fixture()
def metacalibration(fast_analysis):
    amplifier = RandomAmplifierWithSine()
    amplifier.sampling_rate = 500
    amplifier.active_channels = amplifier.active_channels[:2]
    amplifier.start_sampling()
    loop = asyncio.get_event_loop()
    frontend = _TestFrontend(loop=loop)
    visualizer = MagicMock()
    freq_provider = FieldFreqsProvider(min_freq=33,
                                       max_freq=36,
                                       step=2,
                                       n=2)
    blinker = MagicMock()
    meta_calibrator = SSVEPMetaCalibrator(
        frontend=frontend,
        field_freqs_provider=freq_provider,
        sample_rate=amplifier.sampling_rate,
        blinker_controller=blinker,
        visualizer=visualizer,
        channels_for_classification=amplifier.active_channels,
    )

    stop_event = threading.Event()
    thread = threading.Thread(target=_sampling_thread, args=(amplifier, meta_calibrator, stop_event), daemon=True)
    thread.start()
    yield meta_calibrator, amplifier, loop, frontend
    stop_event.set()
    thread.join()


WISDOM_PATH = './calibration'


@pytest.mark.timeout(200)
def test_metacalibrator_calibration(metacalibration):
    meta_calibrator, amplifier, loop, frontend = metacalibration
    calibration = meta_calibrator.calibrate(infinite=False)

    async def _interactions(frontend):
        await asyncio.sleep(0.5)
        frontend.handle_key(Frontend.SPACE)

    loop.create_task(_interactions(frontend))
    loop.run_until_complete(calibration)
    meta_calibrator.save(WISDOM_PATH)


@pytest.mark.timeout(200)
def test_metacalibrator_feedback(metacalibration):
    meta_calibrator, amplifier, loop, frontend = metacalibration
    meta_calibrator.load_wisdom(WISDOM_PATH)

    async def _interactions(frontend):
        await asyncio.sleep(0.5)
        amplifier.set_sine_freq(meta_calibrator.field_freqs[0])
        await frontend.decision.wait()
        assert frontend._last_field == 0
        amplifier.set_sine_freq(meta_calibrator.field_freqs[1])
        await asyncio.sleep(0.5)
        frontend.decision.clear()
        await frontend.decision.wait()
        assert frontend._last_field == 1
        task.cancel()

    interactions = loop.create_task(_interactions(frontend))
    task = loop.create_task(meta_calibrator.endless_feedback())
    all_tasks = asyncio.gather(interactions, task, loop=loop)
    try:
        loop.run_until_complete(all_tasks)
    except CancelledError:
        pass


def test_signal_preprocessor():
    preprocessor = SignalPreprocessor(
        all_channels_names=['1', '2', '3', '4'],
        montage_channels_names=['2', '3'],
        excluded_channel_names=['4'],
        classification_channel_names=['1', '4'],
        sampling_rate=500,
        notch_filters=[]
    )
    data = np.array([
        [1, 2, 3, 4],
        [1, 2, 3, 4],
    ]
    )

    expected = np.array([
        [-1.5, 4],
        [-1.5, 4]
    ])
    preprocessed = preprocessor.get_preprocessed_signal(data)
    assert np.array_equal(preprocessed, expected)
    preprocessor = SignalPreprocessor(
        all_channels_names=['1', '2', '3', '4'],
        classification_channel_names=['1'],
        montage_channels_names=[],
        excluded_channel_names=[],
        sampling_rate=500,
        notch_filters=[50, 125]
    )
    times = np.linspace(-np.pi, np.pi, 500)
    sin = 10 * np.sin(50 * times) + 10 * np.sin(125 * times)
    data = np.array([sin, sin, sin, sin]).T
    preprocessed = preprocessor.get_preprocessed_signal(data)
    assert np.max(np.abs(preprocessed[100:])) < 2
