# -*- coding: utf-8 -*-
# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.
import abc
import queue
import time
import warnings
from collections import namedtuple
from logging import getLogger
from multiprocessing import Process, Queue

try:
    import _tkinter
except (RuntimeError, ModuleNotFoundError) as e:
    warnings.warn(str(e))

from bci_panel.matplotlib_bci_panel import BCIPanel

Request = namedtuple('Request', ['name', 'kwargs'])

LOGGER = getLogger("BCIPanel")
LOGGER.setLevel('INFO')


class BaseBCIPanelProcess(metaclass=abc.ABCMeta):
    SIGNAL_UPDATE_RATE = 1
    SIGNAL_CHUNK_LENGTH = 10
    SEPARATION_TYPE = 'std'

    def update_plot(self, **kwargs):
        raise NotImplementedError

    def update_text(self, **kwargs):
        raise NotImplementedError

    def update_hist(self, **kwargs):
        raise NotImplementedError

    def save_plots(self, **kwargs):
        raise NotImplementedError

    def __del__(self):
        self.close()

    def close(self):
        raise NotImplementedError

    def get_separation(self, mean_std):
        if callable(mean_std):
            mean_std = mean_std()
        return mean_std * 4


class DummyBCIPanelProcess(BaseBCIPanelProcess):
    def update_plot(self, **kwargs):
        LOGGER.debug('Running DummyBCIPanel.update_plot.')

    def update_text(self, **kwargs):
        LOGGER.debug('Running DummyBCIPanel.update_text.')

    def update_hist(self, **kwargs):
        LOGGER.debug('Running DummyBCIPanel.update_hist.')

    def save_plots(self, **kwargs):
        LOGGER.debug('Running DummyBCIPanel.save_plots.')

    def close(self):
        LOGGER.debug('Running DummyBCIPanel.close.')


class BCIPanelProcess(BaseBCIPanelProcess):
    """BCIPanel wrapper - runs it in another thread."""
    time_between_updates = 0.1

    def __init__(self):
        self.queue = Queue()
        self.process = Process(target=self._panel_loop, args=(self.queue,), name="BCIPanel Process")
        self.process.daemon = True
        self.process.start()
        self.queue_open = True
        self._last_update_time = None
        self._panel = None

    def update_plot(self, **kwargs):
        """Same as BCIPanel.update_plot."""
        if self.queue_open:
            self.queue.put(Request('update_plot', kwargs))

    def update_text(self, **kwargs):
        """Same as BCIPanel.update_text."""
        if self.queue_open:
            self.queue.put(Request('update_text', kwargs))

    def update_hist(self, **kwargs):
        """Same as BCIPanel.update_hist."""
        if self.queue_open:
            self.queue.put(Request('update_hist', kwargs))

    def save_plots(self, **kwargs):
        """Same as BCIPanel.save_plots."""
        if self.queue_open:
            self.queue.put(Request('save_plots', kwargs))

    def close(self):
        if not self.queue_open:
            return
        self.queue_open = False
        LOGGER.info("Sending close request. Items on queue: %d", self.queue.qsize())
        self.queue.put(Request('close', {}))
        self.queue.close()
        self.process.join()
        LOGGER.info("Process joined")

    def _panel_loop(self, parent_queue):
        LOGGER.info("Panel process created")
        self.queue_open = False
        self._panel = BCIPanel()
        self.last_update_time = time.time()
        run = True
        while run:
            try:
                request = self.queue.get(timeout=self.time_between_updates)
            except queue.Empty:
                try:
                    self._process_panel_events()
                except _tkinter.TclError:  # except when we closed the figure manually
                    pass
            else:
                run = self._process_request(request)
                if (time.time() - self.last_update_time) > self.time_between_updates:
                    self._process_panel_events()
        self._panel.close()
        parent_queue.cancel_join_thread()
        LOGGER.info("Panel process finished")

    def _process_request(self, request):
        run = request.name != 'close'
        try:
            bound_method = getattr(self._panel, request.name)
            bound_method(**request.kwargs)
        except Exception as ex:
            message = "Error processing request '%s': %s" % (request.name, request.kwargs)
            raise Exception(message) from ex
        return run

    def _process_panel_events(self):
        self._panel.process_events()
        self.last_update_time = time.time()
