# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.
import asyncio
import logging
import os
import pickle
import queue
import re
import threading
from collections import OrderedDict
from concurrent.futures import Future
from pathlib import Path
from typing import Union

import janus
import numpy as np
import socketio
from aiohttp import web
from aiohttp.web import raise_graceful_exit

from bci_panel.bci_panels import Request
from braintech.obci.core.broker.message_handler_mixin import subscribe_message_handler
from braintech.obci.core.broker.messages import BaseMessage, Field
from braintech.obci.core.broker.peer import Peer, PeerInitUrls
from braintech.obci.core.utils.parent_checking import ParentCheckingThread
from braintech.obci.core.utils.streaming_debug import SamplingRateEstimator
from .bci_panels import BCIPanelProcess, BaseBCIPanelProcess

LOGGER = logging.getLogger('bci_panel')
LOGGER.setLevel('DEBUG')
PORT = 7777

CHARTS = '/charts'


class BCIPanelMessage(BaseMessage):
    command = Field(str)
    kwargs = Field(object)

    def serialize_data(self):
        return pickle.dumps((self.command, self.kwargs))

    @classmethod
    def deserialize_data(cls, data: bytes):
        command, kwargs = pickle.loads(data)
        return {'command': command, 'kwargs': kwargs}


class WebBCIPanelMixin:
    SIGNAL_UPDATE_RATE = 0.5
    SIGNAL_CHUNK_LENGTH = 0.5
    SEPARATION_TYPE = 'other'

    def get_separation(self, mean_std):
        return 4


class RemoteWebBCIPanel(WebBCIPanelMixin, BaseBCIPanelProcess):

    def close(self):
        self._send_msg('close')

    def save_plots(self, **kwargs):
        self._send_msg('save_plots', **kwargs)

    def __init__(self, peer):
        self._peer = peer

    def _send_msg(self, command, **kwargs):
        if not self._peer.is_connected:
            LOGGER.warning("Peer %s is not connected", self._peer)
            return
        msg = BCIPanelMessage(command=command,
                              kwargs=kwargs
                              )
        self._peer.send_message(msg)

    def update_hist(self, **kwargs):
        self._send_msg('update_hist', **kwargs)

    def update_plot(self, **kwargs):
        self._send_msg('update_plot', **kwargs)

    def update_text(self, **kwargs):
        self._send_msg('update_text', **kwargs)


class WebBCIPanel(WebBCIPanelMixin, BCIPanelProcess):

    def _is_alive(self):
        return os.getppid() != 1 and self._run

    def _get_from_queue(self, parent_queue):
        while self._is_alive():
            try:
                item = parent_queue.get(timeout=0.5)
                if item.name == 'close':
                    break
                self.server.process_request(item)
            except queue.Empty:
                pass
        self._stop()

    def _stop(self):
        self._run = False
        self.server.stop()

    def _panel_loop(self, parent_queue):
        ParentCheckingThread().start()
        self._run = True
        thread = threading.Thread(target=self._get_from_queue, args=(parent_queue,), daemon=True, name="QueueThread")
        self.server = WebBCIPanelServer(CHARTS)
        thread.start()
        self.server.run("0.0.0.0", PORT)


class WebBCIPanelServer:
    def __init__(self, namespace):
        self._async_queue = None
        self._tmp_queue = []
        self._clear_erp()
        self._last_erp_msg = None
        self.namespace = namespace
        self._create_server()
        self._current_ts = 0
        self._estimator = SamplingRateEstimator("bci_panel")
        self._last_msgs = OrderedDict()

    def process_request(self, req):
        if self._async_queue:
            self._async_queue.sync_q.put(req)
        else:
            self._tmp_queue.append(req)

    def _create_queue(self, loop):
        self._async_queue = janus.Queue(loop=loop)
        for item in self._tmp_queue:
            self._async_queue.sync_q.put(item)
        self._tmp_queue = []

    def run(self, host, port):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        self._create_queue(loop)
        loop.set_exception_handler(self._handle_exception)
        loop.create_task(self._serve_to_clients())
        self._loop = loop
        web.run_app(self._app, host=host, port=port, shutdown_timeout=1.0, handle_signals=False)

    def stop(self):
        self._loop.call_soon_threadsafe(raise_graceful_exit)

    def _create_server(self):
        self._app = web.Application()
        self._sio = socketio.AsyncServer()
        self._sio.attach(self._app)
        logging.getLogger('engineio').setLevel(logging.WARNING)
        logging.getLogger('socketio').setLevel(logging.WARNING)

        self._sio.on('connect', handler=self._on_connect, namespace=self.namespace)

        self._sio.on('disconnect', handler=self._on_disconnect, namespace=self.namespace)

        self._add_routes(self._app.router)

    def _on_connect(self, sid, environ):
        LOGGER.info("connected %s", sid)
        for name, data in self._last_msgs.items():
            self._sio.emit(name, data, room=sid)

    def _on_disconnect(self, sid):
        LOGGER.info('disconnect %s', sid)

    def _add_routes(self, router):
        dist_path = Path(__file__).parent / 'dist'
        if dist_path.exists():
            router.add_static('/', Path(__file__).parent / 'dist')

        else:
            LOGGER.warning("%s does not exist. Not serving static files!", dist_path)

    def _handle_exception(self, loop, context):
        loop.default_exception_handler(context)
        logged = False
        if 'future' in context:
            try:
                context['future'].result()
            except Exception:
                LOGGER.exception("Exception in %s", context['future'])
                logged = True
        if not logged:
            LOGGER.exception("%s:%s", context.get('message'), context.get('exception'))
        self.stop()

    async def _serve_to_clients(self):
        try:
            while True:
                item = await self._async_queue.async_q.get()
                await self._process_item(item)
        except Exception:
            LOGGER.exception('_serve_to_clients_exception')
            raise

    async def _process_item(self, item):
        used_function_names = ['update_plot', 'update_text', 'update_hist']
        functions_map = {
            'features': self._send_features,
            'raw_signal': self._send_eeg,
            'signal impedance': self._send_impedances,
            'mean epochs calib': self._send_calibration_erp,
            'epochs mean': self._send_online_erp,
            'prediction hist': self._send_certainty_histogram,
            'classifier_confidence_batch': self._send_classifier_certainty,
            'AUC': self._send_auc,
            'calibration_progress': self._send_ssvep_calibration_progress,
            'Accuracy and itr': self._send_acc_itr,
            'frequency preference function': self._frequency_preference_function,
        }
        if item.name in used_function_names:
            feature = item.kwargs['feature']
            if feature in functions_map:
                function = functions_map[feature]
                await function(item)

    async def _frequency_preference_function(self, item):
        x = [float('{:0.2f}'.format(v)) for v in item.kwargs['x_data'][0]]
        y = [float(v) for v in item.kwargs['y_data'][0]]
        fpfbase = [float(v) for v in item.kwargs['y_data'][1]]
        data = {
            'x': x,
            'y': y,
            'fpfbase': fpfbase,
        }
        await self.emit('frequency_preference_function', data)

    async def _send_features(self, item):
        labels = item.kwargs['labels']
        label_for_current = 'NOW'
        assert label_for_current in labels
        index_for_current = labels.index(label_for_current)
        all_data = item.kwargs['y_data']
        y = all_data[index_for_current]
        x = [float(v) for v in item.kwargs['x_ticks'][1]]
        data = {
            'x': x,
            'y': y,
        }
        await self.emit('features', data)

    async def _send_eeg(self, item):
        channel_names = item.kwargs['labels']
        sampling_rate = item.kwargs['sampling_rate']
        channel_data = item.kwargs['y_data']
        timestamps = item.kwargs['x_data'][0]
        # self._estimator.new_samples(len(timestamps))

        msg = {
            'channel_names': ['timestamp'] + channel_names,
            'sampling_rate': sampling_rate,
            'max_amplitude': -5,
        }
        timestamps += self._current_ts
        self._current_ts += len(timestamps) / sampling_rate
        await self.emit('raw_signal_description', msg)
        data_to_send = np.vstack((timestamps, channel_data)).astype(np.float32).tobytes()
        await self.emit('raw_signal', data_to_send)

    async def _send_impedances(self, item):
        channel_names = item.kwargs['labels']
        impedance_history_per_channel = item.kwargs['y_data']
        impedances = {}
        for i, channel_name in enumerate(channel_names):
            impedance_history = impedance_history_per_channel[i]
            latest_index = -1
            latest_imp = impedance_history[latest_index]
            if not np.isfinite(latest_imp):
                latest_imp = 9999999
            impedances[channel_name] = latest_imp
        await self.emit('impedances', impedances)

    async def _send_online_erp(self, item):
        target_type = 'tgt' if item.kwargs['field_id'] == 0 else 'ntgt'
        res = re.search(r'N=(\d+)', item.kwargs['title'])
        if res is None:
            self._clear_erp()
            columns = []
        else:
            self._erp_online_data[target_type + '_count'] = int(res.group(1))
            labels = [label + ' ' + target_type for label in item.kwargs['labels']]
            self._erp_online_data[target_type + '_data'] = _make_columns(labels=labels,
                                                                         values_lists=item.kwargs['y_data'])
            columns = (self._erp_online_data['tgt_data']
                       + self._erp_online_data['ntgt_data']
                       + [['x_data'] + _make_list(item.kwargs['x_data'][0])]
                       )
        msg = {
            'targets_count': self._erp_online_data['tgt_count'],
            'nontargets_count': self._erp_online_data['ntgt_count'],
            'calibration': False,
            'columns': columns,
            'amplitude': 4,
        }
        await self._send_erp_msg(msg)

    async def _send_calibration_erp(self, item):

        x_data = _make_list(item.kwargs['x_data'][0])
        y_data = _make_list(item.kwargs['y_data'])

        labels = item.kwargs['labels']
        columns = []
        for i, data_row in enumerate(y_data):
            label = labels[i]
            columns.append([label] + data_row)
        columns.append(['x_data'] + x_data)
        res = re.search(r'TGT=(\d+), NTGT=(\d+)', item.kwargs['title'])
        msg = {
            'targets_count': int(res.group(1)),
            'nontargets_count': int(res.group(2)),
            'calibration': True,
            'columns': columns,
            'amplitude': 5,
            'channels': item.kwargs['channels']
        }
        await self._send_erp_msg(msg)

    async def _send_erp_msg(self, msg):
        self._last_erp_msg = msg
        await self.emit('erp', msg)

    def _clear_erp(self):
        self._erp_online_data = {
            'tgt_count': 0,
            'ntgt_count': 0,
            'tgt_data': [],
            'ntgt_data': [],
        }

    async def _send_certainty_histogram(self, item):
        targets_histogram = _make_list(item.kwargs['targets_histogram'])
        nontargets_histogram = _make_list(item.kwargs['nontargets_histogram'])
        bins_centers = _make_list(item.kwargs['bins_centers'])
        threshold = item.kwargs['threshold']
        values = {
            'targets_histogram': ['targets'] + targets_histogram,
            'nontargets_histogram': ['nontargets'] + nontargets_histogram,
            'bins_centers': ['bins_centers'] + bins_centers,
            'threshold': threshold,
        }
        await self.emit('certainty_histogram', values)

    async def _send_classifier_certainty(self, item):
        choices_values = item.kwargs['choices_values']
        choice_labels = {
            0: 'YES',
            1: 'NO',
        }
        columns = [
            [choice_labels[0], 0, *choices_values.get(0, [])],
            [choice_labels[1], 0, *choices_values.get(1, [])],
        ]
        values = {
            'columns': columns,
            'threshold': item.kwargs['threshold'],
        }
        await self.emit('classifier_certainty', values)

    async def _send_auc(self, item):
        auc_value = item.kwargs['value']
        values = {'value': '{:.2f}'.format(auc_value)}
        await self.emit('AUC', values)

    async def _send_ssvep_calibration_progress(self, item):
        try:
            best_freqs = item.kwargs['data']['Selected best freqs'].strip()
            await self.emit('best_freqs', {'value': best_freqs})
        except KeyError:
            pass

        try:
            mean_example_count_f = float(item.kwargs['data']['Mean example count per freq'].strip())
            mean_example_count = '{:0.2f}'.format(mean_example_count_f)
            await self.emit('mean_example_count', {'value': mean_example_count})
        except KeyError:
            pass

        try:
            mean_auc_raw = item.kwargs['data']['Mean AUC']
            if mean_auc_raw == '{}':
                mean_auc = 'None'
            else:
                mean_auc = mean_auc_raw
            auc_value = {'value': mean_auc}
            await self.emit('AUC', auc_value)
        except KeyError:
            pass

    async def _send_acc_itr(self, item):
        itr = item.kwargs['data']['ITR']
        acc = item.kwargs['data']['Accuracy']
        utility = item.kwargs['data']['Utility']
        await self.emit('ITR', {'value': itr})
        await self.emit('ACC', {'value': acc})
        await self.emit('UTILITY', {'value': utility})

    async def emit(self, name, data):
        self._last_msgs[name] = data
        await self._sio.emit(name, data, namespace=self.namespace)


def _make_list(obj):
    try:
        return obj.tolist()
    except AttributeError:
        return obj


def _make_columns(labels, values_lists):
    columns = []
    for label, data_row in zip(labels, values_lists):
        columns.append([label] + _make_list(data_row))
    return columns


class WebBCIPanelPeer(Peer):
    def __init__(self, urls: Union[str, PeerInitUrls], **kwargs):
        self.server_host = kwargs.pop('host', "0.0.0.0")
        self.server_port = kwargs.pop('port', 8123)
        self.server_namespace = kwargs.pop('namespace', CHARTS)
        super().__init__(urls, **kwargs)

        self.bci_panel = self._create_bci_panel_server()

    def _create_bci_panel_server(self):
        return WebBCIPanelServer(CHARTS)

    @subscribe_message_handler(BCIPanelMessage)
    async def handle_panel_message(self, msg: BCIPanelMessage):
        self.bci_panel.process_request(Request(msg.command, msg.kwargs))

    def run(self):
        self.bci_panel.run(self.server_host, self.server_port)
        super().run()

    def run_in_thread(self):
        result = Future()

        def _run():
            try:
                self.run()
            except Exception as ex:
                result.set_exception(ex)
                raise
            result.set_result(None)

        self._server_thread = threading.Thread(target=_run, daemon=True, name="WebBCIPanelServer")
        self._server_thread.start()
        return result
