# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.

import asyncio
import logging
import time
from queue import Empty

from braintech.obci.experiment.peer.configured_peer import ConfiguredPeer
from braintech.obci.core.broker.messages import TagMsg, BaseMessage, Field
from obci_p300.common import bci_reporter
from obci_p300.common.bci_reporter import ReporterSummaryMsg
from obci_p300.tag_sender import ExperimentTagsSender
from obci_ugm.gui.ugm import ugm_config_manager
from obci_ugm.gui.ugm import ugm_helper
from obci_ugm.utils.audio_files_player import AudioFilesPlayer
from obci_ugm.utils.sequence_provider import RandomSequentialMgr
from .. import parameters
from ..utils.ssvep_blinker_controller import SSVEPBlinkerController

LOGGER = logging.getLogger('SSVEP_calibrator')


class CalibrationStartedDebugMsg(BaseMessage):
    frequency = Field(float)


class CalibratorFrontend:
    FRAMERATE = 30
    INSTRUCTION_TIME = 2

    def __init__(self, calibrator, peer, parameters, ugm_display_config):
        self._peer = peer
        self._parameters = parameters
        self._speech_player = _get_initialized_speech_player(parameters)
        self._calibrator = calibrator
        self.default_config_ugm = (ugm_config_manager.UgmConfigManager(ugm_display_config)
                                   .config_to_message())

    async def reset(self):
        ugm_helper.send_config(self._peer, self.default_config_ugm)

    async def update_status(self, text):
        ugm_helper.send_status(self._peer, text)

    async def present_text(self, text, use_audio, default_duration):
        ugm_helper.send_status(self._peer, text)
        if use_audio and self._speech_player:
            self._speech_player.play(text)
            duration = self._speech_player.get_length_in_seconds(text)
        else:
            duration = default_duration
        await self.wait_if_running(duration)

    async def wait_if_running(self, seconds, start=None):
        start = start or time.time()
        while (time.time() - start < seconds
               and self._calibrator._running):
            frame_start_time = time.time()
            await self.sleep_until_frame_end(frame_start_time)

    async def sleep_until_frame_end(self, frame_start_time):
        time_now = time.time()
        frame_duration = time_now - frame_start_time
        time_left = 1 / self.FRAMERATE - frame_duration
        LOGGER.debug('TIME_LEFT {}'.format(time_left))
        if time_left > 0:
            await asyncio.sleep(time_left)

    async def change_field_color(self, field, color_name):
        color_codes = dict(red='#ff0000', green='#00ff00')
        ugm_field_id = await self._get_ugm_field_id(field)
        color = color_codes[color_name]
        ugm_helper.send_config_for(self._peer, ugm_field_id,
                                   'color', color)

    async def _get_ugm_field_id(self, field):
        if int(field) >= 0:
            return int(self._parameters.first_feedback_field_id) + int(field)
        else:
            return ugm_helper.STATUS_ID

    async def trial_instruction(self, field_id):
        # prepare for trial
        instruction = self._parameters.statuses[field_id]
        await self.present_text(instruction,
                                use_audio=self._parameters.instruction_type_audio,
                                default_duration=self.INSTRUCTION_TIME)

    async def display_good_bye_text(self):
        ugm_helper.send_text(self._peer, self._parameters.bye_text)


class SSVEPCalibrator:
    TRIAL_TIME = 6
    PREPARE_TIME = 0.5
    FRAMERATE = CalibratorFrontend.FRAMERATE
    PAUSE_TIME = 2
    QUESTION_TIME = 5
    FEEDBACK_TIME = 2

    def __init__(self, ssvep_blinker_controller: SSVEPBlinkerController,
                 analysis, communications_peer: ConfiguredPeer,
                 frontend_parameters: parameters.FrontendParameters,
                 reporter: bci_reporter.BCIReporter,
                 ugm_display_config='', pause_on_start=True):
        self.pause_on_start = pause_on_start
        self.analysis = analysis
        self.ssvep_blinker_controller = ssvep_blinker_controller
        self.communications_peer = communications_peer
        self.frontend = self._create_frontend(frontend_parameters, ugm_display_config)  # type: CalibratorFrontend
        self.calibration_fields = list(frontend_parameters.field_to_frequency_map.keys())
        self.side_manager = RandomSequentialMgr(0, len(self.calibration_fields))
        self.calibration_fields_counter = {i: 0 for i in self.calibration_fields}
        self._space_pressed = False
        self._f_pressed = False
        self._running = False
        self._running_finished = None
        self._parameters = frontend_parameters
        self._reporter = reporter
        self._tag_sender = ExperimentTagsSender(self.communications_peer)
        self._last_answer_timestamp = 0

    def _create_frontend(self, parameters, ugm_display_config):
        return CalibratorFrontend(self, self.communications_peer, parameters, ugm_display_config)

    async def start_calibration(self):
        if self._running is False:
            self._running_finished = asyncio.Future()
            self._running = True
            await self._calibrate()
            self._running_finished.set_result(None)

    async def start_feedback(self):
        if self._running is False:
            self._running_finished = asyncio.Future()
            self._running = True
            if self._parameters.questions:
                await self._ask_feedback_questions()
            else:
                await self._endless_feedback()
            self._running_finished.set_result(None)

    async def stop(self):
        self._running = False
        if self._running_finished is not None:
            while not self._running_finished.done():
                await asyncio.sleep(1 / self.FRAMERATE)

    async def _calibrate(self):
        await self.pause(space_ends=self.pause_on_start)
        while self._running:
            await self.frontend.reset()
            calibration_field_id = self.calibration_fields[self.side_manager.get_value()]

            LOGGER.info("Starting trial for field {} freq {}".format(calibration_field_id,
                                                                     self._parameters.field_to_frequency_map[
                                                                         calibration_field_id])
                        )
            await self.frontend.trial_instruction(calibration_field_id)
            await self.pretrial(calibration_field_id)
            await self.trial(calibration_field_id)
            if self.get_f_pressed():
                await self._endless_feedback()
            else:
                await self.pause(space_ends=self.get_space_pressed())
            LOGGER.info('Trial counter: {}'.format(self.calibration_fields_counter))
            LOGGER.info('Sample counter {}'
                        .format([(i[0], len(i[1]))
                                 for i in self.analysis.calibration_buffer.items()]))
            self.calibration_fields_counter[calibration_field_id] += 1

    async def _endless_feedback(self):
        self.analysis.reset_buffers()
        self.ssvep_blinker_controller.start_stimulation()
        while not self.get_f_pressed() and self._running:
            frame_start_time = time.time()

            prediction = self._try_to_get_answer()

            if prediction is None:
                prediction_text = self._parameters.field_to_feedback_text[None]
                await self.frontend.update_status(prediction_text)
            else:
                await self.frontend.reset()
                await self._display_detected_answer(prediction)
                self.ssvep_blinker_controller.start_stimulation()
            await self.frontend.sleep_until_frame_end(frame_start_time)
        self.ssvep_blinker_controller.fullbright()

    def _try_to_get_answer(self):
        try:
            answer_freq = self.analysis.prediction_queue.get(block=False)

            known_answer = answer_freq is not None
            time_hold_done = (time.monotonic() - self._last_answer_timestamp) > self.FEEDBACK_TIME
            if time_hold_done and known_answer:
                self._last_answer_timestamp = time.monotonic()
                self.analysis.reset_buffers()
            else:
                answer_freq = None

        except Empty:
            answer_freq = None
        return answer_freq

    async def _ask_feedback_questions(self):  # noqa: C901
        async def _run():
            questions = _pick_random_questions()
            await _ask_questions(questions)

        def _pick_random_questions():
            questions = self._parameters.questions
            sequence_provider = RandomSequentialMgr(0, len(questions))
            for _ in range(self._parameters.trials_count):
                index = sequence_provider.get_value()
                yield questions[index]

        async def _ask_questions(questions):
            for question in questions:
                if _should_run():
                    await self.pause(space_ends=True)
                    await _ask(question)

        async def _ask(question):
            await _initialize()
            answer_frequency = None
            answer_field = None
            await self._tag_sender.send_text_question_start()
            await _show_question(question)
            self.ssvep_blinker_controller.start_stimulation()
            self.analysis.reset_buffers()
            while answer_frequency is None and _should_run():
                frame_start_time = time.time()
                answer_frequency = self._try_to_get_answer()
                if answer_frequency is not None:
                    answer_field = self._parameters.frequency_to_field_map[answer_frequency]
                await self.frontend.sleep_until_frame_end(frame_start_time)
            self._reporter.trial_finished(time.time(), answer=answer_field)
            actual_answer = _get_answer_field(answer_frequency)
            correct_answer = _get_correct_field(question)
            is_correct = (actual_answer == correct_answer)
            self._reporter.set_answer(actual_answer, correct_answer)
            await self._send_summary(is_correct)
            await self._display_detected_answer(answer_frequency, is_correct=is_correct)

        async def _initialize():
            await self.frontend.reset()
            self.analysis.reset_buffers()

        async def _show_question(question):
            self._reporter.command_displayed(time.time(), question)
            await self.frontend.present_text(question, use_audio=self._parameters.question_type_audio,
                                             default_duration=self.QUESTION_TIME)
            await asyncio.sleep(self._parameters.thinking_pause_after_question_length)
            self._reporter.trial_started(time.time())

        def _should_run():
            return not self.get_f_pressed() and self._running

        def _get_answer_field(answer_frequency):
            answer_field = self._parameters.frequency_to_field_map.get(answer_frequency)
            return int(answer_field) if answer_field else None

        def _get_correct_field(question):
            return int(self._parameters.question_to_answer_field_id[question])

        def _is_correct(question, answer_frequency):
            answer_field = self._parameters.frequency_to_field_map.get(answer_frequency)
            if answer_field:
                correct_field = self._parameters.question_to_answer_field_id[question]
                return int(answer_field) == int(correct_field)
            else:
                return False

        await _run()
        await self.frontend.display_good_bye_text()

    async def _send_summary(self, is_correct):
        decision = self._reporter.fetch_last_decision_tag()
        await self._tag_sender.send_decision(decision, is_correct)
        reporter_summary = self._reporter.get_summary()
        summary_msg = ReporterSummaryMsg(accuracy=reporter_summary['accuracy'],
                                         itr_start_blinking_to_end=reporter_summary['itr_start_to_end'],
                                         utility_start_blinking_to_end=reporter_summary['utility_start_to_end'],
                                         command_count=reporter_summary['command_count'],
                                         correct_count=reporter_summary['correct_count'],
                                         mean_time=reporter_summary['mean_time_start_to_end'],
                                         null_proba=reporter_summary['null_proba'],
                                         )
        self.communications_peer.send_message(summary_msg)

    async def _display_detected_answer(self, answer_frequency, is_correct=True):
        self.ssvep_blinker_controller.fullbright()
        field = self._parameters.frequency_to_field_map[answer_frequency] if answer_frequency is not None else None
        prediction_text = self._parameters.field_to_feedback_text[field]
        if field is not None:
            try:
                await self.communications_peer.send_decision(field)
            except AttributeError:
                pass
            await self._tag_sender.send_visual_feedback_start()
            color_name = 'green'
            await self.frontend.change_field_color(field=field, color_name=color_name)
            await self.frontend.present_text(prediction_text,
                                             use_audio=self._parameters.feedback_type_audio,
                                             default_duration=self.FEEDBACK_TIME)
        await self.frontend.reset()
        await self.frontend.present_text(prediction_text, use_audio=False, default_duration=0)

    def handle_space_pressed(self):
        self._space_pressed = True

    def handle_f_pressed(self):
        self._f_pressed = True

    def get_space_pressed(self):
        if self._space_pressed:
            self._space_pressed = False
            return True
        return False

    def get_f_pressed(self):
        if self._f_pressed:
            self._f_pressed = False
            return True
        return False

    async def await_for_pause_end(self):
        t0 = time.time()
        while (time.time() - t0 < self.PAUSE_TIME
               and self._running):
            await asyncio.sleep(1 / self.FRAMERATE)

    async def await_for_space(self):
        while True and self._running:
            if self.get_space_pressed():
                break
            await asyncio.sleep(1 / self.FRAMERATE)

    async def pause(self, space_ends=False):
        LOGGER.info('Pause')
        self.set_break_state()
        if space_ends:
            await self.await_for_space()
        else:
            await self.await_for_pause_end()
        await asyncio.sleep(1)

    async def send_trial_start(self, trial_start_time, calibration_field):
        field = str(calibration_field)
        frequency = self._parameters.field_to_frequency_map[field]
        tag = TagMsg(
            start_timestamp=trial_start_time,
            end_timestamp=trial_start_time + self.TRIAL_TIME,
            name='ssvep_trial_start',
            channels='-1',
            desc={'calibration_field': field, 'calibration_type': str(frequency)},
        )
        self.communications_peer.send_message(CalibrationStartedDebugMsg(frequency=float(frequency)))
        self.communications_peer.send_message(tag)

    async def send_trial_end(self, trial_end_time):
        tag = TagMsg(
            start_timestamp=trial_end_time,
            end_timestamp=trial_end_time,
            name='ssvep_trial_end',
            channels='-1',
            desc={},
        )
        self.communications_peer.send_message(tag)
        self.communications_peer.send_message(CalibrationStartedDebugMsg(frequency=0.0))

    async def trial(self, field_id):

        trial_start_time = time.time()
        await self.send_trial_start(trial_start_time, field_id)
        await self.frontend.wait_if_running(self.TRIAL_TIME, start=trial_start_time)
        trial_stop_time = time.time()
        await self.send_trial_end(trial_stop_time)
        self.ssvep_blinker_controller.fullbright()

    async def pretrial(self, calibration_field_id):

        trial_start_time = time.time()
        self.start_trial_stimuli()
        while (time.time() - trial_start_time < self.PREPARE_TIME
               and self._running):
            frame_start_time = time.time()
            await self.frontend.sleep_until_frame_end(frame_start_time)

    def set_break_state(self):
        ugm_helper.send_text(self.communications_peer, self._parameters.pause_text)

    def start_trial_stimuli(self):
        self.ssvep_blinker_controller.start_stimulation()

    def close(self):
        self._running = False


def _get_initialized_speech_player(frontend_parameters: parameters.FrontendParameters):
    texts = frontend_parameters.get_speech_texts()
    if texts:
        try:
            from bci_study.utils import speech_generator
        except ImportError:
            LOGGER.warning("Speech Generator not available!")
        else:
            speech_files = speech_generator.get_speech_file_paths(texts)
            return AudioFilesPlayer(speech_files)
