# -*- coding: utf-8 -*-
# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.
import operator

import numpy as np
from scipy.stats import norm

import obci_p300.common.messages
from braintech.obci.core.utils.asyncio import wait_for_condition
from obci_p300.common import bci_reporter
from .ssvep_analysis_bci_panel_visualisation import SSVEPAnalysisVisualisation
from .ssvep_calibrator import SSVEPCalibrator
from .. import parameters
from ..ssvep_peer import SSVEPAnalysisWithCommunicationAndVisualization
from ..utils.ssvep_blinker_controller import SSVEPBlinkerController
from ..utils.visualization import SsvepVisualizationDispatcher


class FieldFreqsProvider:
    MAX_FREQUENCY_RANDOMISATION = 0.1
    PREFERED_FREQUENCY_CENTER = 35
    PREFERED_FREQUENCY_STD = 60

    def __init__(self, min_freq, max_freq, step, n, frequency_randomisation=True, harmonic_avoidance=True):
        self._step = step
        # fighting with stimulus synchronisation - creates visible pulses
        self._frequency_randomisation = frequency_randomisation

        self._available = np.arange(min_freq, max_freq, step, dtype=np.float64)
        self._tried_frequencies_mask = np.zeros_like(self._available, dtype=bool)
        self._candidates_of_good_frequencies_mask = np.zeros_like(self._available, dtype=bool)
        self._frequency_preference_function = self.get_base_frequency_preference_function()

        if harmonic_avoidance:
            self._harmonic_avoidance()
        if self._frequency_randomisation:
            self._available = self._randomise(self._available)
        self.n = n
        self._id = len(self._available) - n

    def get_base_frequency_preference_function(self):
        return norm.pdf(self._available, loc=self.PREFERED_FREQUENCY_CENTER,
                        scale=self.PREFERED_FREQUENCY_STD)

    def update_frequency_preference_from_auc(self, aucs, crit_auc=0.8):
        """:param aucs: dictionary: id: freq (float), value AUC"""

        for key, value in aucs.items():
            freq_id = self.get_available_freqs().index(key)
            self._frequency_preference_function[freq_id] *= (1 - crit_auc) + value
            self._tried_frequencies_mask[freq_id] = True
            self._candidates_of_good_frequencies_mask[freq_id] = (value >= crit_auc)

    def _harmonic_avoidance(self):
        for f0 in self._available:
            for i, f_potential_harmonic in enumerate(self._available):
                if f0 != f_potential_harmonic:
                    divisor = f_potential_harmonic / f0
                    full_divisor = int(divisor)
                    if np.abs(f0 * full_divisor - f_potential_harmonic) < self._step * 0.25:
                        self._available[i:] += self._step * 0.5

    def get_available_freqs(self):
        return list(self._available.astype(float))

    def _randomise(self, freqs):
        randomisation_degree = self.MAX_FREQUENCY_RANDOMISATION
        step_tenth = self._step / 10
        if step_tenth < self.MAX_FREQUENCY_RANDOMISATION:
            randomisation_degree = step_tenth
        freqs = freqs + np.random.uniform(-randomisation_degree, randomisation_degree, len(freqs))
        return freqs

    def get(self):
        not_tried_freq = np.logical_not(self._tried_frequencies_mask)
        not_enough_not_tried_freqs = np.sum(not_tried_freq) < self.n
        enough_good_freqs = np.sum(self._candidates_of_good_frequencies_mask) >= self.n
        if not_enough_not_tried_freqs or enough_good_freqs:
            preference_mask = np.argsort(self._frequency_preference_function)[::-1]
            sorted_available = self._available[preference_mask].astype(float)
            ret = sorted(sorted_available[:self.n])
        else:
            preference_mask = np.argsort(self._frequency_preference_function[not_tried_freq])[::-1]
            sorted_available = self._available[not_tried_freq][preference_mask].astype(float)
            ret = sorted(sorted_available[:self.n])

        return list(ret)

    def get_freq_preference_function(self):
        return np.copy(self._frequency_preference_function)


class SSVEP_Metacalibrator:
    MINIMUM_EXAMPLE_COUNT = 9

    def __init__(self, field_freqs_search,
                 field_freqs_search_min_distance,
                 comms_peer, ugm_display_config, fields_ids_light_indexes,
                 sample_rate, initial_frontend_parameters: parameters.FrontendParameters,
                 reporter: bci_reporter.BCIReporter,
                 bci_panel_class,
                 window=2.0, feedback_every=0.25,
                 learning_every=2.0, channels_for_classification=None, channels_for_montage=None,
                 channel_names=None, critical_auc=0.9, signal_goodness_reporters=None, use_harmonics=True,
                 field_setup_func=None, start_with_pause=True,
                 power_line_freq=50, exclude_from_montage=()):
        self.field_setup_func = field_setup_func
        self.use_harmonics = use_harmonics
        self.signal_goodness_reporters = signal_goodness_reporters
        self.crit_auc = critical_auc
        self.sample_rate = sample_rate
        self.initial_frontend_parameters = initial_frontend_parameters
        self.field_freqs = initial_frontend_parameters.field_freqs
        self.field_freqs_search = field_freqs_search
        self.field_freqs_search_min_distance = field_freqs_search_min_distance
        self.comms_peer = comms_peer
        self.ugm_display_config = ugm_display_config
        self.window = window
        self.feedback_every = feedback_every
        self.learning_every = learning_every
        self.channels_for_classification = channels_for_classification
        self.channels_for_montage = channels_for_montage
        self.channel_names = channel_names
        self.fields_ids_light_indexes = fields_ids_light_indexes
        self.bci_panel_class = bci_panel_class
        self.start_with_pause = start_with_pause

        self.field_freqs_provider = FieldFreqsProvider(self.field_freqs_search[0], self.field_freqs_search[1],
                                                       self.field_freqs_search_min_distance, len(self.field_freqs))

        self.power_line_freq = power_line_freq
        self.exclude_from_montage = exclude_from_montage

        self.statuses = {}

        # to keep best ROC
        self.rocs_per_calibrations = {}  # per frequencies sets
        self.rocs_per_frequencies = {}  # per frequencies
        self._latest_calibration_result = None

        self._running = False

        self.bci_panel = self._create_bci_panel(bci_panel_class)
        self._reporter = reporter

        # for pause on first calib
        self._first_calib = True
        self._current_analysis = None
        self._current_calibrator = None
        self._current_blinker_controller = None

    def _create_bci_panel(self, bci_panel_class):
        return SsvepVisualizationDispatcher(
            channel_names=self.channel_names,
            channels_after_montage=self.channels_for_classification,
            sampling_rate=self.sample_rate,
            stimulation_freqs=self.field_freqs_provider.get_available_freqs(),
            bci_panel_class=bci_panel_class,
        )

    async def start_calibration(self):
        self._running = True
        await self._find_best_frequencies()
        if self._running:
            await self._infinite_calibration_on_best_freqs()

    async def _find_best_frequencies(self):
        while self._running and (not self.are_best_frequencies_found()):
            self._start_new_calib()
            self.comms_peer.create_task(self._current_calibrator.start_calibration())

            await wait_for_condition(self._analysis_calibration_done, float('inf'), 0.1, 'Analysis calibration')
            await wait_for_condition(self._enough_data_in_calibration, float('inf'), 0.1, 'Analysis calibration data')

            self._latest_calibration_result = self._current_analysis.calibration_result

            await self._current_calibrator.stop()

            if self._running:
                self.rocs_per_calibrations[tuple(self.field_freqs)] = self._latest_calibration_result.get_minimum_roc()
                for f in self.field_freqs:
                    roc = self._latest_calibration_result.roc_for_freq(f)
                    self.rocs_per_frequencies[f] = roc
                self.field_freqs_provider.update_frequency_preference_from_auc(self.rocs_per_frequencies,
                                                                               self.crit_auc)
            else:
                break

    def are_best_frequencies_found(self):
        return any(auc > self.crit_auc
                   for auc in self.rocs_per_calibrations.values())

    def _start_new_calib(self):
        field_freqs = self.field_freqs_provider.get()

        self.bci_panel.handle_frequency_preference_function(
            self.field_freqs_provider.get_available_freqs(),
            self.field_freqs_provider.get_freq_preference_function())

        self.comms_peer._logger.info("Starting new calibration for freqs: {}".format(field_freqs))
        self._create_calibration_components(field_freqs, self._first_calib)

        if self._first_calib is True:
            self._first_calib = False

    def _analysis_calibration_done(self):
        return (self._running
                and self._current_analysis is not None
                and self._current_analysis.calibration_result is not None)

    def _enough_data_in_calibration(self):
        return (not self._running
                or self._current_analysis.calibration_result.get_minimum_example_count() > self.MINIMUM_EXAMPLE_COUNT)

    async def _infinite_calibration_on_best_freqs(self):
        try:
            best_freqs = sorted(self.rocs_per_calibrations.items(), key=operator.itemgetter(1), reverse=True)[0][0]
        except IndexError:
            self.comms_peer._logger.exception(
                "No best frequencies something is wrong!!!!. Choosing random frequencies for now")
            best_freqs = [39.0, 41.0, 43.0]

        self._handle_found_best_freqs(best_freqs)
        if set(best_freqs) != set(self._current_analysis.stimulation_freqs):
            self._create_calibration_components(best_freqs, self._first_calib)
        self._current_calibrator.pause_on_start = False
        await self._current_calibrator.start_calibration()

    def _handle_found_best_freqs(self, best_freqs):
        self.comms_peer._logger.info("Found best frequencies. Starting "
                                     "infinite calibration on freqs {}.".format(best_freqs)
                                     )
        self.comms_peer.send_message(obci_p300.common.messages.InitialCalibrationFinished())
        self._current_analysis.bci_panel.handle_best_frequencies_selected(best_freqs)
        self._current_analysis.bci_panel.update_calibration_progress()

    def _create_calibration_components(self, stim_freqs, pause_on_start):
        self.field_freqs = stim_freqs

        self._current_analysis = SSVEPAnalysisWithCommunicationAndVisualization(
            self.comms_peer,
            stimulation_frequencies=self.field_freqs,
            sample_rate=self.sample_rate,
            window=self.window, feedback_every=self.feedback_every,
            learning_every=self.learning_every,
            channels_for_classification=self.channels_for_classification,
            channels_for_montage=self.channels_for_montage,
            channel_names=self.channel_names,
            bci_panel=self.bci_panel,
            signal_goodness_reporters=self.signal_goodness_reporters,
            use_harmonics=self.use_harmonics,
            bci_panel_class=self.bci_panel_class,
            power_line_freq=self.power_line_freq,
            exclude_from_montage=self.exclude_from_montage,
        )

        self.bci_panel.reset_feature_means()

        if self._current_blinker_controller is None:
            if self.field_setup_func is not None:
                status_bar_fields = [40, ]
            else:
                status_bar_fields = list(range(10))

            self._current_blinker_controller = SSVEPBlinkerController(
                stim_freqs=self.field_freqs,
                status_bar_light_indexes=status_bar_fields,
                fields_ids_light_indexes=self.fields_ids_light_indexes,
                field_setup_func=self.field_setup_func,
            )
        else:
            self._current_blinker_controller.update_stim_freqs(self.field_freqs)

        self._current_calibrator = SSVEPCalibrator(
            self._current_blinker_controller,
            self._current_analysis,
            self.comms_peer,
            frontend_parameters=self.current_frontend_parameters,
            reporter=self._reporter,
            ugm_display_config=self.ugm_display_config,
            pause_on_start=pause_on_start and self.start_with_pause
        )

    @property
    def current_frontend_parameters(self):
        initial = self.initial_frontend_parameters
        current = initial.copy()
        current.field_freqs = self.field_freqs
        return current

    async def start_feedback(self):
        await self._current_calibrator.start_feedback()

    def handle_space_pressed(self):
        if self._current_calibrator is not None:
            self._current_calibrator.handle_space_pressed()

    def handle_f_pressed(self):
        if self._current_calibrator is not None:
            self._current_calibrator.handle_f_pressed()

    def handle_sample_packet(self, packet):
        if self._current_analysis is not None:
            self._current_analysis.handle_sample_packet(packet)

    def handle_tag_msg(self, tagmsg):
        if self._current_analysis is not None:
            self._current_analysis.handle_tag_msg(tagmsg)

    def load_wisdom(self, path):
        # to get calibrations freqs from wisdom
        temp_analysis = SSVEPAnalysisVisualisation(
            stimulation_frequencies=self.field_freqs,
            sample_rate=self.sample_rate,
            window=self.window, feedback_every=self.feedback_every,
            learning_every=self.learning_every,
            channels_for_classification=self.channels_for_classification,
            channels_for_montage=self.channels_for_montage,
            channel_names=self.channel_names,
            bci_panel=self.bci_panel,
        )
        temp_analysis.load_wisdom(path)
        # proper load
        wisdom_freqs = temp_analysis.calibration_result.get_available_freqs()
        self._create_calibration_components(wisdom_freqs, True)
        self._current_analysis.load_wisdom(path)

    def save(self, path):
        if self._current_analysis:
            self._current_analysis.save(path)

    async def stop(self):
        self._running = False
        if self._current_calibrator is not None:
            await self._current_calibrator.stop()

    def close(self):
        self.bci_panel.close()
        if self._current_blinker_controller is not None:
            self._current_blinker_controller.close()
