# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.

import json

from bci_panel import bci_panels
from bci_panel import web_bci_panel
from braintech.obci.experiment.peer.configured_peer import ConfiguredPeer
from braintech.obci.core.broker.message_handler_mixin import subscribe_message_handler
from braintech.obci.core.broker.messages import BaseMessage
from braintech.obci.core.broker.messages import Field
from braintech.obci.core.broker.messages import SignalMessage
from braintech.obci.core.broker.messages import TagMsg
from braintech.obci.core.broker.peer import Peer
from braintech.obci.core.utils.properties import bool_property
from braintech.obci.core.utils.properties import join_strings
from braintech.obci.core.utils.properties import param_property
from braintech.obci.core.utils.properties import raw_param_property
from braintech.obci.core.utils.properties import split_floats
from braintech.obci.core.utils.properties import split_string
from obci_p300.common import bci_reporter
from obci_p300.utils import strip_split_string
from obci_ugm.gui.ugm import ugm_config_manager
from obci_ugm.gui.ugm import ugm_helper
from obci_ugm.gui.ugm.ugm_keys import UgmKeys
from obci_ugm.utils.messages import UgmKeyEventMsg
from . import parameters
from .logic.ssvep_analysis_bci_panel_visualisation import SSVEPAnalysisVisualisation
from .logic.ssvep_calibrator import SSVEPCalibrator
from .utils.ssvep_blinker_controller import SSVEPBlinkerController

__all__ = ('SSVEPPeer',)


class InvalidConfiguration(Exception):
    pass


class SSVEPCalibrationResultMessage(BaseMessage):
    roc_auc = Field(float)
    minimal_samples_count = Field(int)


class SSVEPAnalysisWithCommunicationAndVisualization(SSVEPAnalysisVisualisation):
    def __init__(self, peer, *args, **kwargs):
        self.comms_peer = peer
        super().__init__(*args, **kwargs)

    def _handle_calibration_done(self):
        super()._handle_calibration_done()
        msg = SSVEPCalibrationResultMessage(
            roc_auc=self.calibration_result.roc_auc,
            minimal_samples_count=self.calibration_result.get_minimum_example_count(),
        )
        self.comms_peer.send_message(msg)


def int_list_list_deserialize(s):
    def _run():
        for serialized_inner in s.split(';'):
            yield [int(e) for e in serialized_inner.split(',')]

    return list(_run())


def int_list_list_serialize(ll):
    s = []
    for i in ll:
        s.append(','.join([str(k) for k in i]))
    return ';'.join(s)


class BaseSSVEPPeer(Peer):
    thinking_pause_after_question_length = param_property('thinking_pause_after_question_length', float)
    pause_text = 'PAUZA'
    status_bar_light_indexes = list(range(10))

    def __init__(self, *args, config=None, amplifier_config=None, **kwargs):
        if config and amplifier_config:
            self._set_static_parameters(config, amplifier_config)
        super().__init__(*args, **kwargs)
        self.analysis = None
        self.calibrator = None
        self.reporter = None

    def _set_static_parameters(self, config, amplifier_config):
        def _deserialize_list(items: str, deserializer):
            if items:
                return [deserializer(i) for i in items.split(';')]
            else:
                return []

        self.use_harmonics = bool(int(config['use_harmonics']))
        self.first_feedback_field_id = int(config['first_feedback_field_id'])
        self.use_artifact_rejection_based_on_signal = bool(int(config['use_artifact_rejection_based_on_signal']))
        self.use_artifact_rejection_based_on_camera = bool(int(config['use_artifact_rejection_based_on_camera']))
        self.use_camera_control_state_detector = bool(int(config['use_camera_control_state_detector']))
        self.channel_names = _deserialize_list(amplifier_config['channel_names'], str)
        self.channel_gains = _deserialize_list(amplifier_config['channel_gains'], str)
        self.channel_offsets = _deserialize_list(amplifier_config['channel_offsets'], float)

        self.calibration_mode = bool(int(config['calibration']))
        self.channels_for_montage = _deserialize_list(config['channels_for_montage'], str)
        self.channels_for_classification = _deserialize_list(config['channels_for_classification'], str)

        self.wisdom_path = config['wisdom_path']
        self.summary_dump_path = config['summary_dump_path']
        self.trials_dump_path = config['trials_dump_path']
        self.sampling_rate = int(amplifier_config['sampling_rate'])
        self.ugm_config = config['ugm_config']
        self.plots_image_path = config['plots_image_path']
        self.field_freqs = _deserialize_list(config['field_freqs'], float)
        self.status_text = config['status_text']
        self.field_statuses = _deserialize_list(config['field_statuses'], str)
        self.fields_ids_light_indexes = int_list_list_deserialize(','.join(config['fields_ids_light_indexes']))
        self.trials_count = int(config['trials_count'])
        self.questions = _deserialize_list(config['questions'], str)
        raw_correct_answers = config['questions_correct_answers_active_field_ids']
        self.questions_correct_answers_active_field_ids = [int(field_id)
                                                           for field_id in raw_correct_answers.split(';')
                                                           if field_id]
        self.instruction_type_audio = bool(int(config['instruction_type_audio']))
        self.question_type_audio = bool(int(config['question_type_audio']))
        self.feedback_type_audio = bool(int(config['feedback_type_audio']))
        self.feedback_texts = _deserialize_list(config['feedback_texts'], str)
        self.bye_text = config["bye_text"]
        self.use_web_bci_panel = bool(int(config['use_web_bci_panel']))
        self.use_windowed_bci_panel = bool(int(config['use_windowed_bci_panel']))

    async def _connections_established(self):
        await super()._connections_established()
        self._verify_configuration()
        self.ssvep_blinker_controller = self.create_blinker_controller()
        self._setup_analysis()

    def _setup_analysis(self):
        self.analysis = self.create_analysis()
        if not self.calibration_mode:
            if self.analysis is not None:
                self.analysis.load_wisdom(self.wisdom_path)
            self.reporter = self.create_bci_reporter()
        if self.calibrator:
            self.calibrator.analysis = self.analysis

    def _verify_configuration(self):
        if len(self.questions) != len(self.questions_correct_answers_active_field_ids):
            raise InvalidConfiguration("Number of questions doesn't match correct answers.")

    def _get_signal_goodness_detectors(self):
        return []

    def create_analysis(self):
        return SSVEPAnalysisWithCommunicationAndVisualization(
            stimulation_frequencies=self.field_freqs,
            sample_rate=self.sampling_rate,
            window=1.0,
            feedback_every=0.5,
            learning_every=1.0,
            channels_for_classification=self.channels_for_classification,
            channels_for_montage=self.channels_for_montage,
            peer=self,
            channel_names=self.channel_names,
            signal_goodness_reporters=self._get_signal_goodness_detectors(),
            bci_panel_class=self.bci_panel_class,
        )

    @property
    def bci_panel_class(self):
        assert not (self.use_web_bci_panel and self.use_windowed_bci_panel)
        if self.use_web_bci_panel:
            return web_bci_panel.WebBCIPanel
        elif self.use_windowed_bci_panel:
            return bci_panels.BCIPanelProcess
        else:
            return bci_panels.DummyBCIPanelProcess

    def create_blinker_controller(self):
        return SSVEPBlinkerController(stim_freqs=self.field_freqs,
                                      status_bar_light_indexes=self.status_bar_light_indexes,
                                      fields_ids_light_indexes=self.fields_ids_light_indexes)

    def create_bci_reporter(self):
        return bci_reporter.BCIReporter(number_of_choices=len(self.field_freqs),
                                        logger=self._logger)

    async def _start(self):
        self.ssvep_ugm_config = (ugm_config_manager.UgmConfigManager(self.ugm_config)
                                 .config_to_message())
        await super()._start()
        self.create_task(self.logic_process())

    async def _stop(self):
        await self._soft_stop()
        if self.reporter:
            if self.summary_dump_path:
                _save_as_json(self.reporter.get_summary(), self.summary_dump_path)
            if self.trials_dump_path:
                _save_as_json(self.reporter.get_trials(), self.trials_dump_path)
        await super()._stop()

    async def _soft_stop(self):
        if self.calibrator:
            await self.calibrator.stop()

    async def logic_process(self):
        ugm_helper.send_config(self, self.ssvep_ugm_config)
        self.calibrator = self._create_calibrator()
        await self._soft_start()

    async def _soft_start(self):
        if self.calibration_mode:
            await self.calibrator.start_calibration()
        else:
            await self.calibrator.start_feedback()

    def _create_calibrator(self):
        return SSVEPCalibrator(
            self.ssvep_blinker_controller,
            self.analysis,
            self,
            frontend_parameters=self.frontend_parameters,
            reporter=self.reporter,
            ugm_display_config=self.ugm_config,
        )

    @property
    def frontend_parameters(self) -> parameters.FrontendParameters:
        return parameters.FrontendParameters(
            status_text=self.status_text,
            field_statuses=self.field_statuses,
            field_freqs=self.field_freqs,
            pause_text=self.pause_text,
            trials_count=self.trials_count,
            questions=self.questions,
            questions_correct_answers_active_field_ids=self.questions_correct_answers_active_field_ids,
            instruction_type_audio=self.instruction_type_audio,
            question_type_audio=self.question_type_audio,
            feedback_type_audio=self.feedback_type_audio,
            feedback_texts=self.feedback_texts,
            bye_text=self.bye_text,
            first_feedback_field_id=self.first_feedback_field_id,
            thinking_pause_after_question_length=self.thinking_pause_after_question_length
        )

    @subscribe_message_handler(SignalMessage)
    def handle_signal(self, msg):
        if self.analysis is not None:
            self.analysis.handle_sample_packet(msg.data)

    @subscribe_message_handler(TagMsg)
    def handle_tag(self, msg):
        if self.analysis is not None:
            self.analysis.handle_tag_msg(msg)

    @subscribe_message_handler(UgmKeyEventMsg)
    def handle_ugm_keys(self, msg: UgmKeyEventMsg):
        if self.is_running and self.calibrator is not None:
            if msg.getUgmKey() == UgmKeys.Key_Space:
                self.calibrator.handle_space_pressed()
            elif msg.getUgmKey() == UgmKeys.Key_F:
                self.calibrator.handle_f_pressed()

    async def _shutting_down(self):
        if self.analysis is not None:
            await self._loop.run_in_executor(executor=None, func=self._stop_and_save_analysis)
        if self.calibrator is not None:
            self.calibrator.close()
        await super()._shutting_down()

    def _stop_and_save_analysis(self):
        if hasattr(self.analysis, 'bci_panel'):
            self.analysis.bci_panel.save_plots(self.plots_image_path)
        self.analysis.save(self.wisdom_path)
        self.analysis.close()


class SSVEPPeer(ConfiguredPeer, BaseSSVEPPeer):
    use_harmonics = param_property('use_harmonics',
                                   deserializer=lambda x: bool(int(x)),
                                   serializer=str)
    first_feedback_field_id = param_property('first_feedback_field_id',
                                             deserializer=int,
                                             serializer=str)
    channel_names = param_property('channel_names', deserializer=split_string)
    channel_gains = param_property('channel_gains', deserializer=split_floats)
    channel_offsets = param_property('channel_offsets', deserializer=split_floats)

    calibration_mode = bool_property('calibration')
    channels_for_montage = raw_param_property('channels_for_montage',
                                              deserializer=split_string)
    channels_for_classification = param_property('channels_for_classification',
                                                 deserializer=split_string,
                                                 serializer=join_strings)
    wisdom_path = param_property('wisdom_path')
    summary_dump_path = param_property('summary_dump_path', str)
    trials_dump_path = param_property('trials_dump_path', str)
    sampling_rate = param_property('sampling_rate', deserializer=float, serializer=str)
    ugm_config = param_property('ugm_config')
    plots_image_path = param_property('plots_image_path')
    field_freqs = param_property('field_freqs', deserializer=split_floats,
                                 serializer=join_strings)
    status_text = param_property('status_text')
    field_statuses = param_property('field_statuses', deserializer=split_string,
                                    serializer=join_strings)
    fields_ids_light_indexes = param_property('fields_ids_light_indexes',
                                              deserializer=int_list_list_deserialize,
                                              serializer=int_list_list_serialize)
    trials_count = param_property('trials_count', int)
    questions = raw_param_property('questions', deserializer=strip_split_string,
                                   serializer=join_strings)
    instruction_type_audio = bool_property('instruction_type_audio')
    question_type_audio = bool_property('question_type_audio')
    feedback_type_audio = bool_property('feedback_type_audio')
    feedback_texts = param_property('feedback_texts', deserializer=split_string,
                                    serializer=join_strings)
    questions_correct_answers_active_field_ids = raw_param_property(
        'questions_correct_answers_active_field_ids',
        deserializer=lambda s: [int(field_id) for field_id in s.split(';')
                                if field_id],
        serializer=lambda d: ';'.join([str(i) for i in d]),
    )
    bye_text = raw_param_property("bye_text")
    use_web_bci_panel = bool_property('use_web_bci_panel')
    use_windowed_bci_panel = bool_property('use_windowed_bci_panel')


def _save_as_json(data, pathname):
    with open(pathname, 'w') as json_file:
        json.dump(data, json_file, sort_keys=True, indent=4)
