import logging
import random
import threading
import time

from obci_ugm.gui.ugm.blinking import ugm_blinking_id_manager
from obci_ugm.gui.ugm.blinking import ugm_blinking_time_manager

logger = logging.getLogger(__name__)


class P300Blinker:
    HALFBLINKID = -1

    def __init__(self, blinker_config, messaging_peer):
        self.connection = messaging_peer

        self.time_mgr = ugm_blinking_time_manager.UgmBlinkingTimeManager()
        self.id_mgr = ugm_blinking_id_manager.UgmBlinkingIdManager()
        self.mgrs = [self.time_mgr, self.id_mgr]
        self._set_managers_configs(self.mgrs, blinker_config)

        self._blink_duration = blinker_config['blink_duration']
        assert self._blink_duration > 0
        self._min_max_halfblinks = blinker_config['min_max_hallfblinks_between_targets']
        self._halfblink_counter = self._min_max_halfblinks[0]
        self._send_halfblinks = blinker_config['send_halfblinks']

        self._blinking_thread = None  # type: threading.Thread
        self._exit_flag = threading.Event()

    def _set_managers_configs(self, managers, general_config):
        class Config(dict):
            def get_param(self, key):
                return self.get(key)

        configs = Config(
            blink_max_break=general_config['blink_max_break'],
            blink_min_break=general_config['blink_min_break'],
            blink_id_type=general_config['blink_id_type'],
            blink_id_count=general_config['blink_id_count'],
            blink_repeat_limit=general_config['blink_repeat_limit'],
            blink_count_type=general_config['blink_count_type'],
            blink_count_min=general_config['blink_count_min'],
            blink_count_max=general_config['blink_count_max'],
        )
        for m in managers:
            m.set_configs(configs)

    def start_blinking(self):
        if self._blinking_thread is None:
            self._blinking_thread = threading.Thread(target=self._run_blinking_loop)
            self._blinking_thread.start()
        else:
            logger.warning("Can't start blinking: blinking thread is already running.")

    def _run_blinking_loop(self):
        start_time = time.time()
        while not self._exit_flag.is_set():
            blink_id = self._schedule_blink(start_time)
            self._blink(blink_id)
            start_time = self._unblink()
        self._exit_flag.clear()
        self._reset_managers()

    def _schedule_blink(self, start_time):
        """Schedule the next blink with possibility of 'half blink'."""
        # do halfblinks sometimes
        if self._halfblink_counter > 0:
            blink_id = self.HALFBLINKID
            self._halfblink_counter -= 1
        else:
            blink_id = self.id_mgr.get_id()
            self._halfblink_counter = random.randint(*self._min_max_halfblinks)

        desired_delay = self.time_mgr.get_time()
        delay_still = desired_delay - (time.time() - start_time)
        delay_still = max(0, delay_still)
        self._exit_flag.wait(timeout=delay_still)
        return blink_id

    def _blink(self, blink_id):
        start_time = time.time()
        update_time = self._set_blinked_in_gui(blink_id)
        is_halfblink = blink_id == self.HALFBLINKID
        if (is_halfblink and self._send_halfblinks) or not is_halfblink:
            self.connection.send_blink(blink_id, update_time)
        desired_time = start_time + self._blink_duration
        timeout_seconds = desired_time - time.time()
        timeout_seconds = max(0, timeout_seconds)
        self._exit_flag.wait(timeout=timeout_seconds)

    def _set_blinked_in_gui(self, field_id):
        logger.info('Set blink for field %s.', field_id)
        return time.time()

    def _unblink(self):
        return self._set_unblinked_in_gui()

    def _set_unblinked_in_gui(self):
        logger.info('Unset blinks.')
        return time.time()

    def clear_gui(self):
        self._set_unblinked_in_gui()

    def _reset_managers(self):
        for m in self.mgrs:
            m.reset()

    def stop_blinking(self):
        if self._blinking_thread is None:
            logger.warning("Can't stop blinking: no blinking thread is running.")
        else:
            self._exit_flag.set()
            self._blinking_thread.join()
            self._blinking_thread = None
