import tempfile

from obci_demo.amplifiers import get_available_amplifiers


class AmplifierNotFound(Exception):
    pass


class InvalidSamplingRateSelected(Exception):
    pass


def get_amplifier_config(amplifier_id, analysis_channels, montage_channels, sampling_rate):
    amplifier = validate_amplifier_id(amplifier_id)
    sampling_rate = validate_sampling_rate(sampling_rate, amplifier_id)
    used_channels = montage_channels + analysis_channels
    used_channels_indexes = [amplifier['channels'].index(i) for i in used_channels]
    used_gains = [amplifier['gains'][i] for i in used_channels_indexes]
    used_offsets = [amplifier['offsets'][i] for i in used_channels_indexes]
    amplifier_class_name = amplifier['class_name']
    return {
        'amplifier_class': amplifier_class_name,
        'amplifier_id': amplifier_id,
        'channel_names': used_channels,
        'sampling_rate': sampling_rate,
        'samples_per_packet': None,
        'sample_type': 'FLOAT',
        'channels_info': None,
        'gains': used_gains,
        'offsets': used_offsets,
    }


def get_wisdom_path(wisdom_path, prefix):
    if wisdom_path is None:
        wisdom_path = tempfile.TemporaryDirectory(prefix=prefix + '-').name
        wisdom_path += '/obci_demo_classifier.dump'
    return wisdom_path


def generate(amplifier_id, analysis_channels, montage_channels,
             exclude_from_montage, sampling_rate, filter_frequency, wisdom_path):
    amplifier_config = get_amplifier_config(amplifier_id, analysis_channels, montage_channels, sampling_rate)
    return {
        'amplifier': amplifier_config,
        'p300': {
            'wisdom_path': get_wisdom_path(wisdom_path, 'obci_demo'),
            'epochs_dump_path': '',
            'channels_for_classification': analysis_channels,
            'channels_for_montage': montage_channels,
            'exclude_from_montage': exclude_from_montage,
            'power_line_freq': filter_frequency,
            'instruction_text': "Count appearances of the face on the field {calibration_option}",
            'field_labels': ['YES', 'NO'],
            'calibration': True,
            'blinker': {
                'blink_max_break': 0.350,
                'blink_min_break': 0.200,
                'blink_id_type': 'random_limited',
                'blink_id_count': 2,
                'blink_repeat_limit': 2,
                'blink_count_type': 'inf',
                'blink_count_min': 64,
                'blink_count_max': 80,
                'blink_duration': 0.100,
                'min_max_hallfblinks_between_targets': [1, 4],
                'send_halfblinks': True,
            },
            'channel_gains': amplifier_config['gains'],
            'channel_offsets': amplifier_config['offsets'],
        },
        'gui': {
            'feedback_time': 3,
            'pause_time': 2,
        },
        'controller': {
            'title': "BCI P300 Demo"
        }
    }


def validate_amplifier_id(amplifier_id):
    available = get_available_amplifiers()
    if amplifier_id not in available:
        raise AmplifierNotFound(':{}. Available amplifiers IDs: {}'.format(amplifier_id, available.keys()))
    return available[amplifier_id]


def validate_sampling_rate(sampling_rate, amplifier_id):
    available = get_available_amplifiers()[amplifier_id]['sampling_rates']
    if sampling_rate is None:
        sampling_rate = available[0]
    normalized_available = set(float(sr) for sr in available)
    if float(sampling_rate) not in normalized_available:
        raise InvalidSamplingRateSelected('Available sampling rates: {}, tried to select: {}'
                                          .format(available, sampling_rate))
    return sampling_rate
