import argparse
import logging
import os
import sys

import coloredlogs

from obci_demo import amplifiers, brokers, config_generation
from obci_demo.gui import bci_demo_app

logger = logging.getLogger(__name__)


def _is_editable(package_name):
    """Is distribution an editable install?"""
    for path_item in sys.path:
        egg_link = os.path.join(path_item, package_name + '.egg-link')
        if os.path.isfile(egg_link):
            return True
    return False


DEV = _is_editable('obci-demo')


class AnotherInstanceIsRunning(Exception):
    pass


def run(run_func=bci_demo_app.run, parser=None):
    if brokers.is_any_running():
        raise AnotherInstanceIsRunning('Only one instance of this program '
                                       'is allowed to run at a time.')
    else:
        coloredlogs.install(level=logging.INFO)
        if parser is None:
            parser = get_argument_parser()
        args, remaining = parser.parse_known_args()
        config = parser.generate_config(args)
        run_func(config)


def comma_list(string):
    return list(filter(None, string.split(',')))


def get_argument_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("--list-amplifiers", action=PrintAmplifiers,
                        help="List available amplifiers and return.", nargs=0)
    parser.add_argument("--amplifier-id", default='RandomAmplifierWithSine',
                        help="Existing ID of one of available amplifiers.")
    parser.add_argument("--analysis-channels", default='Random2,Random3,Random4,Random5',
                        help="Channels used for analysis, comma separated.", type=comma_list)
    parser.add_argument("--montage-channels", default='',
                        help="Channels used for montage, comma separated.", type=comma_list)
    parser.add_argument("--exclude-from-montage", default='',
                        help="Analysis channels to be excluded from montage, comma separated.", type=comma_list)
    parser.add_argument("--filter", default=50, type=int, choices=[50, 60],
                        help="Filter frequency to remove power network artifacts.")
    parser.add_argument("--sampling-rate", type=float,
                        help="Chosen sampling rate of selected amplifier.")
    parser.add_argument("--wisdom_path", default=None, type=str,
                        help="Path to wisdom file")
    parser.add_argument("--start", action='store_true', default=False,
                        help="Start sampling immediately")
    parser.generate_config = generate_config
    return parser


class PrintAmplifiers(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        descriptions = amplifiers.get_available_amplifiers()
        for i, description in enumerate(descriptions):
            name = description['amplifier_params']['channels_info']['name']
            amplifier_id = description['amplifier_params']['additional_params']['amplifier_id']
            channel_names = [channel['name'] for channel in
                             description['amplifier_params']['channels_info']['channels']]
            sampling_rates = description['amplifier_params']['channels_info']['sampling_rates']
            print('=== Amplifier #{} ==='.format(i))
            print('Name:', name)
            print('amplifier_id:', amplifier_id)
            print('Channel names:', channel_names)
            print('Sampling rates:', sampling_rates)
            print()
        parser.exit()


def generate_config(args: argparse.Namespace) -> dict:
    config = config_generation.generate(
        amplifier_id=args.amplifier_id,
        analysis_channels=args.analysis_channels,
        montage_channels=args.montage_channels,
        exclude_from_montage=args.exclude_from_montage,
        sampling_rate=args.sampling_rate,
        filter_frequency=args.filter,
        wisdom_path=args.wisdom_path,
    )
    if args.start:
        config['start'] = True
    return config


if __name__ == '__main__':
    run()
