# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved
from concurrent.futures import ThreadPoolExecutor
from enum import Enum
from functools import wraps
from logging import getLogger
from traceback import print_exc

from PySide2 import QtCore
from PySide2.QtCore import QJsonDocument, QJsonValue

from braintech.obci.core.utils.qt import invoke_in_main_thread
from obci_demo.amplifiers import get_available_amplifiers
from obci_demo.gui.bci_frontend import BCIFrontend
from obci_demo.gui.web_view import unwrap_json
from obci_demo.obci_demo_experiment import InvalidState, OBCIDemoExperiment

logger = getLogger(__name__)


def catch_invalid(func):
    @wraps(func)
    def wrapper(self, *args):
        if self._current_task and not self._current_task.done():
            return "Previous task did not finish yet"
        try:
            return func(self, *args) or "OK"
        except InvalidState as ex:
            return str(ex)

    return wrapper


class DemoStatus(Enum):
    STOPPED = 'stopped'
    READY = 'ready'
    CALIBRATING = 'calibrating'
    COMMUNICATING = 'communicating'
    CALIBRATED = 'calibrated'

    def __str__(self):
        return self.value


class BCIDemoController(QtCore.QObject):

    def __init__(self, config, frontend: BCIFrontend, experiment: OBCIDemoExperiment):
        QtCore.QObject.__init__(self)
        self._status = DemoStatus.STOPPED.value
        self._config = config
        self._frontend = frontend
        self._experiment = experiment
        experiment.controller = self
        self._current_task = None
        self._executor = ThreadPoolExecutor()

        # self._current_task.add_done_callback(lambda _: self.start_calibration())
        self._frontend.key_pressed.connect(self._experiment.key_pressed)
        self._frontend.stop_pressed.connect(lambda: self.stop())  # connecting self.stop directly causes crash...
        self.settings = {
            'bci_panel_addresses': self._experiment.bci_panel.get_addresses(),
            'title': config['controller']['title'],
        }
        self.amplifiers = get_available_amplifiers()
        self.calibration_finished.connect(lambda: self.stop())
        if 'start' in config:
            self._start_amplifier(None)

    def _get_amplifiers(self):
        return self._amplifiers

    @unwrap_json
    def _set_amplifiers(self, amps):
        self._amplifiers = QJsonDocument(amps)
        self.amplifiers_changed.emit()

    amplifiers_changed = QtCore.Signal()

    amplifiers = QtCore.Property(QJsonDocument, _get_amplifiers, _set_amplifiers, notify=amplifiers_changed)

    def _get_settings(self):
        return self._settings

    @unwrap_json
    def _set_settings(self, settings):
        self._settings = QJsonDocument(settings)
        self.settings_changed.emit()

    settings_changed = QtCore.Signal()

    settings = QtCore.Property(QJsonDocument, _get_settings, _set_settings, notify=settings_changed)

    def _get_status(self):
        return self._status

    def _set_status(self, status):
        if status.value == self._status:
            return
        logger.info("Changing state from %s to %s", self._status, status.value)
        self._status = status.value
        self.status_changed.emit()

    status_changed = QtCore.Signal()
    status = QtCore.Property(str, _get_status, _set_status, notify=status_changed)

    @QtCore.Slot(QJsonValue, result=str)
    @unwrap_json
    @catch_invalid
    def start_amplifier(self, config):
        self._run_async(self._start_amplifier, config)

    def _start_amplifier(self, config):
        if config:
            exp_config = self._experiment.generate_config(config)
        else:
            exp_config = self._config
        self._experiment.start(exp_config)
        self._update_ready_status()

    @QtCore.Slot(result=str)
    @catch_invalid
    def stop_amplifier(self):
        if self._experiment:
            self._run_async(self._stop_runner)
        else:
            return "No Peers runner to stop"

    def _stop_runner(self):
        self._experiment.stop()
        self.status = DemoStatus.STOPPED

    @QtCore.Slot(result=str)
    @catch_invalid
    def start_calibration(self):
        self._run_async(self._start, calibration=True)

    def _start(self, calibration=True):
        if calibration:
            self._experiment.start_calibration()
            self.status = DemoStatus.CALIBRATING
        else:
            self._experiment.start_bci()
            self.status = DemoStatus.COMMUNICATING
        invoke_in_main_thread(self._frontend.show)

    def _expect_status(self, status):
        if DemoStatus(self.status) != status:
            raise InvalidState("Wrong status %s" % self.status)

    @QtCore.Slot(result=str)
    @catch_invalid
    def start_bci(self):
        self._expect_status(DemoStatus.CALIBRATED)
        self._run_async(self._start, calibration=False)

    @QtCore.Slot(result=str)
    @catch_invalid
    def stop(self):
        self._experiment.finish()
        self._update_ready_status()
        self._frontend.hide()

    def _update_ready_status(self):
        if self._experiment.is_calibrated():
            self.status = DemoStatus.CALIBRATED
        else:
            self.status = DemoStatus.READY

    def _run_async(self, func, *args, **kwargs):
        @wraps(func)
        def _wrapped(*args, **kwargs):
            try:
                func(*args, **kwargs)
            except Exception:
                print_exc()
                raise

        self._current_task = self._executor.submit(_wrapped, *args, **kwargs)

    def close(self):
        self._stop_runner()
        self._experiment.shutdown()
        if self._current_task:
            self._current_task.result()

    calibration_finished = QtCore.Signal()
