# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved
import asyncio
import io
import mimetypes
import threading
from collections import defaultdict
from logging import getLogger

from PySide2 import QtWebChannel, QtCore
from PySide2.QtCore import QJsonDocument, QResource
from aiohttp import web

from bci_panel.web_bci_panel import WebBCIPanelServer, WebBCIPanelPeer, CHARTS
from braintech.obci.core.broker.message_handler_mixin import subscribe_message_handler
from braintech.obci.core.utils.net import get_all_ip4_addresses
from obci_demo.blinking import P300Blinker
from obci_demo.brokers import P300GuiBroker
from obci_demo.gui.web_view import WebView, unwrap_json
from obci_demo.obci_demo_experiment import OBCIDemoExperiment
from obci_p300.common.messages import DecisionVerboseMsg

LOGGER = getLogger(__name__)


class BCIFrontend(QtCore.QObject):
    TITLE = "BCI Stimulus"

    def __init__(self, field_labels):
        super().__init__(None)
        self.view = WebView(None)
        self.view.setWindowTitle(self.TITLE)
        self.page = self.view.page()
        channel = QtWebChannel.QWebChannel(self)
        self._channel = channel
        self.page.setWebChannel(channel)
        channel.registerObject("frontend", self)

        self.view.hide()
        self._field_labels = field_labels
        self.field_count = len(field_labels)
        self._field_config = None
        self._mode = None
        self._instruction = ''
        self.clear()
        self.view_updated = threading.Condition()
        self._last_update = 0
        self.key_pressed.connect(self._handle_key)
        self.view.window_event.connect(self._window_event)
        self._handlers = defaultdict(list)
        self._event.connect(self._call_handlers)

    def load(self, url):
        self.view.load(QtCore.QUrl(url))

    def _get_field_config(self):
        return self._field_config

    @unwrap_json
    def _set_field_config(self, val):
        self._field_config = QJsonDocument.fromVariant(val)
        self.field_config_changed.emit()

    field_config_changed = QtCore.Signal()
    field_config = QtCore.Property(QJsonDocument, _get_field_config, _set_field_config, notify=field_config_changed)

    def _get_instruction(self):
        return self._instruction

    def _set_instruction(self, val):
        self._instruction = val
        self.instruction_changed.emit()

    instruction_changed = QtCore.Signal()
    instruction = QtCore.Property(str, _get_instruction, _set_instruction, notify=instruction_changed)

    def _get_mode(self):
        return self._mode

    def _set_mode(self, mode):
        if mode != self._mode:
            self._mode = mode
            self.mode_changed.emit()

    mode_changed = QtCore.Signal()
    mode = QtCore.Property(str, _get_mode, _set_mode, notify=mode_changed)

    def show(self):
        LOGGER.info("Show Frontend")
        self.clear()
        self.view.showNormal()

    def hide(self):
        LOGGER.info("Hide Frontend")
        self.view.hide()

    def _get_base_config(self):
        return [{'state': '', 'label': l} for l in self._field_labels]

    def clear(self):
        self.field_config = self._get_base_config()

    def select(self, field_num):
        config = self._get_base_config()
        config[field_num]['state'] = 'selected'
        self.field_config = config

    @QtCore.Slot(float)
    def display_updated(self, draw_time):
        with self.view_updated:
            self._last_update = draw_time
            self.view_updated.notify_all()

    def wait_for_redraw(self):
        with self.view_updated:
            self.view_updated.wait(0.5)
            return self._last_update

    @QtCore.Slot()
    def toggle_fullscreen(self):
        if self.view.isFullScreen():
            self.view.showNormal()
        else:
            self.view.showFullScreen()
        self.fullscreen_changed.emit()

    def _fullscreen(self):
        return self.view.isFullScreen()

    fullscreen_changed = QtCore.Signal()
    fullscreen = QtCore.Property(bool, _fullscreen, notify=fullscreen_changed)

    def _handle_key(self, key_name):
        if key_name in ['Escape', 'Esc'] and self.view.isFullScreen():
            self.view.showNormal()

    def _window_event(self, name):
        if name == 'maximized':
            self.view.showFullScreen()
            self.fullscreen_changed.emit()
        elif name == 'closed':
            self.stop()

    key_pressed = QtCore.Signal(str)

    @QtCore.Slot(str)
    def key_pressed_emit(self, key_name):
        self.key_pressed.emit(key_name)

    stop_pressed = QtCore.Signal()

    @QtCore.Slot()
    def stop(self):
        self.stop_pressed.emit()

    _event = QtCore.Signal(str, QJsonDocument)

    @QtCore.Slot(str, QJsonDocument)
    def toggle(self, event_name, data=None):
        if data is None:
            data = {}
        self._event.emit(event_name, QJsonDocument(data))

    def on(self, event_name, callback):
        self._handlers[event_name].append(callback)

    def _call_handlers(self, event_name, data):
        for h in self._handlers[event_name]:
            h(data.toVariant())


class P300BCIFrontend(BCIFrontend):
    TITLE = "BCI P300 Stimulus"

    def blink(self, states):
        self.field_config = states


class P300FrontendBlinker(P300Blinker):

    def __init__(self, blinker_config, messaging_peer):
        super().__init__(blinker_config, messaging_peer)
        self.frontend = None  # type: P300BCIFrontend

    def _send_config(self, config):
        self.frontend.field_config = config

    def _set_blinked_in_gui(self, field_id):
        config = ['fullblink' if field_id == i else 'halfblink' for i in range(self.frontend.field_count)]
        self.frontend.blink(config)
        return self.frontend.wait_for_redraw()

    def _set_unblinked_in_gui(self):
        self.frontend.clear()
        return self.frontend.wait_for_redraw()


class FrontendBroker(P300GuiBroker):
    def _create_blinker(self, config):
        return P300FrontendBlinker(config, self)

    def __init__(self, config, *args, **kwargs):
        super().__init__(config, *args, **kwargs)
        self._frontend = None  # type: BCIFrontend

    @property
    def frontend(self):
        return self._frontend

    @frontend.setter
    def frontend(self, frontend):
        self._frontend = frontend
        self._blinker.frontend = frontend

    @subscribe_message_handler(DecisionVerboseMsg)
    def handle_p300_decision_message(self, message: DecisionVerboseMsg):
        super(FrontendBroker, self).handle_p300_decision_message(message)

        async def delayed_select():
            await asyncio.sleep(0.2)
            self.frontend.select(int(message.score))

        if message.decision_type != 'LEARN':
            self.create_task(delayed_select())

    def _set_gui_message(self, instruction):
        super(FrontendBroker, self)._set_gui_message(instruction)
        self._frontend.instruction = instruction


class OBCIDemoExperimentWithFrontend(OBCIDemoExperiment):
    BrokerClass = FrontendBroker

    def __init__(self, config, frontend):
        super().__init__(config, True)
        self._frontend = frontend
        self.broker.frontend = frontend
        self.controller = None

    def show_calibration_finished(self):
        self._logic.show_calibration_done()

    def calibration_finished(self):
        if self.controller:
            self.controller.calibration_finished.emit()

    def _create_bci_panel(self):
        return self.add_peer(QtWebBCIPanelPeer, peer_id='bci_panel')  # type: QtWebBCIPanelPeer


class QtWebBCIPanelServer(WebBCIPanelServer):
    _last_sid = None
    MAX_CLIENTS = 100
    ROOM_ID = 'bci_panel'

    def __init__(self, namespace):
        super().__init__(namespace)
        self.prefix = ''
        self._clients = []

    def _add_routes(self, router):
        router.add_get('/{path:.*$}', self._handle_qrc)

    async def _handle_qrc(self, request):
        path = request.match_info['path']
        if not path:
            path = 'index.html'
        path = ':/%s/%s' % (self.prefix, path)
        resource = QResource(path)
        ct, encoding = mimetypes.guess_type(path)
        if resource.isValid():
            return web.Response(body=io.BytesIO(resource.data()), content_type=ct, charset=encoding)
        return web.Response(status=404, text="404")

    def _on_connect(self, sid, environ):
        if len(self._clients) >= self.MAX_CLIENTS:
            self._sio.leave_room(self._clients[0], self.ROOM_ID, self.namespace)
            self._sio.disconnect(self._clients[0])
            self._clients.pop(0)
        self._clients.append(sid)
        self._sio.enter_room(sid, self.ROOM_ID, self.namespace)
        super(QtWebBCIPanelServer, self)._on_connect(sid, environ)

    def _on_disconnect(self, sid):
        super(QtWebBCIPanelServer, self)._on_disconnect(sid)
        self._sio.leave_room(sid, self.ROOM_ID, self.namespace)
        if sid in self._clients:
            self._clients.remove(sid)

    async def emit(self, name, data):
        await self._sio.emit(name, data, namespace=self.namespace, room=self.ROOM_ID)


class QtWebBCIPanelPeer(WebBCIPanelPeer):
    def _create_bci_panel_server(self):
        return QtWebBCIPanelServer(CHARTS)

    def set_prefix(self, prefix):
        self.bci_panel.prefix = prefix

    def get_addresses(self):
        if self.server_host == '127.0.0.1':
            local_ips = ['127.0.0.1']
        elif self.server_host == '0.0.0.0':
            local_ips_all = get_all_ip4_addresses()
            excluded = ['172',
                        '127',
                        '169']
            local_ips = [i for i in local_ips_all if i.split('.')[0] not in excluded]
        else:
            local_ips = [self.server_host]
        return ['%s:%d' % (ip, self.server_port) for ip in local_ips]
