import asyncio
import logging

import obci_p300.common.messages
from bci_panel.web_bci_panel import RemoteWebBCIPanel
from braintech.obci.core.broker import peer
from braintech.obci.core.broker.message_handler_mixin import subscribe_message_handler
from braintech.obci.core.broker.messages import BaseMessage, Field
from obci_p300 import p300_peer
from obci_p300.common import messages
from obci_ugm.utils import sequence_provider

logger = logging.getLogger(__name__)


class P300Message(BaseMessage):
    key = Field(str)
    value = Field(str)


class PreconfiguredP300AnalysisPeer(p300_peer.BaseP300Peer):

    def _set_static_parameters(self, config: dict, amplifier_config: dict):
        # timeout after decision when nothing should be accepted and no decision sent.
        self._post_decision_blinks_ignore_time = 3
        # confidence (probability) level for internal decision to be counted to decision stop:
        # decision False Positive Rate - maximal level of error for internal decision.
        self.allowed_classifier_FPR = 0.2

        self.sampling_rate = amplifier_config['sampling_rate']
        # available channels
        self.channel_names = amplifier_config['channel_names']
        self.channel_gains = amplifier_config['gains']
        self.channel_offsets = amplifier_config['offsets']
        # used channels
        self.channels_for_classification = config['channels_for_classification']
        # channel montage:
        # montage type
        self.montage = 'custom'
        # montage channels
        self.montage_channels = config["channels_for_montage"]

        # learn every n targets
        self.learn_every = 5

        # pre event baseline in seconds (negative) seconds
        self.baseline = -0.2

        # post event window to consider seconds
        self.window = 0.9

        # maximum averaged epochs
        # can be inf
        self.maximum_to_average = 10
        # identical decisions to get final answer
        self.decision_stop = 3
        # downsample to
        self.targetFs = 24

        self.ignored_blink_ids = [-1]

        self.proba_used_for_internal_decision = 'weighted_mean'
        self.use_linear_proba = True
        self.use_web_bci_panel = True
        self.use_windowed_bci_panel = False
        self.ssvep_cs_field_id_to_blinker_map = {}
        self.visualise_classification_epochs_bool = False
        self.offline_learning = False
        self.offline_learning_dataset_path = ''
        self.plots_image_path = ''
        self.epochs_dump_path = ''
        self.wisdom_path = config['wisdom_path']

        self.power_line_freq = config['power_line_freq']
        self.exclude_from_montage = config['exclude_from_montage']
        self.wisdom_load = config.get('wisdom_load', False)
        self.training_index = None

    def _create_bci_panel(self):
        self.bci_panel = RemoteWebBCIPanel(self)

    def set_calibration(self, calibration):
        if calibration:
            self.training_index = 0
            self._analysis.training_index = 0
            self.wisdom_load = False
        else:
            self.training_index = None
            self._analysis.training_index = None
            self.wisdom_load = True
        self.create_task(self._setup_before_running()).result()
        if calibration:
            self.reset_to_blank_state()


class P300LogicPeer(peer.Peer):
    pause_text = 'PAUZA'
    instruction_time = 2.5
    PREPARE_TIME = 2
    READING_SPEED = 863 / 60  # characters per second

    class InvalidConfiguration(Exception):
        pass

    def __init__(self, config, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._break_duration = (config['gui']['pause_time']
                                + config['gui']['feedback_time'])
        self._field_labels = config['p300']['field_labels']
        self._raw_instruction_text = config['p300']['instruction_text']
        self._is_calibration = config['p300']['calibration']
        self._calibration_sequence = sequence_provider.RandomLimitedMgr(
            start=0, count=2, repeat_limit=2)
        self._task = None
        self.autostart = self.autoshutdown = False
        self._validate_configuration()
        self._last_key = None

    async def _initialize(self):
        self._decision = asyncio.Event()
        self._cv_started = asyncio.Event()
        self._key_pressed = asyncio.Condition()

        await super(P300LogicPeer, self)._initialize()

    def _validate_configuration(self):
        if not self._break_duration and self._break_duration > 0:
            raise self.InvalidConfiguration("Break can't end: _break_duration={}."
                                            .format(self._break_duration))

    async def _start(self):
        await super()._start()
        self._task = self.create_task(self.blinking_task())

    async def _wait_for_space(self):
        with await self._key_pressed:
            self._last_key = None
            await self._key_pressed.wait_for(lambda: self._last_key == ' ')

    async def blinking_task(self):
        await self._display_text("Get ready for {session}. Press space when ready to continue.".format(
            session="calibration" if self._is_calibration else "BCI session"
        ), pause=False)
        await self._wait_for_space()
        while self.is_running:
            if self._is_calibration:
                new_target_index = self._pick_instruction_index()
                await self._broadcast_current_index(new_target_index)
                await self._display_instruction(new_target_index)
            else:
                await self._display_text("Select your answer")
            self._decision.clear()
            self._cv_started.clear()
            try:
                await self._start_blinking()
                if self._is_calibration:
                    await self._cv_started.wait()
                    await self._stop_blinking()
                    await self._display_text("Calculating")
                await self._decision.wait()
            finally:
                await self._stop_blinking()
            await self._take_a_break()

    def _pick_instruction_index(self):
        return self._calibration_sequence.get_value()

    async def _broadcast_current_index(self, index):
        message = messages.CalibrationIndexUpdate(
            calibration_index=str(index),
        )
        await self._send_message(message)

    async def _display_instruction(self, target_index):
        instruction = self._format_instruction(target_index)
        await self._display_text(instruction)

    def _format_instruction(self, target_index):
        field_text = self._field_labels[target_index]
        return self._raw_instruction_text.format(calibration_option=field_text)

    async def _start_blinking(self):
        await self._send_message(P300Message(key='start_blinking', value=''))

    async def _take_a_break(self):
        if self._is_calibration:
            text = "BREAK <br/> Press space to continue"
            await self._display_text(text, pause=False)
            await self._wait_for_space()
        else:
            await self._display_text("PAUZA", pause=False)
            await self._wait_for_space()
            await self._send_message(P300Message(key='clear'))
        # await asyncio.sleep(self._break_duration)

    async def _display_text(self, text, pause=True):
        await self._send_message(P300Message(key='instruction', value=text))
        if pause:
            await asyncio.sleep(len(text) / self.READING_SPEED)

    @subscribe_message_handler(obci_p300.common.messages.CVStartedMsg)
    async def handle_cv_started(self, _):
        self._cv_started.set()

    @subscribe_message_handler(obci_p300.common.messages.DecisionVerboseMsg)
    async def handle_decision(self, msg):
        if self._is_calibration:
            self._decision.set()
        else:
            if msg.decision == 'CLASSIFICATION':
                self._decision.set()

    async def key_pressed_async(self, key_name):
        with await self._key_pressed:
            self._last_key = key_name
            self._key_pressed.notify_all()

    def key_pressed(self, key_name):
        self.create_task(self.key_pressed_async(key_name))

    async def _stop(self):
        if self._task is not None:
            self._task.cancel()
        await super()._stop()

    async def _stop_blinking(self):
        await self._send_message(P300Message(key='stop_blinking', value=''))

    def start_calibration(self):
        self._logger.info("START CALIBRATION")
        self._is_calibration = True
        self.start()

    def show_calibration_done(self):
        if self._task is not None:
            self._task.cancel()
            self._task = None
        if self._is_calibration:
            self.create_task(self._display_text('Calibration done'))

    def start_bci(self):
        self._logger.info("START BCI")
        self._is_calibration = False
        self.start()
