# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved
import multiprocessing
import os
import queue
import sys
import threading
import time
import warnings
from collections import namedtuple
from functools import partial
from logging import getLogger

from braintech.obci.core.broker.asyncio_task_manager import ShuttingDownException
from braintech.obci.core.broker.base_peer import PeerState
from braintech.obci.core.broker.broker import Broker

Request = namedtuple('Request', 'attr,args,kwargs')
Response = namedtuple('Response', 'exception,value')

SHUTDOWN = 'SHUTDOWN'
LOGGER = getLogger(__name__)
MP_CONTEXT = multiprocessing.get_context()
if os.name == 'posix':
    try:
        from PySide2.QtWidgets import QApplication

        app = QApplication.instance()
        if app is not None:
            warnings.warn("QApplication already created. Using SPAWN multiprocessing context, which might be slow")
            MP_CONTEXT = multiprocessing.get_context('spawn')
        elif sys.platform.startswith('darwin'):
            MP_CONTEXT = multiprocessing.get_context('spawn')
        else:
            MP_CONTEXT = multiprocessing.get_context('forkserver')
    except ImportError:
        pass


class SubprocessPeer:
    KNOWN_PROPERTIES = ['id', 'is_running', 'is_ready']

    @staticmethod
    def _process_requests(peer, req_queue: multiprocessing.Queue, resp_queue: multiprocessing.Queue):
        p_pid = os.getppid()
        while True:
            try:
                req = req_queue.get(timeout=0.25)  # type:Request
            except queue.Empty:
                if os.getppid() != p_pid:
                    LOGGER.warning("Parent process is probably dead. Shutting down")
                    peer.shutdown()
                    return
            else:
                resp = SubprocessPeer._process_request(req, peer)
                if req.attr == 'shutdown':
                    break
                else:
                    resp_queue.put(resp)

    @staticmethod
    def _process_request(req, peer):
        try:
            val = getattr(peer, req.attr)
            if callable(val) or req.args or req.kwargs:
                val = val(*req.args, **req.kwargs)
            return Response(False, val)
        except Exception as ex:
            return Response(True, ex)

    @staticmethod
    def _run_peer(PeerClass, req_queue, resp_queue, args, kwargs):
        peer = PeerClass(*args, **kwargs)
        comm_thread = threading.Thread(target=SubprocessPeer._process_requests, args=(peer, req_queue, resp_queue),
                                       daemon=True)
        comm_thread.start()
        try:
            peer.run()
        finally:
            resp_queue.put(Response(False, SHUTDOWN))

    def __init__(self, PeerClass, *args, **kwargs):
        if kwargs.pop("force_spawn", False):
            context = multiprocessing.get_context('spawn')
        else:
            context = MP_CONTEXT
        self._req_queue = context.Queue()
        self._res_queue = context.Queue()
        self._process = context.Process(target=SubprocessPeer._run_peer,
                                        args=(PeerClass, self._req_queue, self._res_queue, args, kwargs),
                                        name=PeerClass.__name__)
        self._process.daemon = True
        self._process.start()
        self.id = getattr(self, 'id')

    def _peer_method(self, name, *args, **kwargs):
        if not self._process.is_alive():
            raise ShuttingDownException
        self._req_queue.put(Request(name, args, kwargs))
        response = self._res_queue.get()  # type:Response
        if response.exception:
            raise response.value
        else:
            return response.value

    def __getattr__(self, attr):
        if attr in self.KNOWN_PROPERTIES:
            return self._peer_method(attr)
        return partial(self._peer_method, attr)

    def shutdown(self):
        resp = self._peer_method('shutdown')
        self._process.join()
        return resp


class SimpleExperiment:
    def _create_broker(self):
        return Broker()

    def __init__(self, broker=None):
        if broker is None:
            broker = self._create_broker()
        self.broker = broker
        self._peers = []

    def _add_peer(self, as_subprocess, PeerClass, *args, **kwargs):
        kwargs['urls'] = self.broker.broker_ip
        if as_subprocess:
            peer = SubprocessPeer(PeerClass, *args, **kwargs)
        else:
            peer = PeerClass(*args, **kwargs)
        self._peers.append(peer)
        return peer

    def add_peer(self, PeerClass, *args, **kwargs):
        return self._add_peer(False, PeerClass, *args, **kwargs)

    def add_subprocess_peer(self, PeerClass, *args, **kwargs):
        return self._add_peer(True, PeerClass, *args, **kwargs)

    def remove_peer(self, peer):
        self._peers.remove(peer)
        peer.shutdown()

    def shutdown(self):
        try:
            for p in self._peers:
                try:
                    p.shutdown()
                except ShuttingDownException:
                    pass
        except AttributeError:
            pass
        self._peers = []
        if self.broker:
            self.broker.shutdown()
        self.broker = None

    def wait_for(self, peers, state, timeout=10):
        end_time = time.monotonic() + timeout
        while True:
            for p in peers:
                if state == PeerState.ready:
                    if not p.is_ready:
                        break
                elif not p.is_running:
                    break
            else:
                return
            time.sleep(0.1)
            if time.monotonic() > end_time:
                raise TimeoutError()

    def __del__(self):
        self.shutdown()
