import collections
import enum
import logging
import numpy
import scipy.stats
import warnings
from typing import Optional

from braintech.obci.core.broker.messages import BaseMessage, Field
from obci_ugm.utils.math import xlog2


class ReporterSummaryMsg(BaseMessage):
    accuracy = Field(float)
    itr_start_blinking_to_end = Field(float)
    utility_start_blinking_to_end = Field(float)
    command_count = Field(int)
    correct_count = Field(int)
    mean_time = Field(float)
    null_proba = Field(float)

    @property
    def dict_data(self):
        itr = self.itr_start_blinking_to_end
        acc = self.accuracy
        utility = self.utility_start_blinking_to_end
        # Keys are used in bci_panel.bci_panel.WebBCIPanel#_send_acc_itr
        return collections.OrderedDict([
            ('ITR', '{:.2f} bit/min'.format(itr)),
            ('Accuracy', '{:.0f} %'.format(acc * 100)),
            ('Utility', '{:.2f}'.format(utility)),
            ('Command count', '{}'.format(self.command_count)),
            ('Correct count', '{}'.format(self.correct_count)),
            ('Mean answer time', '{:.2f}s'.format(self.mean_time)),
            ('Binom test p', '{:.4f}'.format(self.null_proba))
        ])


class BCIReporterTrialInfo:
    """
    Contains information on a single trial of BCI system.
    All custom (paradigm-related) information should be put under "additional_info".
    """
    def __init__(self, timestamp_command: float, timestamp_start: float):
        self.timestamp_command = timestamp_command
        self.timestamp_start = timestamp_start
        self.timestamp_end = None
        self.time_command_to_end = numpy.nan
        self.time_start_to_end = numpy.nan
        self.answer = None
        self.correct_answer = None
        self.is_correct = -1
        self.additional_info = {}


class BCIReporterBase:
    """
    Gathers information about BCI run.
    """
    def __init__(self, logger: logging.Logger):
        """
        Create a new reporter instance.
        :param logger: logger interface to be used by BCIReporter
        """
        self.trials = []
        self.current_trial = None
        self.timestamp_command = None
        self.command = None
        self.logger = logger

    def command_displayed(self, timestamp: float, command: str):
        """
        Record information of the command being displayed.
        :param timestamp: timestamp (UTC) of command displaying
        :param command: textual representation of the command
        """
        self.timestamp_command = timestamp
        self.command = command

    def trial_started(self, timestamp: float):
        """
        Record information of the new trial being started.
        If the command was not displayed since the previous trial,
        the time of the command displaying (needed for some statistics)
        will be considered the same as the time of the trial starting.
        :param timestamp: timestamp (UTC) of trial starting
        """
        if self.timestamp_command is None:
            if self.logger:
                self.logger.warning('command_displayed was not called before trial_started')
            # if command has not been displayed we have to pretend it was
            timestamp_command = timestamp
        else:
            timestamp_command = self.timestamp_command
        self.current_trial = BCIReporterTrialInfo(timestamp_command, timestamp)
        self.timestamp_command = None

    def trial_finished(self, timestamp: float, **info):
        """
        Record information of last trial being finished.
        Calling this method more than once for a given trial will have no effect,
        neither will calling this method if no trial has been started.
        :param timestamp: timestamp (UTC) of trial finishing
        :param info: additional (paradigm-related) information for the trial
        """
        if self.current_trial is not None and self.current_trial.timestamp_end is None:
            self.current_trial.timestamp_end = timestamp
            self.current_trial.time_start_to_end = timestamp - self.current_trial.timestamp_start
            self.current_trial.time_command_to_end = timestamp - self.current_trial.timestamp_command
            self.current_trial.additional_info = info
            self.trials.append(self.current_trial)

    def set_answer(self, answer, correct_answer=None):
        """
        Record what was the answer (and optionally, the correct answer) in the last trial.
        If correct answer is also given, then the answer correctness will be computed as well.
        :param answer: actual answer
        :param correct_answer: correct answer or None if unknown
        """
        if self.current_trial is not None:
            self.current_trial.answer = answer
            self.current_trial.correct_answer = correct_answer
            if correct_answer is not None:
                self.current_trial.is_correct = (answer == correct_answer)

    def set_answer_correctness(self, is_correct: int):
        """
        Record whether the last trial resulted in a correct answer.
        Calling this method if no trial has been started will have no effect.
        :param is_correct: 1 if correct, 0 if not correct, -1 if "don't know"
        """
        if self.current_trial is not None:
            self.current_trial.is_correct = is_correct

    def fetch_last_decision_tag(self) -> Optional[dict]:
        """
        Return information (in form of a tag fields) about the most recent trial.
        Calling this method more than once for a given trial will return None,
        as will calling this method if no trial has been started.
        :return: dictionary of trial tag information or None
        """
        if self.current_trial is None or self.current_trial.timestamp_end is None:
            return None
        try:
            return {
                'p_start_timestamp': self.current_trial.timestamp_start,
                'p_end_timestamp': self.current_trial.timestamp_end,
                'p_tag_name': 'trial',
                'p_tag_desc': {
                    'command': self.command,
                    'answer': self.current_trial.answer,
                    'correct_answer': self.current_trial.correct_answer,
                    'time_command_to_end': self.current_trial.time_command_to_end,
                    'time_start_to_end': self.current_trial.time_start_to_end,
                    # cannot compute third variant of time (next_start - prev_start)
                    # since the timestamp of next command is not yet available
                    **self.current_trial.additional_info
                }
            }
        finally:
            self.current_trial = None

    def get_trials(self) -> list:
        """
        Generate detailed information of all trials performed so far.
        :return: array of trial information (one dict per trial)
        """
        results = list(map(vars, self.trials))
        last_trial = None
        for trial in results:
            # compute time_start_to_start from adjacent trials
            if last_trial is not None:
                last_trial['time_start_to_start'] = trial['timestamp_start'] - last_trial['timestamp_start']
            last_trial = trial
        return results


class BCIReporter(BCIReporterBase):
    """
    Gathers information about BCI run and is able to generate statistics.
    """
    def __init__(self, number_of_choices: int, logger: logging.Logger):
        """
        Create a new reporter instance.
        :param number_of_choices: number of choices (fields) in BCI system
        :param logger: logger interface to be used by BCIReporter
        """
        super().__init__(logger)
        self.number_of_choices = number_of_choices

    def get_summary(self) -> dict:
        """
        Generate information summary of the BCI system performance so far,
        based on all trial information gathered by this object.
        :return: dictionary with summary information
        """
        command_count = len(self.trials)
        correct_count = sum(trial.is_correct == 1 for trial in self.trials)
        accuracy = correct_count / command_count if command_count else numpy.nan
        mean_times = {}
        with warnings.catch_warnings():
            # warnings can appear if self.trials have not enough data
            warnings.simplefilter("ignore", category=RuntimeWarning)
            mean_times[IntervalType.START_TO_START] = numpy.nanmean(
                numpy.diff([trial.timestamp_start for trial in self.trials]))
            mean_times[IntervalType.COMMAND_TO_END] = numpy.nanmean(
                [trial.time_command_to_end for trial in self.trials])
            mean_times[IntervalType.START_TO_END] = numpy.nanmean([trial.time_start_to_end for trial in self.trials])
        bit_rate = self._compute_bit_rate(self.number_of_choices, accuracy)
        utility = self._compute_utility(self.number_of_choices, accuracy)
        chance_level = 1 / self.number_of_choices
        summary = {
            'command_count': command_count,
            'correct_count': correct_count,
            'accuracy': accuracy,
            'null_proba': scipy.stats.binom_test(correct_count, command_count,
                                                 chance_level) if command_count else numpy.nan,
        }
        for metric_type, mean_time in mean_times.items():
            mean_time_minutes = mean_time / 60
            suffix = '_' + metric_type.name.lower()
            summary['mean_time' + suffix] = mean_time
            summary['itr' + suffix] = bit_rate / mean_time_minutes
            summary['utility' + suffix] = utility / mean_time_minutes
        return summary

    @staticmethod
    def _compute_bit_rate(N, P):
        """
        Compute BCI system bit-rate measure.
        :param N: number of choices (fields) in BCI
        :param P: accuracy of BCI system
        :return: bit-rate measure in bits per trial
        """
        return numpy.log2(N) + xlog2(P) + (N - 1) * xlog2((1 - P) / (N - 1))

    @staticmethod
    def _compute_utility(N, P):
        """
        Compute BCI system "utility" measure according to Dal Seno et al. (2009).
        :param N: number of choices (fields) in BCI
        :param P: accuracy of BCI system
        :return: utility measure
        """
        return (2 * P - 1) * numpy.log2(N - 1)


class BCIReporterP300(BCIReporter):
    """
    Gathers information about BCI run for P300 BCI.
    """

    def get_summary(self) -> dict:
        """
        Generate information summary of the P300 BCI system performance so far,
        based on all trial information gathered by this object,
        including mean blink count across all trials.
        :return: dictionary with summary information
        """
        summary = super().get_summary()
        with warnings.catch_warnings():
            # warnings can appear if no trial have valid blink count
            warnings.simplefilter("ignore", category=RuntimeWarning)
            summary['mean_blink_count'] = numpy.nanmean([
                sum(trial.additional_info['blink_counts'].values())
                if 'blink_counts' in trial.additional_info else numpy.nan
                for trial in self.trials
            ])
        return summary


class IntervalType(enum.Enum):
    START_TO_START = 0
    COMMAND_TO_END = 1
    START_TO_END = 2
