# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.
import asyncio

from PySide2.QtWidgets import QApplication

from braintech.obci.core.broker.message_handler_mixin import subscribe_message_handler
from braintech.obci.core.broker.messages import TagMsg
from braintech.obci.experiment.peers.drivers.amplifiers.amplifier_peer import (BaseAmplifierPeer, AmplifierPeer,
                                                                               register_amplifier_peer)
from braintech.obci.core.utils.qt import invoke_in_main_thread
from obci_p300.common.messages import CalibrationIndexUpdate
from obci_p300.common.random_amp_with_sine.random_amplifier_with_sine_app import RandomAmplifierWithSineAppWindow
from .random_amplifier_with_sine import RandomAmplifierWithSine

__all__ = ('RandomAmplifierWithSinePeer',)
P300_LATENCY = 0.3


class BaseRandomAmplifierWithSinePeer(BaseAmplifierPeer):
    """Random amplifier - sends random data (uniform [0;1)) over all channels."""

    AmplifierClass = RandomAmplifierWithSine

    def __init__(self, *args, **kwargs):
        self._autocalibrate_ssvep = False
        self._autocalibrate_p300 = False
        self._window = None
        super().__init__(*args, **kwargs)

    def p300_autocalibration_check_handle(self, state):
        self._autocalibrate_p300 = state > 0

    def blink_text_updated_handle(self):
        self.react_to = str(self._window.blink_reaction.toPlainText())

    @subscribe_message_handler(CalibrationIndexUpdate)
    def handle_training_index_change(self, msg: CalibrationIndexUpdate):
        index_string = msg.calibration_index
        if self._autocalibrate_p300:
            self._window.update_react_to.emit(str(index_string))

    @subscribe_message_handler(TagMsg)
    async def handle_tag_to_blink_converter(self, msg: TagMsg):
        if msg.name == 'blink':
            blink_info = msg.desc['stimuli_id']
            self._window.update_last_blink.emit(str(blink_info))
            await self._handle_react_to(blink_info)

    async def _handle_react_to(self, blink_info):
        if str(self.react_to) == str(blink_info):
            await asyncio.sleep(P300_LATENCY)
            self.generate_p300()
        if isinstance(blink_info, (list, set, tuple)):
            if str(self.react_to) in [str(i) for i in blink_info]:
                await asyncio.sleep(P300_LATENCY)
                self.generate_p300()

    def amplifier_frequency_select_handle(self, f):
        if self._amplifier is not None:
            self._amplifier.set_sine_freq(f)

    def amplifier_sine_amplitude_handle(self, a):
        if self._amplifier is not None:
            self._amplifier.set_sine_amplitude(a)

    def amplifier_noise_std_handle(self, std):
        if self._amplifier is not None:
            self._amplifier.set_noise_std(std)

    def p300_amplitude_handle(self, amplitude):
        if self._amplifier is not None:
            self._amplifier.set_p300_amplitude(amplitude)

    def generate_p300(self):
        if self._amplifier is not None:
            self._amplifier.generate_p300()

    def _create_window(self):
        self._window = RandomAmplifierWithSineAppWindow(self.amplifier_frequency_select_handle,
                                                        self.amplifier_sine_amplitude_handle,
                                                        self.amplifier_noise_std_handle,
                                                        self.p300_amplitude_handle,
                                                        self.generate_p300,
                                                        self.p300_autocalibration_check_handle,
                                                        self.blink_text_updated_handle)
        self.react_to = str(self._window.blink_reaction.toPlainText())

    def run(self):
        app = QApplication.instance()
        if app is None:
            app = QApplication([])
            app.setQuitOnLastWindowClosed(True)
            exec_ = True
        else:
            exec_ = False
        self._create_window()
        self._window.show()
        if exec_:
            app.exec_()
        super().run()

    async def _shutting_down(self):
        if self._window is not None:
            invoke_in_main_thread(self._window.close)
        await super()._shutting_down()


@register_amplifier_peer
class RandomAmplifierWithSinePeer(AmplifierPeer, BaseRandomAmplifierWithSinePeer):
    pass
