# -*- coding: utf-8 -*-
# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.

import logging
import time
import traceback

import numpy as np
import scipy.stats
import scipy.signal as ss
from scipy.spatial.distance import euclidean
from sklearn.exceptions import NotFittedError
from sklearn.metrics import roc_auc_score
from sklearn.metrics import roc_curve

from obci_p300.abstract_classifier import AbstractClassifier

from sklearn.discriminant_analysis import LinearDiscriminantAnalysis

from sklearn.model_selection import LeaveOneOut
from sklearn.model_selection import LeavePOut
from scipy.stats import scoreatpercentile

LOGGER = logging.getLogger('P300EasyClassifier')
LOGGER.setLevel(logging.DEBUG)

MINIMUM_SENSIBLE_EXAMPLE_COUNT = 12


class NotCalibratedError(Exception):
    pass


def critical_auc(n1, n2, p=0.01):
    """Calculating critical AUC from U statistic.

    U statistic is generated using normal distribution.
    """
    meanrank = n1 * n2 / 2.0
    T = 1
    sd = np.sqrt(T * n1 * n2 * (n1 + n2 + 1) / 12.0)
    z = scipy.stats.norm.ppf(1 - p)
    u = z * sd + meanrank
    return u / (n1 * n2)


def _tags_to_array(tags):
    """
    :param tags:
    :return:  3D numpy array from OBCI smart tags: epochs x channels x time
    """
    min_length = min(i.get_samples().shape[1] for i in tags)
    # really don't like this, but epochs generated by smart tags can vary in length by 1 sample
    array = np.dstack([i.get_samples()[:, :min_length] for i in tags])
    return np.rollaxis(array, 2)


def _remove_artifact_epochs(data, labels):
    """Provisional version.

    :param data:  3D numpy array epoch x channels x time,
    :param labels: list of epochs labels
    :return: clean data and labels
    """
    mask = np.ones(len(data), dtype=bool)
    for id, i in enumerate(data):
        if np.max(np.abs(i - i[:, 0][:, None])) > 2000:
            mask[id] = False
    newlabels = [l for l, m in zip(labels, mask) if m]
    newdata = data[mask]
    return newdata, newlabels


def _feature_extraction(data, Fs, bas=-0.1, window=0.4, targetFs=34):
    """
    :param data: 3D numpy array epoch x channels x time,
    :param Fs:  sampling in Hz
    :param bas: baseline in seconds
    :param window: timewindow after baseline to select in seconds
    :param targetFs: target sampling in Hz (will be approximated)
    :return: spatiotemporal features array epoch x features
    """
    features = []
    for epoch in data:
        features.append(feature_extraction_singular(epoch, Fs, bas,
                                                    window, targetFs))
    return np.array(features)


def feature_extraction_singular(epoch, Fs, bas=-0.1, window=0.5, targetFs=30, ):
    """Performs feature extraction on epoch (array channels x time).

    :param epoch:
    :param Fs:  sampling in Hz
    :param bas: baseline in seconds
    :param window: time-window after baseline to select in seconds
    :param targetFs: target sampling in Hz (will be approximated)
    :return: 1D array downsampled, len = downsampled samples x channels
    epoch minus mean of baseline, downsampled by factor int(Fs/targetFs)
    samples used - from end of baseline to window timepoint
    """

    detrended = scipy.signal.detrend(epoch, axis=-1, type='linear')

    mean = np.mean(detrended[:, :int(-bas * Fs)], axis=1)
    decimation_factor = int(1.0 * Fs / targetFs)
    selected = detrended[:, int(-bas * Fs):int((-bas + window) * Fs)] - mean[:, None]

    features = ss.decimate(selected, decimation_factor, axis=1, ftype='fir', zero_phase=False)
    return features.flatten()


def _feature_reduction_mask(ft, labels, mode):
    """
    :param ft: features 2d array nsamples x nfeatures
    :param labels: nsamples array of labels 0, 1,
    :param mode: 'auto', int
    :return: features mask
    """
    tscore, p = scipy.stats.ttest_ind(ft[labels == 1], ft[labels == 0])
    if mode == 'auto':
        mask = p < 0.05
        if mask.sum() < 1:
            raise Exception('Feature reduction produced zero usable features')
    elif isinstance(mode, int):
        mask_ind = np.argsort(p)[-mode:]
        mask = np.zeros_like(p, dtype=bool)
        mask[mask_ind] = True
    return mask


class LDALinearProba(LinearDiscriminantAnalysis):
    """Modified sklearn LDA, which returns predict proba with scaled linear part to training data."""

    def fit(self, X, y):
        super().fit(X, y)
        assert len(self.classes_) == 2
        decs = self.decision_function(X)

        self._lower_score = scoreatpercentile(decs, 10)
        self._upper_score = scoreatpercentile(decs, 90)

    def predict_proba(self, X):
        if not hasattr(self, 'coef_') or self.coef_ is None:
            raise NotFittedError("This %(name)s instance is not fitted "
                                 "yet" % {'name': type(self).__name__})
        assert len(self.classes_) == 2
        prob = self.decision_function(X)

        offset = self._lower_score
        scale = self._upper_score - self._lower_score
        prob -= offset
        prob /= scale

        # normalisation function is linear for decision_functions from -2 to 2
        prob *= 4
        prob += -2

        prob *= -1
        np.exp(prob, prob)
        prob += 1
        np.reciprocal(prob, prob)

        return np.column_stack([1 - prob, prob])

    def predict_decision_func(self, X):
        if not hasattr(self, 'coef_') or self.coef_ is None:
            raise NotFittedError("This %(name)s instance is not fitted "
                                 "yet" % {'name': type(self).__name__})
        assert len(self.classes_) == 2
        return self.decision_function(X)


class P300EasyClassifier(AbstractClassifier):
    """Easy and modular P300 classifier attributes"""

    def __init__(self, clf=None, learn_every: int = 20, targetFs=24, use_linear_proba=True):

        """ # relearn every # provided targets
        :param clf: scikit learn type classifier, None if you want to use standard LDA with shrinkage
        :param learn every: provided targets
        :param targetFs: target sampling rate after downsampling for feature extraction
        """
        self.creation_time = time.time()
        self._clf_params = dict(solver='lsqr', shrinkage='auto')
        if use_linear_proba:
            self._clf_type = LDALinearProba
        else:
            self._clf_type = LinearDiscriminantAnalysis
        if clf is None:
            self.clf = self._clf_type(**self._clf_params)
        else:
            self.clf = clf
        self.calibrated = False
        self.targetFs = targetFs
        self.learn_every = learn_every
        self.learning_buffor_features = list()
        self.learning_buffor_classes = list()

        # receiver operating curve
        self.roc_fpr = np.array([], ndmin=1)
        self.roc_tpr = np.array([], ndmin=1)
        self.roc_auc = 0.5
        self.p_val = 1

        self._last_n_targets_used_to_learn = 0

        self.AUC_history = []
        self.crit_AUC_history = []
        self.p_val_history = []
        self.history_n = []

        self.rocs = []
        self.roc_aucs = []
        self.n_averaged_rocs = []

    def classify(self, features):
        """
        :param features: 1D vector of extracted features
        :return: probability of target
        """
        return self.clf.predict_proba(features.reshape(1, -1))[0][1]

    def decision_func(self, features):
        return self.clf.predict_decision_func(features.reshape(1, -1))[0]

    def normalized_euclidean_distance_from_target(self, features):
        norm = euclidean(self.clf.means_[1], self.clf.means_[0]) * 10
        return np.clip((norm - euclidean(self.clf.means_[1], features)) / norm, 0, 1) ** 2

    def _fit_and_cv(self):
        LOGGER.info('Classifier: fitting and validating clf with new data')

        features = np.array(self.learning_buffor_features)
        classes = np.array(self.learning_buffor_classes)

        rocs, roc_aucs, n_averaged = self._get_roc_and_aucs_per_averages(features, classes)

        self.rocs = rocs
        self.roc_aucs = roc_aucs
        self.n_averaged_rocs = n_averaged

        validator = LeaveOneOut()
        predictions = []
        prediction_r_cls = []  # real classes in order of predictions

        tested_clf = self._clf_type(**self._clf_params)

        for train, test in validator.split(features, classes):
            tested_clf.fit(features[train], classes[train])
            predictions.extend(tested_clf.predict_proba(features[test])[:, 1])
            prediction_r_cls.extend(classes[test])

        try:

            self.roc_fpr, self.roc_tpr, _ = roc_curve(prediction_r_cls, predictions)
        except ValueError:
            traceback.print_exc()
            self.roc_tpr = [0, 1]
            self.roc_fpr = [0, 1]

        try:
            self.roc_auc = roc_auc_score(prediction_r_cls, predictions)
        except ValueError:
            traceback.print_exc()
            self.roc_auc = 0.5

        tgts = np.array(predictions)[np.array(prediction_r_cls) == 1]
        ntgts = np.array(predictions)[np.array(prediction_r_cls) == 0]

        _, p_val = scipy.stats.mannwhitneyu(tgts, ntgts, alternative='two-sided')

        self.p_val = p_val
        self.calib_tgts_prediction = tgts
        self.calib_ntgts_prediction = ntgts

        n_targets = np.sum(classes == 1)
        n_nontargets = np.sum(classes == 0)

        self.history_n.append(n_targets)
        self.AUC_history.append(self.roc_auc)
        self.crit_AUC_history.append(critical_auc(n_targets, n_nontargets))
        self.p_val_history.append(self.p_val)

        self.clf.fit(features, classes)
        self.calibrated = True
        LOGGER.info('Classifier: LOO validation ROC AUC: {}, n_targets {}, n_nontargets: {}'.format(self.roc_auc,
                                                                                                    n_targets,
                                                                                                    n_nontargets
                                                                                                    ))
        self.dec_thresholds = np.linspace(0, 1, 200)
        # percentage of nontargets which would cross the decision threshold
        self.thr_fpr = [self.fpr_for_threshold(i) for i in self.dec_thresholds]

    def _get_roc_and_aucs_per_averages(self, features, classes):
        rocs = []  # list of (roc_tpr, roc_fpr)
        roc_aucs = []  # list of floats
        n_averaged = []  # list of ints

        minimum_amount_of_examples = np.min([np.sum(classes == i) for i in np.unique(classes)])
        average_to = np.clip((minimum_amount_of_examples - MINIMUM_SENSIBLE_EXAMPLE_COUNT), 1, 10)

        for n in range(1, average_to + 1):
            roc_fpr, roc_tpr, roc_auc = self._roc_curve_from_cv_averaged(features, classes, n)
            rocs.append((roc_tpr, roc_fpr))
            roc_aucs.append(roc_auc)
            n_averaged.append(n)

        return rocs, roc_aucs, n_averaged

    def _roc_curve_from_cv_averaged(self, features, classes, n, max_iters=100):

        predictions = []
        prediction_r_cls = []  # real classes in order of predictions

        tested_clf = self._clf_type(**self._clf_params)

        target_mask = (classes == 1)
        nontarget_mask = (classes == 0)

        for mask in [target_mask, nontarget_mask]:
            validator = LeavePOut(n)
            validator_gen = validator.split(features[mask], classes[mask])
            n_validator_splits = validator.get_n_splits(features[mask], classes[mask])
            if n_validator_splits > 1000:
                self._random_permutation_cv_backend(max_iters, features, mask, classes, tested_clf,
                                                    predictions, prediction_r_cls, n)
            else:
                self._full_permutation_cv_backend(max_iters, n_validator_splits, validator_gen, features, classes,
                                                  mask, tested_clf, predictions, prediction_r_cls)
        try:

            roc_fpr, roc_tpr, _ = roc_curve(prediction_r_cls, predictions)
        except ValueError:
            traceback.print_exc()
            roc_tpr = [0, 1]
            roc_fpr = [0, 1]

        try:
            roc_auc = roc_auc_score(prediction_r_cls, predictions)
        except ValueError:
            traceback.print_exc()
            roc_auc = 0.5

        return roc_fpr, roc_tpr, roc_auc

    def _random_permutation_cv_backend(self, max_iters, features, mask, classes, tested_clf, predictions,
                                       prediction_r_cls, n):
        # better for high amounts of samples
        for i in range(int(max_iters / 2)):
            permuted_features = np.random.permutation(features[mask])
            permuted_training_features = permuted_features[n:]
            permuted_testing_features = permuted_features[:n]

            training_features = np.concatenate([permuted_training_features, features[np.logical_not(mask)]])
            training_classes = np.concatenate([classes[mask][n:], classes[np.logical_not(mask)]])
            tested_clf.fit(training_features, training_classes)
            averaged_sample = np.mean(permuted_testing_features, axis=0)
            predictions.extend(tested_clf.predict_proba([averaged_sample])[:, 1])
            prediction_r_cls.append(classes[mask][0])

    def _full_permutation_cv_backend(self, max_iters, n_validator_splits, validator_gen, features, classes, mask,
                                     tested_clf, predictions, prediction_r_cls):
        split_weight = max_iters / 2 / n_validator_splits
        split_id_weighted = 0
        old_split_id = 0
        for train, test in validator_gen:  # better for small amounts of samples
            split_id_weighted += split_weight
            if int(split_id_weighted) != old_split_id:
                old_split_id = int(split_id_weighted)
                training_features = np.concatenate([features[mask][train], features[np.logical_not(mask)]])
                training_classes = np.concatenate([classes[mask][train], classes[np.logical_not(mask)]])
                tested_clf.fit(training_features, training_classes)
                averaged_sample = np.mean(features[mask][test], axis=0)
                predictions.extend(tested_clf.predict_proba([averaged_sample])[:, 1])
                prediction_r_cls.append(classes[mask][test][0])

    def fpr_for_threshold(self, thr):
        """Calculate False Positive Rate, for given classifier target probability threshold.

        :param thr: float [0;1] - classifier threshold
        """
        if not self.calibrated:
            raise NotCalibratedError("Classifier is not fitted yet, run calibration")

        ntgts = self.calib_ntgts_prediction
        percentile_at_threshold = scipy.stats.percentileofscore(ntgts, thr)
        # we return fraction of nontargets which are to the right of the threshold
        return 1 - percentile_at_threshold / 100

    def threshold_for_fpr(self, fpr, average_with_fnr_if_higher=True):
        """Calculate target probability threshold, for given classifier False Positive Rate.

        :param fpr: float [0;1] target False Positive Rate
        :param average_with_fnr_if_higher better threshold at high AUC using 0.15 FNR and averaging trick
        """
        fnr = 0.15

        if not self.calibrated:
            raise NotCalibratedError("Classifier is not fitted yet, run calibration")
        ntgts = self.calib_ntgts_prediction
        tgts = self.calib_tgts_prediction
        percentile_at_fpr = 100.0 * (1 - fpr)
        percentile_at_fnr = 100.0 * fnr
        threshold_fpr = scipy.stats.scoreatpercentile(ntgts, percentile_at_fpr)
        threshold_fnr = scipy.stats.scoreatpercentile(tgts, percentile_at_fnr)

        if (threshold_fnr > threshold_fpr) and average_with_fnr_if_higher:
            threshold = 0.5 * threshold_fnr + 0.5 * threshold_fpr
        else:
            threshold = threshold_fpr
        return threshold

    def _learn_every_condition(self):
        n_targets = sum(self.learning_buffor_classes)
        if self._last_n_targets_used_to_learn == n_targets:
            return False
        self._last_n_targets_used_to_learn = n_targets
        return sum(self.learning_buffor_classes) % self.learn_every == 0

    def learn(self, chunk, target, cv_start_callback=None):
        """For online learning thread.

        :param chunk: numpy 1D features array
        :param target: name of the class
        :return score: if learned or None, if not
        """
        if target == 'target':
            self.learning_buffor_classes.append(1)
        else:
            self.learning_buffor_classes.append(0)
        self.learning_buffor_features.append(chunk)
        targets_count = sum(self.learning_buffor_classes)
        nontargets_count = sum(i == 0 for i in self.learning_buffor_classes)
        enough = (targets_count > 4 and nontargets_count > 3)
        if self._learn_every_condition() and enough:
            if cv_start_callback is not None:
                cv_start_callback()
            self._fit_and_cv()
            return self.roc_auc

        return None

    def calibrate(self, targets, nontargets, bas=-0.1, window=0.4, Fs=None):
        """Offline learning function.

        :param targets: 3D arrays (epoch x channel x time) or list of OBCI smart tags. If arrays - need to provide Fs
        (sampling frequency) in Hz
        :param nontargets: 3D arrays (epoch x channel x time) or list of OBCI smart tags. If arrays - need to provide Fs
        (sampling frequency) in Hz
        :param bas: baseline in seconds(negative), in other words start offset
        :param window: seconds after event to consider in classification
        :param Fs: needed if 3D numpy array is provided
        :return score
        """
        if Fs is None:
            Fs = float(targets[0].get_param('sampling_frequency'))
            target_data = _tags_to_array(targets)
            nontarget_data = _tags_to_array(nontargets)
        data = np.vstack((target_data, nontarget_data))
        self.epoch_l = data.shape[2]
        labels = np.zeros(len(data))
        labels[:len(target_data)] = 1
        data, labels = _remove_artifact_epochs(data, labels)
        features = _feature_extraction(data, Fs, bas, window, self.targetFs)
        self.clf.fit(features, labels)
        # building a list of features
        # will be saved and can be extended in online learning later
        self.learning_buffor_classes = [i for i in labels]
        self.learning_buffor_features = [i for i in features]
        self.calibrated = True
        return self.clf.score(features, labels)
