import enum
import os
import pathlib
import shlex
import subprocess
import sys
import tempfile
import typing
from collections import OrderedDict

from fabric import api
from fabric import context_managers

try:
    import devops  # noqa
except ImportError:
    print("devops module not found! Not all commands are available!\n"
          "Install devops module using https://pip.braintech.pl/obci/stable", file=sys.stderr)
    devpi = None
else:
    from devops.ci import install
    from devops.ci import output
    from devops import devpi

    sys.stdout = output.unbuffered_stdout


@enum.unique
class ReleaseType(enum.Enum):
    NIGHTLY = 'nightly'
    TESTING = 'testing'
    MASTER = 'master'


PACKAGES = OrderedDict([
])

APT_REQUIREMENTS_PACKAGES = {
    'common': ['git', 'python3-pip'],
    'codestyle': ([]),
    'test': install.COMMON_APT_DEPENDENCIES[install.Stage.TEST] + [
        'python3-tk',
        # tests
        'svarog',
        'scrot',
        # to build cpp amps
        'libftdi1-dev',
        'libftdi-dev',
        'libbluetooth-dev',
    ],
    'deploy': install.COMMON_APT_DEPENDENCIES[install.Stage.DEPLOY],
    'docs': []
}
try:
    APT_REQUIREMENTS_PACKAGES['test'].remove('python3-xaut')
except ValueError:
    pass
PIP_REQUIREMENTS = {
    'build': ['setuptools'],
    'codestyle': [
        'flake8~=3.0.4',
        'flake8_docstrings~=1.0.2',
        'pydocstyle~=1.0.0',
        'pylint~=1.7.2',
    ],
    'test': [
        'flaky~=3.4.0',
        'mne~=0.17.0',
        'pytest~=3.1.2',
        'pytest-cov~=2.6.0',
        'pytest-qt~=2.4.1',
        'pytest-timeout~=1.2.0',
        'pytest-env==0.6.2',
        'nose~=1.3.7',
        'python-redmine',
        'python-xlib',
        'xaut==0.5.0.post2',
        'psychopy~=3.0.0',
        'Cython==0.29.13',
    ],
    'docs': [
        'setuptools~=40.0',
        'Sphinx~=1.4.8',
        'sphinx-rtd-theme~=0.1.9',
        'sphinxcontrib-asyncio~=0.2.0',
    ]
}


def check_obci_with_pylint():
    check_command = 'git diff --name-only origin/development | grep ".py$"'
    changed_python_files_raw = api.local(check_command, capture=True).stdout
    api.puts('Changed files: \n{}'.format(changed_python_files_raw))
    if changed_python_files_raw:
        files_inline = ' '.join(changed_python_files_raw.splitlines())
        with context_managers.settings(warn_only=True):
            result = api.local('pylint {}'.format(files_inline),
                               capture=True)
        api.puts(result.stdout)
        api.puts(result.stderr)
        pylint_no_error_return_code = 0
        pylint_convention_error_return_code = 16
        allowed_return_codes = [pylint_no_error_return_code,
                                pylint_convention_error_return_code]
        if result.return_code not in allowed_return_codes:
            api.puts('List of return codes: '
                     'https://pylint.readthedocs.io/en/latest/user_guide/run.html'
                     '?highlight=return%20codes#exit-codes')
            api.abort("Not allowed return code: {}.".format(result.return_code))
    else:
        api.puts("No *.py files changed.")


def check_obci_codestyle():
    api.local('flake8')

#TODO REPLACE HERE
def check_obci_docstyle():
    api.local('flake8 '
              '--enable-extensions=D '
              '--select D braintech.obci.core.broker/ obci/peers/ obci/peers/test/ '
              'obci/utils/ obci/signal_processing/ '
              'obci/control/gui/report_error.py')


def build_obci():
    build_pip_packages()


def build_pip_packages():
    devpi.use_branch_index(tmp=True)
    # when building deb package, set dsn from Gitlab-CI Project Variable
    api.local('sed -i "s#^SENTRY_DSN = \'\'\$#SENTRY_DSN = \'$OBCI_SENTRY_TEMPLATE_DSN\'#g" '
              './braintech-obci-lab/braintech/obci/lab/settings.py')
    api.local('mkdir -p packages')
    for package, package_path in PACKAGES.items():
        devpi.upload(package_path)
        api.local('mv %s/dist/* ./packages/' % package_path)


def generate_obci_leandocs():
    api.local('mkdir -p leandocs')
    rst_files_to_copy = ','.join([
        'introduction',
        'bci_framework_commands',
        'new_peer_architecture',
        'peer_configuration',
        'messages',
        'bci_framework_and_guide',
        'peer_states',
    ])
    api.local('cp -r docs/_static docs/images docs/{%s}.rst docs/Makefile docs/conf.py docs/make.bat leandocs'
              % rst_files_to_copy, shell='/bin/bash')
    api.local('cp docs/lean/index.rst leandocs')
    with context_managers.lcd('leandocs'):
        api.local('patch < ../docs/lean/conf.patch')
    api.local('sphinx-apidoc --separate -o leandocs/apidoc obci')
    with context_managers.lcd('leandocs'):
        api.local('make html')
        api.local('rm -rf _build/html/apidoc')


def generate_obci_docs():
    api.local('sphinx-apidoc --separate -o docs/apidoc obci')
    api.local('make --directory=docs html')


def run_tests(mark_expression=''):
    # Run graphic tests in X virtual framebuffer
    base_command = ('xvfb-run -e/tmp/xvfb.err -a -s "-screen 0 1400x900x24 -ac '
                    '+extension RANDR +extension GLX +render -noreset" -- '
                    'py.test')
    if os.name == 'nt':
        base_command = 'pytest'
    if mark_expression:
        api.local('{} -m "{}"'.format(base_command, mark_expression.replace('-', ' ')))
    else:
        api.local(base_command)


def test_obci_downstream():
    api.local('apt-get install python3-pip')
    api.local('pip3 install requests==2.18.4')
    api.local('python3 ./scripts/ci/tests_downstream.py')


def deploy_obci_docs(release_type_raw):
    release_type = ReleaseType(release_type_raw)
    env = _get_deployment_env_variables(release_type)
    with context_managers.shell_env(**env):
        _deploy_docs(release_type)


def _deploy_docs(release_type: ReleaseType):
    with context_managers.prefix('set -u'):
        if release_type == ReleaseType.MASTER:
            obci_version = api.local('git describe --tag --first-parent | head -c 3',
                                     capture=True)
            with context_managers.shell_env(OBCI_VERSION=str(obci_version)):
                api.local('sshpass -p $SSHPASS ssh -o StrictHostKeyChecking=no '
                          '$DEB_USER@$DEB_SRV '
                          '"mkdir -p -v ~/incoming/docs/$OBCI_VERSION"')
                api.local('sshpass -p $SSHPASS scp -r -o StrictHostKeyChecking=no '
                          '-v ./docs/_build/html/* '
                          '$DEB_USER@$DEB_SRV:~/incoming/docs/$OBCI_VERSION')
                api.local('sshpass -p $SSHPASS ssh -o StrictHostKeyChecking=no '
                          '$DEB_USER@$DEB_SRV '
                          '"bash ~/docs_rebuild.sh $OBCI_VERSION"')
        else:
            api.local('sshpass -p $SSHPASS ssh -o StrictHostKeyChecking=no '
                      '$DEB_USER@$DEB_SRV '
                      '"mkdir -p -v ~/incoming/docs/"')
            api.local('sshpass -p $SSHPASS scp -r -o StrictHostKeyChecking=no '
                      '-v ./docs/_build/html/* '
                      '$DEB_USER@$DEB_SRV:~/incoming/docs/')
            api.local('sshpass -p $SSHPASS ssh -o StrictHostKeyChecking=no '
                      '$DEB_USER@$DEB_SRV '
                      '"rm -r ~/docs && mv ~/incoming/docs ~/docs"')


def deploy_pip_packages():
    devpi_index = devpi.current_index()['name']
    devpi.use_branch_index(tmp=True)
    for package, dir in PACKAGES.items():
        devpi.push(dir, devpi_index)


def _get_deployment_env_variables(release_type: ReleaseType) -> typing.Dict[str, str]:
    env = dict(RELEASE='xenial')
    envs_per_type = {
        ReleaseType.NIGHTLY: dict(
            DEB_SRV=os.environ.get('DEB_NIGHTLY_SRV'),
            DEB_USER=os.environ.get('DEB_NIGHTLY_USER'),
            SSHPASS=os.environ.get('DEB_NIGHTLY_PASSWD'),
        ),
        ReleaseType.TESTING: dict(
            DEB_SRV=os.environ.get('DEB_TESTING_SRV'),
            DEB_USER=os.environ.get('DEB_TESTING_USER'),
            SSHPASS=os.environ.get('DEB_TESTING_PASSWD'),
        ),
        ReleaseType.MASTER: dict(
            DEB_SRV=os.environ.get('DEB_STABLE_SRV'),
            DEB_USER=os.environ.get('DEB_STABLE_USER'),
            SSHPASS=os.environ.get('DEB_STABLE_PASSWD'),
        ),
    }
    env.update(envs_per_type[release_type])
    assert all(v is not None for v in env.values()), "Environment values are missing: {}".format(env)
    return env


def setup_stage(stage):
    add_stable_deb_repo()
    install.setup_stage(stage, APT_REQUIREMENTS_PACKAGES, PIP_REQUIREMENTS)
    if stage == 'test':
        install.run_in_virtual_x_server('pip3 install pyautogui')
        for path in PACKAGES.values():
            install.local("pip3 install -e {path}".format(path=path))
        api.local('install_all')  # Install all obci dependencies


def deploy_obci_leandocs():
    with context_managers.lcd('leandocs/_build'):
        api.local('mv html BCIframework')
        api.local('''lftp<<EOT
        open -u $LEANMANUALUSER,$LEANMANUALPSWD $LEANMANUALHOST
        mkdir BCIframework
        cd BCIframework
        mirror -R -n BCIframework/ .
        quit
        EOT''')


def rebuild_image_latest():
    command_fmt = 'docker build --no-cache ' \
                  '-t obci-image:latest {path}'
    with tempfile.TemporaryDirectory() as tmpdir_name:
        generate_dockerfile(str(pathlib.Path(tmpdir_name) / 'Dockerfile'))
        command = command_fmt.format(
            path=tmpdir_name,
        )
        subprocess.run(shlex.split(command), check=True)


def generate_dockerfile(path):
    aggregator = _generate_collected_obci_requirements()
    for reqs in PIP_REQUIREMENTS.values():
        aggregator.collect([(install.PIP3, r) for r in reqs])
    uninstall = ['braintech-obci-experiment', 'braintech-obci-core', 'braintech-obci-signal-processing',
                 'braintech-obci-lab', 'devops']
    for package, package_path in PACKAGES.items():
        sys.path.insert(0, package_path)
        from setup import INSTALL_REQUIREMENTS
        del sys.modules['setup']
        sys.path.pop(0)
        for reqs in INSTALL_REQUIREMENTS.values():
            aggregator.collect([(install.PIP3, r) for r in reqs if r not in PACKAGES])
        uninstall.append(package)
    template = """FROM devops:latest
RUN apt-get -qq update && apt-get install -y \\
    {apt_requirements} \\
    && rm -rf /var/lib/apt/lists/*
RUN pip2 install \\
    {pip2_requirements} \\
    --index {DEVPI_INDEX}
RUN xvfb-run -e/tmp/xvfb.err -a -s "{xvfb_settings}" -- pip3 install \\
    {pip3_requirements} \\
    --index {DEVPI_INDEX}
RUN pip3 uninstall -y {uninstall}""".format(
        apt_requirements=aggregator.flatten(install.APT),
        pip2_requirements=aggregator.flatten(install.PIP2),
        pip3_requirements=aggregator.flatten(install.PIP3),
        xvfb_settings=('-screen 0 1400x900x24 -ac +extension RANDR '  # PyAutoGui
                       '+extension GLX +render -noreset'),
        DEVPI_INDEX=os.environ['DEVPI_URL'] + '/' + devpi.current_index()['name'],
        uninstall=" ".join(uninstall)
    )
    open(path, 'w').write(template)
    return template


def _generate_collected_obci_requirements():
    aggregator = install.RequirementsAggregator()
    for reqs in APT_REQUIREMENTS_PACKAGES.values():
        aggregator.collect([(install.APT, r) for r in reqs])
    return aggregator


def install_editable():
    for package, dir in PACKAGES.items():
        install.local('pip3 install  -e %s/.' % dir)


def setup_dev():
    if 'DEVPI_URL' not in os.environ:
        print("Provide DEVPI_URL environment variable!\n"
              "(https://redmine.titanis.pl/projects/brain4/wiki/Has%C5%82a_do_repozytori%C3%B3w#Devpi)")
        return
    devpi_index = "{DEVPI_URL}/obci/dev".format(**os.environ)

    global devpi
    global install
    if devpi is None:
        user = '' if os.name == 'nt' else '--user'
        api.local("pip3 install {user} --index {devpi_index} devops".format(
            user=user,
            devpi_index=devpi_index
        ))
        from devops import devpi
        from devops.ci import install

    aggregator = _generate_collected_obci_requirements()
    install.local("apt-get install " + aggregator.flatten(install.APT))
    devpi.use_index(devpi_index)
    install_editable()


def add_stable_deb_repo():
    install.local(
        'echo "deb http://$HTTP_USER:$HTTP_PASSWD@obci-stable.braintech.pl/xenial /"'
        '> /etc/apt/sources.list.d/obci-stable-xenial.list')
    install.local('apt-key adv --keyserver keyserver.ubuntu.com --recv-keys 1452AFDF')
